@@ -4,15 +4,19 @@
 #ifndef NTP_H
 #define NTP_H
 
-#include "ntp_types.h"
+#include <stddef.h>
 #include <math.h>
+
+#include <ntp_fp.h>
+#include <ntp_types.h>
+#include <ntp_stdlib.h>
 #ifdef OPENSSL
-#include "ntp_crypto.h"
+#include <ntp_crypto.h>
 #endif /* OPENSSL */
 #include <ntp_random.h>
+#include <ntp_net.h>
 
 #include <isc/boolean.h>
-#include <isc/list.h>
 
 /*
  * Calendar arithmetic - contributed by G. Healton
@@ -124,20 +128,20 @@ typedef char s_char;
 #define	NTP_MINCLOCK	3	/* min survivors */
 #define	NTP_MAXCLOCK	10	/* max candidates */
 #define	NTP_MAXASSOC	50	/* max associations */
-#define MINDISPERSE	.005	/* min dispersion increment */
+#define MINDISPERSE	.001	/* min distance */
 #define MAXDISTANCE	1.5	/* max root distance (select threshold) */
 #define CLOCK_SGATE	3.	/* popcorn spike gate */
 #define HUFFPUFF	900	/* huff-n'-puff sample interval (s) */
 #define MAXHOP		2	/* anti-clockhop threshold */
 #define MAX_TTL		8	/* max ttl mapping vector size */
 #define	BEACON		7200	/* manycast beacon interval */
-#define NTP_MAXEXTEN	1024	/* max extension field size */
+#define NTP_MAXEXTEN	2048	/* max extension field size */
 
 /*
  * Miscellaneous stuff
  */
 #define NTP_MAXKEY	65535	/* max authentication key number */
-
+#define	KEY_TYPE_MD5	NID_md5	/* MD5 digest NID */
 /*
  * Limits of things
  */
@@ -152,7 +156,7 @@ typedef char s_char;
  * dispersion component (frequency error plus precision). The frequency
  * error component is computed as CLOCK_PHI times the difference between
  * the epoch of the time measurement and the reference time. The
- * precision componen is computed as the square root of the mean of the
+ * precision component is computed as the square root of the mean of the
  * squares of a zero-mean, uniform distribution of unit maximum
  * amplitude. Whether this makes statistical sense may be arguable.
  */
@@ -166,34 +170,40 @@ typedef char s_char;
 
 #define	EVENT_TIMEOUT	0	/* one second, that is */
 
+
 /*
  * The interface structure is used to hold the addresses and socket
- * numbers of each of the interfaces we are using.
+ * numbers of each of the local network addresses we are using.
+ * Because "interface" is a reserved word in C++ and has so many
+ * varied meanings, a change to "endpt" (via typedef) is under way.
+ * Eventually the struct tag will change from interface to endpt_tag.
+ * endpt is unrelated to the select algorithm's struct endpoint.
  */
+typedef struct interface endpt;
 struct interface {
-	SOCKET fd;			/* socket this is opened on */
-	SOCKET bfd;			/* socket for receiving broadcasts */
-	struct sockaddr_storage sin;	/* interface address */
-	struct sockaddr_storage bcast;	/* broadcast address */
-	struct sockaddr_storage mask;	/* interface mask */
-	char name[32];			/* name of interface */
-	short  family;			/* Address family */
-	int flags;			/* interface flags */
-	int last_ttl;			/* last TTL specified */
-	u_int32 addr_refid;		/* IPv4 addr or IPv6 hash */
-	int num_mcast;			/* No. of IP addresses in multicast socket */
-	u_long starttime;	        /* current_time as of creation of interface structure */
-	volatile long received;		/* number of incoming packets */
-	long sent;			/* number of outgoing packets */
-	long notsent;			/* number of send failures */
-	u_int scopeid;			/* Scope used for Multicasting */
-	u_int ifindex;			/* interface index */
-	u_int ifnum;		        /* sequential interface instance count */
-        u_char phase;		        /* phase in update cycle */
-	isc_boolean_t ignore_packets;	/* Specify whether the packet should be ignored */
-        ISC_LIST(struct peer) peers;    /* list of peers for the interface */
-        u_int peercnt;		        /* number of peers referencinf this interface - informational only */
-        ISC_LINK(struct interface) link;     /* interface list */
+	endpt *		elink;		/* endpt list link */
+	endpt *		mclink;		/* per-AF_* multicast list */
+	SOCKET		fd;		/* socket descriptor */
+	SOCKET		bfd;		/* for receiving broadcasts */
+	u_int32		ifnum;		/* endpt instance count */
+	sockaddr_u	sin;		/* unicast address */
+	sockaddr_u	mask;		/* subnet mask */
+	sockaddr_u	bcast;		/* broadcast address */
+	char		name[32];	/* name of interface */
+	u_short		family;		/* AF_INET/AF_INET6 */
+	u_short		phase;		/* phase in update cycle */
+	u_int32		flags;		/* interface flags */
+	int		last_ttl;	/* last TTL specified */
+	u_int32		addr_refid;	/* IPv4 addr or IPv6 hash */
+	int		num_mcast;	/* mcast addrs enabled */
+	u_long		starttime;	/* current_time at creation */
+	volatile long	received;	/* number of incoming packets */
+	long		sent;		/* number of outgoing packets */
+	long		notsent;	/* number of send failures */
+	u_int		ifindex;	/* for IPV6_MULTICAST_IF */
+	isc_boolean_t	ignore_packets; /* listen-read-drop this? */
+	struct peer *	peers;		/* list of peers using endpt */
+	u_int		peercnt;	/* count of same */
 };
 
 /*
@@ -206,8 +216,10 @@ struct interface {
 #define INT_MULTICAST	0x010	/* can multicast out this interface */
 #define	INT_BCASTOPEN	0x020	/* broadcast socket is open */
 #define INT_MCASTOPEN	0x040	/* multicasting enabled */
-#define INT_WILDCARD    0x080   /* wildcard interface - usually skipped */
-#define INT_MCASTIF     0x100	/* bound directly to MCAST address */
+#define INT_WILDCARD	0x080	/* wildcard interface - usually skipped */
+#define INT_MCASTIF	0x100	/* bound directly to MCAST address */
+#define INT_PRIVACY	0x200	/* RFC 4941 IPv6 privacy address */
+
 /*
  * Define flasher bits (tests 1 through 11 in packet procedure)
  * These reveal the state at the last grumble from the peer and are
@@ -242,11 +254,11 @@ struct interface {
  * spec.
  */
 struct peer {
-	struct peer *next;	/* pointer to next association */
+	struct peer *next;	/* link pointer in peer hash */
 	struct peer *ass_next;	/* link pointer in associd hash */
-	struct sockaddr_storage srcadr; /* address of remote host */
-	struct interface *dstadr; /* pointer to address on local host */
-        ISC_LINK(struct peer) ilink; /* interface link list */
+	struct peer *ilink;	/* list of peers for interface */
+	sockaddr_u srcadr;	/* address of remote host */
+	endpt *	dstadr;		/* local address */
 	associd_t associd;	/* association ID */
 	u_char	version;	/* version number */
 	u_char	hmode;		/* local association mode */
@@ -322,24 +334,30 @@ struct peer {
 	 * Ephemeral state variables
 	 */
 	u_char	status;		/* peer status */
+	u_char	new_status;	/* under-construction status */
 	u_char	reach;		/* reachability register */
 	int	flash;		/* protocol error test tally bits */
 	u_long	epoch;		/* reference epoch */
 	int	burst;		/* packets remaining in burst */
 	int	retry;		/* retry counter */
+	int	flip;		/* interleave mode control */
 	int	filter_nextpt;	/* index into filter shift register */
 	double	filter_delay[NTP_SHIFT]; /* delay shift register */
 	double	filter_offset[NTP_SHIFT]; /* offset shift register */
 	double	filter_disp[NTP_SHIFT]; /* dispersion shift register */
 	u_long	filter_epoch[NTP_SHIFT]; /* epoch shift register */
 	u_char	filter_order[NTP_SHIFT]; /* filter sort index */
-	l_fp	org;		/* originate time stamp */
 	l_fp	rec;		/* receive time stamp */
 	l_fp	xmt;		/* transmit time stamp */
+	l_fp	dst;		/* destination timestamp */
+	l_fp	aorg;		/* origin timestamp */
+	l_fp	borg;		/* alternate origin timestamp */
 	double	offset;		/* peer clock offset */
 	double	delay;		/* peer roundtrip delay */
 	double	jitter;		/* peer jitter (squares) */
 	double	disp;		/* peer dispersion */
+	double	xleave;		/* interleave delay */
+	double	bias;		/* bias for NIC asymmetry */
 
 	/*
 	 * Variables used to correct for packet length and asymmetry.
@@ -424,19 +442,22 @@ struct peer {
  * Values for peer.flags
  */
 #define	FLAG_CONFIG	0x0001	/* association was configured */
-#define FLAG_PREEMPT	0x0002	/* preemptable association */
+#define	FLAG_PREEMPT	0x0002	/* preemptable association */
 #define	FLAG_AUTHENTIC	0x0004	/* last message was authentic */
 #define	FLAG_REFCLOCK	0x0008	/* this is actually a reference clock */
-#define	FLAG_SYSPEER	0x0010	/* system peer */
-#define FLAG_PREFER	0x0020	/* prefer peer */
-#define FLAG_BURST	0x0040	/* burst mode */
-#define FLAG_FIXPOLL	0x0080	/* stick at minpoll */
-#define FLAG_IBURST	0x0100	/* initial burst mode */
-#define FLAG_NOSELECT	0x0200	/* never select */
-#define FLAG_TRUE	0x0400	/* force truechimer */
-#ifdef OPENSSL
-#define FLAG_SKEY	0x1000  /* autokey authentication */
-#define FLAG_ASSOC	0x2000	/* autokey request */
+#define	FLAG_BC_VOL	0x0010	/* broadcast client volleying */
+#define	FLAG_PREFER	0x0020	/* prefer peer */
+#define	FLAG_BURST	0x0040	/* burst mode */
+#define	FLAG_PPS	0x0080	/* steered by PPS */
+#define	FLAG_IBURST	0x0100	/* initial burst mode */
+#define	FLAG_NOSELECT	0x0200	/* never select */
+#define	FLAG_TRUE	0x0400	/* force truechimer */
+#define	FLAG_SKEY	0x0800  /* autokey authentication */
+#define	FLAG_XLEAVE	0x1000	/* interleaved protocol */
+#define	FLAG_XB		0x2000	/* interleaved broadcast */
+#define	FLAG_XBOGUS	0x4000	/* interleaved bogus packet */
+#ifdef	OPENSSL
+#define FLAG_ASSOC	0x8000	/* autokey request */
 #endif /* OPENSSL */
 
 /*
@@ -507,92 +528,6 @@ struct peer {
 #define REFCLK_NEOCLOCK4X	44	/* NeoClock4X DCF77 or TDF receiver */
 #define REFCLK_MAX		44	/* NeoClock4X DCF77 or TDF receiver */
 
- /*
- * Macro for sockaddr_storage structures operations
- */
-#ifdef ISC_PLATFORM_HAVESCOPEID
-#define SOCKSCOPE(sock1, sock2) (((struct sockaddr_in6 *)sock1)->sin6_scope_id == ((struct sockaddr_in6 *)sock2)->sin6_scope_id)
-#else
-#define SOCKSCOPE(sock1, sock2) 1
-#endif
-
-#define SOCKCMP(sock1, sock2) \
-	(((struct sockaddr_storage *)sock1)->ss_family \
-	    == ((struct sockaddr_storage *)sock2)->ss_family ? \
- 	((struct sockaddr_storage *)sock1)->ss_family == AF_INET ? \
- 	memcmp(&((struct sockaddr_in *)sock1)->sin_addr, \
-	    &((struct sockaddr_in *)sock2)->sin_addr, \
-	    sizeof(struct in_addr)) == 0 : \
-	memcmp(&((struct sockaddr_in6 *)sock1)->sin6_addr, \
-	    &((struct sockaddr_in6 *)sock2)->sin6_addr, \
-	       sizeof(struct in6_addr)) == 0 && SOCKSCOPE(sock1, sock2) : \
-	0)
-
-#define SOCKNUL(sock1) \
-	(((struct sockaddr_storage *)sock1)->ss_family == AF_INET ? \
- 	(((struct sockaddr_in *)sock1)->sin_addr.s_addr == 0) : \
- 	(IN6_IS_ADDR_UNSPECIFIED(&((struct sockaddr_in6 *)sock1)->sin6_addr)))
-
-#define SOCKLEN(sock) \
-	(((struct sockaddr_storage *)sock)->ss_family == AF_INET ? \
- 	(sizeof(struct sockaddr_in)) : (sizeof(struct sockaddr_in6)))
-
-#define ANYSOCK(sock) \
-	memset(((struct sockaddr_storage *)sock), 0, \
-	    sizeof(struct sockaddr_storage))
-
-#define ANY_INTERFACE_CHOOSE(sock) \
-	(((struct sockaddr_storage *)sock)->ss_family == AF_INET ? \
- 	any_interface : any6_interface)
-
-/*
- * We tell reference clocks from real peers by giving the reference
- * clocks an address of the form 127.127.t.u, where t is the type and
- * u is the unit number.  We define some of this here since we will need
- * some sanity checks to make sure this address isn't interpretted as
- * that of a normal peer.
- */
-#define	REFCLOCK_ADDR	0x7f7f0000	/* 127.127.0.0 */
-#define	REFCLOCK_MASK	0xffff0000	/* 255.255.0.0 */
-
-#define	ISREFCLOCKADR(srcadr)	((SRCADR(srcadr) & REFCLOCK_MASK) \
-					== REFCLOCK_ADDR)
-
-/*
- * Macro for checking for invalid addresses.  This is really, really
- * gross, but is needed so no one configures a host on net 127 now that
- * we're encouraging it the the configuration file.
- */
-#define	LOOPBACKADR	0x7f000001
-#define	LOOPNETMASK	0xff000000
-
-#define	ISBADADR(srcadr)	(((SRCADR(srcadr) & LOOPNETMASK) \
-				    == (LOOPBACKADR & LOOPNETMASK)) \
-				    && (SRCADR(srcadr) != LOOPBACKADR))
-
-/*
- * Utilities for manipulating addresses and port numbers
- */
-#define	NSRCADR(src)	(((struct sockaddr_in *)src)->sin_addr.s_addr) /* address in net byte order */
-#define	NSRCPORT(src)	(((struct sockaddr_in *)src)->sin_port)	/* port in net byte order */
-#define	SRCADR(src)	(ntohl(NSRCADR((src))))	/* address in host byte order */
-#define	SRCPORT(src)	(ntohs(NSRCPORT((src))))	/* host port */
-
-#define CAST_V4(src)	((struct sockaddr_in *)&(src))
-#define CAST_V6(src)	((struct sockaddr_in6 *)&(src))
-#define GET_INADDR(src)  (CAST_V4(src)->sin_addr.s_addr)
-#define GET_INADDR6(src) (CAST_V6(src)->sin6_addr)
-
-#define SET_HOSTMASK(addr, family)	\
-	do { \
-		memset((char *)(addr), 0, sizeof(struct sockaddr_storage)); \
-		(addr)->ss_family = (family); \
-		if ((family) == AF_INET) \
-			GET_INADDR(*(addr)) = 0xffffffff; \
-		else \
-			memset(&GET_INADDR6(*(addr)), 0xff, \
-			    sizeof(struct in6_addr)); \
-	} while(0)
 
 /*
  * NTP packet format.  The mac field is optional.  It isn't really
@@ -615,18 +550,18 @@ struct pkt {
 	l_fp	rec;		/* receive time stamp */
 	l_fp	xmt;		/* transmit time stamp */
 
-#define	LEN_PKT_NOMAC	12 * sizeof(u_int32) /* min header length */
-#define	LEN_PKT_MAC	LEN_PKT_NOMAC +  sizeof(u_int32)
-#define MIN_MAC_LEN	3 * sizeof(u_int32)	/* DES */
-#define MAX_MAC_LEN	5 * sizeof(u_int32)	/* MD5 */
+#define	LEN_PKT_NOMAC	(12 * sizeof(u_int32)) /* min header length */
+#define MIN_MAC_LEN	(1 * sizeof(u_int32))	/* crypto_NAK */
+#define MAX_MD5_LEN	(5 * sizeof(u_int32))	/* MD5 */
+#define	MAX_MAC_LEN	(6 * sizeof(u_int32))	/* SHA */
 
 	/*
 	 * The length of the packet less MAC must be a multiple of 64
-	 * with an RSA modulus and Diffie-Hellman prime of 64 octets
+	 * with an RSA modulus and Diffie-Hellman prime of 256 octets
 	 * and maximum host name of 128 octets, the maximum autokey
 	 * command is 152 octets and maximum autokey response is 460
 	 * octets. A packet can contain no more than one command and one
-	 * response, so the maximum total extension field length is 672
+	 * response, so the maximum total extension field length is 864
 	 * octets. But, to handle humungus certificates, the bank must
 	 * be broke.
 	 */
@@ -685,6 +620,10 @@ struct pkt {
 #define	EVNT_LEAP	11	/* leap event */
 #define	EVNT_CLOCKRESET	12	/* clock step */
 #define	EVNT_KERN	13	/* kernel event */
+#define	EVNT_TAI	14	/* TAI */
+#define	EVNT_LEAPVAL	15	/* stale leapsecond values */
+#define	EVNT_CLKHOP	16	/* clockhop */
+
 /*
  * Peer event codes
  */
@@ -697,10 +636,13 @@ struct pkt {
 #define	PEVNT_RATE	(7 | PEER_EVENT) /* rate exceeded */
 #define	PEVNT_DENY	(8 | PEER_EVENT) /* access denied */
 #define PEVNT_ARMED	(9 | PEER_EVENT) /* leap armed */
-#define	PEVNT_NEWPEER	(10 | PEER_EVENT) /* sys peer*/
+#define	PEVNT_NEWPEER	(10 | PEER_EVENT) /* sys peer */
 #define	PEVNT_CLOCK	(11 | PEER_EVENT) /* clock event */
 #define	PEVNT_AUTH	(12 | PEER_EVENT) /* bad auth */
 #define	PEVNT_POPCORN	(13 | PEER_EVENT) /* popcorn */
+#define	PEVNT_XLEAVE	(14 | PEER_EVENT) /* interleave mode */
+#define	PEVNT_XERR	(15 | PEER_EVENT) /* interleave error */
+#define	PEVNT_TAI	(16 | PEER_EVENT) /* TAI */
 
 /*
  * Clock event codes
@@ -811,7 +753,7 @@ struct mon_data {
 	int	count;			/* total packet count */
 	u_long	firsttime;		/* first time found */
 	u_long	lasttime;		/* last time found */
-	struct sockaddr_storage rmtadr;	/* address of remote host */
+	sockaddr_u rmtadr;		/* address of remote host */
 	struct interface *interface;	/* interface on which this arrived */
 	u_short	rmtport;		/* remote port last came from */
 	u_char	mode;			/* packet mode */
@@ -839,51 +781,59 @@ struct mon_data {
 /*
  * Structure used for restrictlist entries
  */
-struct restrictlist {
-	struct restrictlist *next;	/* link to next entry */
-	u_int32 addr;			/* Ipv4 host address (host byte order) */
-	u_int32 mask;			/* Ipv4 mask for address (host byte order) */
-	u_long count;			/* number of packets matched */
-	u_short flags;			/* accesslist flags */
-	u_short mflags;			/* match flags */
-};
+typedef struct res_addr4_tag {
+	u_int32		addr;		/* IPv4 addr (host order) */
+	u_int32		mask;		/* IPv4 mask (host order) */
+} res_addr4;
 
-struct restrictlist6 {
-	struct restrictlist6 *next;	/* link to next entry */
-	struct in6_addr addr6;		/* Ipv6 host address */
-	struct in6_addr mask6;		/* Ipv6 mask address */
-	u_long count;			/* number of packets matched */
-	u_short flags;			/* accesslist flags */
-	u_short mflags;			/* match flags */
+typedef struct res_addr6_tag {
+	struct in6_addr addr;		/* IPv6 addr (net order) */
+	struct in6_addr mask;		/* IPv6 mask (net order) */
+} res_addr6;
+
+typedef struct restrict_u_tag	restrict_u;
+struct restrict_u_tag {
+	restrict_u *		link;	/* link to next entry */
+	u_int32			count;	/* number of packets matched */
+	u_short			flags;	/* accesslist flags */
+	u_short			mflags;	/* match flags */
+	union {				/* variant starting here */
+		res_addr4 v4;
+		res_addr6 v6;
+	} u;
 };
+#define	V4_SIZEOF_RESTRICT_U	(offsetof(restrict_u, u)	\
+				 + sizeof(res_addr4))
+#define	V6_SIZEOF_RESTRICT_U	(offsetof(restrict_u, u)	\
+				 + sizeof(res_addr6))
 
 
 /*
  * Access flags
  */
-#define	RES_IGNORE		0x001	/* ignore packet */
-#define	RES_DONTSERVE		0x002	/* access denied */
-#define	RES_DONTTRUST		0x004	/* authentication required */
-#define	RES_VERSION		0x008	/* version mismatch */
-#define	RES_NOPEER		0x010	/* new association denied */
-#define RES_LIMITED		0x020	/* packet rate exceeded */
-
+#define	RES_IGNORE		0x0001	/* ignore packet */
+#define	RES_DONTSERVE		0x0002	/* access denied */
+#define	RES_DONTTRUST		0x0004	/* authentication required */
+#define	RES_VERSION		0x0008	/* version mismatch */
+#define	RES_NOPEER		0x0010	/* new association denied */
+#define RES_LIMITED		0x0020	/* packet rate exceeded */
 #define RES_FLAGS		(RES_IGNORE | RES_DONTSERVE |\
 				    RES_DONTTRUST | RES_VERSION |\
 				    RES_NOPEER | RES_LIMITED)
 
-#define	RES_NOQUERY		0x040	/* mode 6/7 packet denied */
-#define	RES_NOMODIFY		0x080	/* mode 6/7 modify denied */
-#define	RES_NOTRAP		0x100	/* mode 6/7 set trap denied */
-#define	RES_LPTRAP		0x200	/* mode 6/7 low priority trap */
+#define	RES_NOQUERY		0x0040	/* mode 6/7 packet denied */
+#define	RES_NOMODIFY		0x0080	/* mode 6/7 modify denied */
+#define	RES_NOTRAP		0x0100	/* mode 6/7 set trap denied */
+#define	RES_LPTRAP		0x0200	/* mode 6/7 low priority trap */
 
-#define RES_KOD			0x400	/* send kiss of death packet */
-#define RES_TIMEOUT		0x800	/* timeout this entry */
+#define RES_KOD			0x0400	/* send kiss of death packet */
+#define	RES_MSSNTP		0x0800	/* enable MS-SNTP authentication */
+#define RES_TIMEOUT		0x1000	/* timeout this entry */
 
 #define	RES_ALLFLAGS		(RES_FLAGS | RES_NOQUERY |\
 				    RES_NOMODIFY | RES_NOTRAP |\
 				    RES_LPTRAP | RES_KOD |\
-				    RES_TIMEOUT)
+				    RES_MSSNTP | RES_TIMEOUT)
 
 /*
  * Match flags
