@@ -44,9 +44,10 @@
 #define BSD_TTYS
 #endif /* SYSV_TTYS STREAM BSD_TTYS */
 
-#define SAMPLE(x)	pp->filter[pp->coderecv++ % MAXSTAGE] = (x); \
-			if (pp->coderecv % MAXSTAGE == pp->codeproc % MAXSTAGE) \
-				pp->codeproc++;
+#define SAMPLE(x)	pp->coderecv = (pp->coderecv + 1) % MAXSTAGE; \
+			pp->filter[pp->coderecv] = (x); \
+			if (pp->coderecv == pp->codeproc) \
+				pp->codeproc = (pp->codeproc + 1) % MAXSTAGE;
 
 /*
  * Macros to determine the clock type and unit numbers from a
@@ -64,6 +65,7 @@ struct clktype {
 	const char *clocktype;	/* long description */
 	const char *abbrev;	/* short description */
 };
+extern struct clktype clktypes[];
 
 /*
  * Configuration flag values
@@ -108,7 +110,7 @@ struct refclockstat {
 	double	fudgetime1;	/* configure fudge time1 */
 	double	fudgetime2;	/* configure fudge time2 */
 	int32	fudgeval1;	/* configure fudge value1 */
-	int32	fudgeval2;	/* configure fudge value2 */
+	u_int32	fudgeval2;	/* configure fudge value2 */
 	u_char	currentstatus;	/* clock status */
 	u_char	lastevent;	/* last exception event */
 	u_char	leap;		/* leap bits */
@@ -121,8 +123,8 @@ struct refclockstat {
  */
 struct refclockio {
 	struct	refclockio *next; /* link to next structure */
-	void	(*clock_recv) P((struct recvbuf *)); /* completion routine */
-	int 	(*io_input)   P((struct recvbuf *)); /* input routine -
+	void	(*clock_recv) (struct recvbuf *); /* completion routine */
+	int 	(*io_input)   (struct recvbuf *); /* input routine -
 				to avoid excessive buffer use
 				due to small bursts
 				of refclock input data */
@@ -162,17 +164,20 @@ struct refclockbug {
  * modules to be installed/loaded in the kernel. If specified, but not
  * installed, the code runs as if unspecified.
  */
-#define LDISC_STD	0x0	/* standard */
-#define LDISC_CLK	0x1	/* tty_clk \n intercept */
-#define LDISC_CLKPPS	0x2	/* tty_clk \377 intercept */
-#define LDISC_ACTS	0x4	/* tty_clk #* intercept */
-#define LDISC_CHU	0x8	/* depredated */
+#define LDISC_STD	0x00	/* standard */
+#define LDISC_CLK	0x01	/* tty_clk \n intercept */
+#define LDISC_CLKPPS	0x02	/* tty_clk \377 intercept */
+#define LDISC_ACTS	0x04	/* tty_clk #* intercept */
+#define LDISC_CHU	0x08	/* depredated */
 #define LDISC_PPS	0x10	/* ppsclock, ppsapi */
 #define LDISC_RAW	0x20	/* raw binary */
+#define LDISC_ECHO	0x40	/* enable echo */
+#define	LDISC_REMOTE	0x80	/* remote mode */
+#define	LDISC_7O1      0x100    /* 7-bit, odd parity for Z3801A */
 
 struct refclockproc {
 	struct	refclockio io;	/* I/O handler structure */
-	caddr_t	unitptr;	/* pointer to unit structure */
+	void *	unitptr;	/* pointer to unit structure */
 	u_char	leap;		/* leap/synchronization code */
 	u_char	currentstatus;	/* clock status */
 	u_char	lastevent;	/* last exception event */
@@ -180,20 +185,19 @@ struct refclockproc {
 	const char *clockdesc;	/* clock description */
 
 	char	a_lastcode[BMAX]; /* last timecode received */
-	u_short	lencode;	/* length of last timecode */
+	int	lencode;	/* length of last timecode */
 
 	int	year;		/* year of eternity */
 	int	day;		/* day of year */
 	int	hour;		/* hour of day */
 	int	minute;		/* minute of hour */
 	int	second;		/* second of minute */
-	int	msec;		/* millisecond of second */
-	long	usec;		/* microsecond of second (alt) */
+	long	nsec;		/* nanosecond of second */
 	u_long	yearstart;	/* beginning of year */
 	int	coderecv;	/* put pointer */
 	int	codeproc;	/* get pointer */
-	l_fp	lastref;	/* timecode timestamp */
-	l_fp	lastrec;	/* local timestamp */
+	l_fp	lastref;	/* reference timestamp */
+	l_fp	lastrec;	/* receive timestamp */
 	double	offset;		/* mean offset */
 	double	disp;		/* sample dispersion */
 	double	jitter;		/* jitter (mean squares) */
@@ -227,14 +231,14 @@ struct refclockproc {
 #define	NOFLAGS	0		/* flag for null flags */
 
 struct refclock {
-	int (*clock_start)	P((int, struct peer *));
-	void (*clock_shutdown)	P((int, struct peer *));
-	void (*clock_poll)	P((int, struct peer *));
-	void (*clock_control)	P((int, struct refclockstat *,
-				    struct refclockstat *, struct peer *));
-	void (*clock_init)	P((void));
-	void (*clock_buginfo)	P((int, struct refclockbug *, struct peer *));
-	u_long clock_flags;
+	int (*clock_start)	(int, struct peer *);
+	void (*clock_shutdown)	(int, struct peer *);
+	void (*clock_poll)	(int, struct peer *);
+	void (*clock_control)	(int, struct refclockstat *,
+				    struct refclockstat *, struct peer *);
+	void (*clock_init)	(void);
+	void (*clock_buginfo)	(int, struct refclockbug *, struct peer *);
+	void (*clock_timer)	(int, struct peer *);
 };
 
 /*
@@ -243,25 +247,28 @@ struct refclock {
 /*
  * auxiliary PPS interface (implemented by refclock_atom())
  */
-extern	int	pps_sample P((l_fp *));
-extern	int	io_addclock_simple P((struct refclockio *));
-extern	int	io_addclock	P((struct refclockio *));
-extern	void	io_closeclock	P((struct refclockio *));
+extern	int	pps_sample (l_fp *);
+extern	int	io_addclock_simple (struct refclockio *);
+extern	int	io_addclock	(struct refclockio *);
+extern	void	io_closeclock	(struct refclockio *);
 
 #ifdef REFCLOCK
-extern	void	refclock_buginfo P((struct sockaddr_in *,
-				    struct refclockbug *));
-extern	void	refclock_control P((struct sockaddr_in *,
+extern	void	refclock_buginfo (sockaddr_u *,
+				    struct refclockbug *);
+extern	void	refclock_control (sockaddr_u *,
 				    struct refclockstat *,
-				    struct refclockstat *));
-extern	int	refclock_open	P((char *, int, int));
-extern	void	refclock_transmit P((struct peer *));
-extern	int	refclock_ioctl	P((int, int));
-extern 	int	refclock_process P((struct refclockproc *));
-extern 	void	refclock_process_offset P((struct refclockproc *, l_fp, l_fp, double));
-extern	void	refclock_report	P((struct peer *, int));
-extern	int	refclock_gtlin	P((struct recvbuf *, char *, int,
-				    l_fp *));
+				    struct refclockstat *);
+extern	int	refclock_open	(char *, u_int, u_int);
+extern	int	refclock_setup	(int, u_int, u_int);
+extern	void	refclock_timer	(struct peer *);
+extern	void	refclock_transmit (struct peer *);
+extern	int	refclock_ioctl	(int, u_int);
+extern 	int	refclock_process (struct refclockproc *);
+extern 	int	refclock_process_f (struct refclockproc *, double);
+extern 	void	refclock_process_offset (struct refclockproc *, l_fp, l_fp, double);
+extern	void	refclock_report	(struct peer *, int);
+extern	int	refclock_gtlin	(struct recvbuf *, char *, int, l_fp *);
+extern	int	refclock_gtraw  (struct recvbuf *, char *, int, l_fp *);
 #endif /* REFCLOCK */
 
 #endif /* NTP_REFCLOCK_H */
