@@ -65,47 +65,46 @@
 /*
  * Compatability shims with the rfc2553 API to simplify ntp.
  */
-#ifndef _NTP_RFC2553_H_
-#define _NTP_RFC2553_H_
+#ifndef NTP_RFC2553_H
+#define NTP_RFC2553_H
 
 /*
  * Ensure that we include the configuration file before we check
  * for IPV6
  */
 #include <config.h>
-
 #include <netdb.h>
+#include <isc/net.h>
 
 #include "ntp_types.h"
 
-/*
- * Don't include any additional IPv6 definitions
- * We are defining our own here.
- */
-#define ISC_IPV6_H 1
 
  /*
  * If various macros are not defined we need to define them
  */
 
 #ifndef AF_INET6
-#define AF_INET6	AF_MAX
-#define PF_INET6	AF_INET6
+# define AF_INET6	AF_MAX
+# define PF_INET6	AF_INET6
 #endif
 
 #if !defined(_SS_MAXSIZE) && !defined(_SS_ALIGNSIZE)
 
-#define	_SS_MAXSIZE	128
-#define	_SS_ALIGNSIZE	(sizeof(ntp_uint64_t))
-#ifdef HAVE_SA_LEN_IN_STRUCT_SOCKADDR
-#define	_SS_PAD1SIZE	(_SS_ALIGNSIZE - sizeof(u_char) - sizeof(ntp_u_int8_t))
-#define	_SS_PAD2SIZE	(_SS_MAXSIZE - sizeof(u_char) - sizeof(ntp_u_int8_t) - \
+# define	_SS_MAXSIZE	128
+# define	_SS_ALIGNSIZE	(sizeof(ntp_uint64_t))
+# ifdef ISC_PLATFORM_HAVESALEN
+#  define	_SS_PAD1SIZE	(_SS_ALIGNSIZE - sizeof(u_char) - sizeof(ntp_u_int8_t))
+#  define	_SS_PAD2SIZE	(_SS_MAXSIZE - sizeof(u_char) - sizeof(ntp_u_int8_t) - \
 				_SS_PAD1SIZE - _SS_ALIGNSIZE)
-#else
-#define	_SS_PAD1SIZE	(_SS_ALIGNSIZE - sizeof(short))
-#define	_SS_PAD2SIZE	(_SS_MAXSIZE - sizeof(short) - \
+# else
+#  define	_SS_PAD1SIZE	(_SS_ALIGNSIZE - sizeof(short))
+#  define	_SS_PAD2SIZE	(_SS_MAXSIZE - sizeof(short) - \
 				_SS_PAD1SIZE - _SS_ALIGNSIZE)
-#endif /* HAVE_SA_LEN_IN_STRUCT_SOCKADDR */
+# endif /* ISC_PLATFORM_HAVESALEN */
+#endif
+
+#ifndef INET6_ADDRSTRLEN
+# define	INET6_ADDRSTRLEN	46	/* max len of IPv6 addr in ascii */
 #endif
 
 /*
@@ -115,7 +114,7 @@
 
 #ifndef HAVE_STRUCT_SOCKADDR_STORAGE
 struct sockaddr_storage {
-#ifdef HAVE_SA_LEN_IN_STRUCT_SOCKADDR
+#ifdef ISC_PLATFORM_HAVESALEN
 	ntp_u_int8_t	ss_len;		/* address length */
 	ntp_u_int8_t	ss_family;	/* address family */
 #else
@@ -135,7 +134,7 @@ struct sockaddr_storage {
 /*
  * Flag values for getaddrinfo()
  */
-#ifndef AI_NUMERICHOST
+#ifndef AI_PASSIVE
 #define	AI_PASSIVE	0x00000001 /* get address to use bind() */
 #define	AI_CANONNAME	0x00000002 /* fill ai_canonname */
 #define	AI_NUMERICHOST	0x00000004 /* prevent name resolution */
@@ -144,90 +143,16 @@ struct sockaddr_storage {
     (AI_PASSIVE | AI_CANONNAME | AI_NUMERICHOST | AI_ADDRCONFIG)
 
 #define	AI_ADDRCONFIG	0x00000400 /* only if any address is assigned */
+#endif	/* !AI_PASSIVE */
+
+#ifndef AI_NUMERICHOST		/* such as AIX 4.3 */
+#define AI_NUMERICHOST	0
 #endif
 
 #ifndef ISC_PLATFORM_HAVEIPV6
-/*
- * Definition of some useful macros to handle IP6 addresses
- */
-#ifdef ISC_PLATFORM_NEEDIN6ADDRANY
+
 #ifdef SYS_WINNT
-#define IN6ADDR_ANY_INIT 	{{ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 }}
-#else
-#define IN6ADDR_ANY_INIT \
-	{{{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
-	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }}}
-#endif
-#endif
-
-
-/*
- * IPv6 address
- */
-#ifdef SYS_WINNT
-#define in6_addr in_addr6
-#else
-
-struct in6_addr {
-	union {
-		ntp_u_int8_t   __u6_addr8[16];
-		ntp_u_int16_t  __u6_addr16[8];
-		ntp_u_int32_t  __u6_addr32[4];
-	} __u6_addr;			/* 128-bit IP6 address */
-};
-
-#define s6_addr   __u6_addr.__u6_addr8
-#endif
-
-#if defined(ISC_PLATFORM_HAVEIPV6) && defined(ISC_PLATFORM_NEEDIN6ADDRANY)
-extern const struct in6_addr in6addr_any;
-#endif
-
-#define SIN6_LEN
-#ifndef HAVE_SOCKADDR_IN6
-struct sockaddr_in6 {
-#ifdef HAVE_SA_LEN_IN_STRUCT_SOCKADDR
-	ntp_u_int8_t	sin6_len;	/* length of this struct(sa_family_t)*/
-	ntp_u_int8_t	sin6_family;	/* AF_INET6 (sa_family_t) */
-#else
-	short		sin6_family;	/* AF_INET6 (sa_family_t) */
-#endif
-	ntp_u_int16_t	sin6_port;	/* Transport layer port # (in_port_t)*/
-	ntp_u_int32_t	sin6_flowinfo;	/* IP6 flow information */
-	struct in6_addr	sin6_addr;	/* IP6 address */
-	ntp_u_int32_t	sin6_scope_id;	/* scope zone index */
-};
-#endif
-
-/*
- * Unspecified
- */
-#ifndef IN6_IS_ADDR_UNSPECIFIED
-#define IN6_IS_ADDR_UNSPECIFIED(a)	\
-	((*(const ntp_u_int32_t *)(const void *)(&(a)->s6_addr[0]) == 0) &&	\
-	 (*(const ntp_u_int32_t *)(const void *)(&(a)->s6_addr[4]) == 0) &&	\
-	 (*(const ntp_u_int32_t *)(const void *)(&(a)->s6_addr[8]) == 0) &&	\
-	 (*(const ntp_u_int32_t *)(const void *)(&(a)->s6_addr[12]) == 0))
-#endif
-/*
- * Multicast
- */
-#ifndef IN6_IS_ADDR_MULTICAST
-#define IN6_IS_ADDR_MULTICAST(a)	((a)->s6_addr[0] == 0xff)
-#endif
-/*
- * Unicast link / site local.
- */
-#ifndef IN6_IS_ADDR_LINKLOCAL
-#define IN6_IS_ADDR_LINKLOCAL(a)	(\
-(*((u_long *)((a)->s6_addr)    ) == 0xfe) && \
-((*((u_long *)((a)->s6_addr) + 1) & 0xc0) == 0x80))
-#endif
-
-#ifndef IN6_IS_ADDR_SITELOCAL
-#define IN6_IS_ADDR_SITELOCAL(a)	(\
-(*((u_long *)((a)->s6_addr)    ) == 0xfe) && \
-((*((u_long *)((a)->s6_addr) + 1) & 0xc0) == 0xc0))
+# define in6_addr in_addr6
 #endif
 
 struct addrinfo {
@@ -260,12 +185,12 @@ struct addrinfo {
 #define	EAI_MAX		14
 
 
-int	getaddrinfo P((const char *, const char *,
-			 const struct addrinfo *, struct addrinfo **));
-int	getnameinfo P((const struct sockaddr *, u_int, char *,
-			 size_t, char *, size_t, int));
-void	freeaddrinfo P((struct addrinfo *));
-char	*gai_strerror P((int));
+int	getaddrinfo (const char *, const char *,
+			 const struct addrinfo *, struct addrinfo **);
+int	getnameinfo (const struct sockaddr *, u_int, char *,
+			 size_t, char *, size_t, int);
+void	freeaddrinfo (struct addrinfo *);
+char	*gai_strerror (int);
 
 /*
  * Constants for getnameinfo()
@@ -287,6 +212,17 @@ char	*gai_strerror P((int));
 #define NI_WITHSCOPEID	0x00000020
 #endif
 
-#endif /* ISC_PLATFORM_HAVEIPV6 */
+#endif /* !ISC_PLATFORM_HAVEIPV6 */
 
-#endif /* !_NTP_RFC2553_H_ */
+/* 
+ * Set up some macros to look for IPv6 and IPv6 multicast
+ */
+
+#if defined(ISC_PLATFORM_HAVEIPV6) && defined(WANT_IPV6)
+# define INCLUDE_IPV6_SUPPORT
+# if defined(IPV6_JOIN_GROUP) && defined(IPV6_LEAVE_GROUP)
+#  define INCLUDE_IPV6_MULTICAST_SUPPORT
+# endif	/* IPV6 Multicast Support */
+#endif  /* IPv6 Support */
+
+#endif /* !NTP_RFC2553_H */
