@@ -40,14 +40,14 @@ struct clktype clktypes[] = {
 	  "ATOM_LEITCH" },
 	{ REFCLK_MSF_EES,	"EES M201 MSF Receiver (14)",
 	  "MSF_EES" },
-	{ REFCLK_GPSTM_TRUE,	"TrueTime (old GPSTM driver) Receiver (15)",
-	  "GPSTM_TRUE" },
+	{ REFCLK_NONE,		"not used (15)",
+	  "NOT_USED" },
 	{ REFCLK_IRIG_BANCOMM,	"Bancomm GPS/IRIG Receiver (16)",
 	  "GPS_BANC" },
 	{ REFCLK_GPS_DATUM,	"Datum Precision Time System (17)",
 	  "GPS_DATUM" },
-	{ REFCLK_NIST_ACTS,	"NIST Automated Computer Time Service (18)",
-	  "ACTS_NIST" },
+	{ REFCLK_ACTS,		"Automated Computer Time Service (18)",
+	  "ACTS_MODEM" },
 	{ REFCLK_WWV_HEATH,	"Heath WWV/WWVH Receiver (19)",
 	  "WWV_HEATH" },
 	{ REFCLK_GPS_NMEA,	"Generic NMEA GPS Receiver (20)",
@@ -56,20 +56,20 @@ struct clktype clktypes[] = {
 	  "GPS_VME" },
 	{ REFCLK_ATOM_PPS,	"PPS Clock Discipline (22)",
 	  "PPS" },
-	{ REFCLK_PTB_ACTS,	"PTB Automated Computer Time Service (23)",
-	  "ACTS_PTB" },
-	{ REFCLK_USNO,		"Naval Observatory dialup (24)",
-	  "ACTS_USNO" },
-	{ REFCLK_TRUETIME,	"TrueTime (generic) Receivers (25)",
-	  "TRUETIME" },
+	{ REFCLK_NONE,		"not used (23)",
+	  "NOT_USED" },
+	{ REFCLK_NONE,		"not used (24)",
+	  "NOT_USED" },
+	{ REFCLK_NONE,		"not used (25)",
+	  "NOT_USED" },
 	{ REFCLK_GPS_HP,	"HP 58503A GPS Time & Frequency Receiver (26)",
 	  "GPS_HP" },
 	{ REFCLK_ARCRON_MSF,	"ARCRON MSF (and DCF77) Receiver (27)",
 	  "MSF_ARCRON" },
 	{ REFCLK_SHM,		"Clock attached thru shared Memory (28)",
 	  "SHM" },
-        { REFCLK_PALISADE,      "Trimble Navigation Palisade GPS (29)",
-          "GPS_PALISADE" },
+	{ REFCLK_PALISADE,	"Trimble Navigation Palisade GPS (29)",
+	  "GPS_PALISADE" },
 	{ REFCLK_ONCORE,	"Motorola UT Oncore GPS (30)",
 	  "GPS_ONCORE" },
 	{ REFCLK_GPS_JUPITER,	"Rockwell Jupiter GPS (31)",
@@ -78,20 +78,28 @@ struct clktype clktypes[] = {
 	  "CHRONOLOG" },
 	{ REFCLK_DUMBCLOCK,	"Dumb generic hh:mm:ss local clock (33)",
 	  "DUMBCLOCK" },
-	{ REFCLK_ULINK,         "Ultralink M320 WWVB receiver (34)",
+	{ REFCLK_ULINK,		"Ultralink M320 WWVB receiver (34)",
 	  "ULINK_M320"},
 	{ REFCLK_PCF,		"Conrad parallel port radio clock (35)",
 	  "PCF"},
 	{ REFCLK_WWV_AUDIO,	"WWV/H Audio Demodulator/Decoder (36)",
 	  "WWV_AUDIO"},
-	{ REFCLK_FG,	"Forum Graphic GPS Dating Station (37)",
+	{ REFCLK_FG,		"Forum Graphic GPS Dating Station (37)",
 	  "GPS_FG"},
 	{ REFCLK_HOPF_SERIAL,	"hopf Elektronic serial line receiver (38)",
 	  "HOPF_S"},
 	{ REFCLK_HOPF_PCI,	"hopf Elektronic PCI receiver (39)",
 	  "HOPF_P"},
-	{ REFCLK_JJY,	"JJY receiver (40)",
-	  "JJY" },
+	{ REFCLK_JJY,		"JJY receiver (40)",
+	  "JJY"},
+	{ REFCLK_TT560,		"TrueTime 560 IRIG-B decoder (41)",
+	  "TT_IRIG"},
+	{ REFCLK_ZYFER,		"Zyfer GPStarplus receiver (42)",
+	  "GPS_ZYFER" },
+	{ REFCLK_RIPENCC,	"RIPE NCC Trimble driver (43)",
+	  "GPS_RIPENCC" },
+	{ REFCLK_NEOCLOCK4X,	"NeoClock4X DCF77 / TDF receiver (44)",
+	  "NEOCLK4X"},
 	{ -1,			"", "" }
 };
 
@@ -104,7 +112,7 @@ clockname(
   
 	for (clk = clktypes; clk->code != -1; clk++) {
 		if (num == clk->code)
-		    return (clk->abbrev);
+			return (clk->abbrev);
 	}
 	return (NULL);
 }
