@@ -11,6 +11,7 @@
 #include "ntp_tty.h"
 #include "ntp_refclock.h"
 #include "ntp_stdlib.h"
+#include "ntp_assert.h"
 
 #include <stdio.h>
 
@@ -30,68 +31,62 @@
 # endif
 #endif /* TTYCLK */
 
-#ifdef HAVE_PPSCLOCK_H
-#include <sys/ppsclock.h>
-#endif /* HAVE_PPSCLOCK_H */
-
 #ifdef KERNEL_PLL
 #include "ntp_syscall.h"
 #endif /* KERNEL_PLL */
 
+#ifdef HAVE_PPSAPI
+#include "ppsapi_timepps.h"
+#include "refclock_atom.h"
+#endif /* HAVE_PPSAPI */
+
 /*
  * Reference clock support is provided here by maintaining the fiction
- * that the clock is actually a peer. As no packets are exchanged with a
- * reference clock, however, we replace the transmit, receive and packet
- * procedures with separate code to simulate them. Routines
+ * that the clock is actually a peer.  As no packets are exchanged with
+ * a reference clock, however, we replace the transmit, receive and
+ * packet procedures with separate code to simulate them.  Routines
  * refclock_transmit() and refclock_receive() maintain the peer
  * variables in a state analogous to an actual peer and pass reference
- * clock data on through the filters. Routines refclock_peer() and
+ * clock data on through the filters.  Routines refclock_peer() and
  * refclock_unpeer() are called to initialize and terminate reference
- * clock associations. A set of utility routines is included to open
+ * clock associations.  A set of utility routines is included to open
  * serial devices, process sample data, edit input lines to extract
- * embedded timestamps and to peform various debugging functions.
+ * embedded timestamps and to perform various debugging functions.
  *
  * The main interface used by these routines is the refclockproc
- * structure, which contains for most drivers the decimal equivalants of
- * the year, day, month, hour, second and millisecond/microsecond
- * decoded from the ASCII timecode. Additional information includes the
- * receive timestamp, exception report, statistics tallies, etc. In
- * addition, there may be a driver-specific unit structure used for
+ * structure, which contains for most drivers the decimal equivalants
+ * of the year, day, month, hour, second and millisecond/microsecond
+ * decoded from the ASCII timecode.  Additional information includes
+ * the receive timestamp, exception report, statistics tallies, etc. 
+ * In addition, there may be a driver-specific unit structure used for
  * local control of the device.
  *
  * The support routines are passed a pointer to the peer structure,
- * which is used for all peer-specific processing and contains a pointer
- * to the refclockproc structure, which in turn containes a pointer to
- * the unit structure, if used. The peer structure is identified by an
- * interface address in the dotted quad form 127.127.t.u (for now only IPv4
- * addresses are used, so we need to be sure the address is it), where t is
- * the clock type and u the unit. Some legacy drivers derive the
- * refclockproc structure pointer from the table typeunit[type][unit].
- * This interface is strongly discouraged and may be abandoned in
- * future.
+ * which is used for all peer-specific processing and contains a
+ * pointer to the refclockproc structure, which in turn contains a
+ * pointer to the unit structure, if used.  The peer structure is 
+ * identified by an interface address in the dotted quad form 
+ * 127.127.t.u, where t is the clock type and u the unit.
  */
-#define MAXUNIT 	4	/* max units */
 #define FUDGEFAC	.1	/* fudge correction factor */
+#define LF		0x0a	/* ASCII LF */
 
-int fdpps;			/* pps file descriptor */
-int cal_enable;			/* enable refclock calibrate */
+#ifdef PPS
+int	fdpps;			/* ppsclock legacy */
+#endif /* PPS */
 
-/*
- * Type/unit peer index. Used to find the peer structure for control and
- * debugging. When all clock drivers have been converted to new style,
- * this dissapears.
- */
-static struct peer *typeunit[REFCLK_MAX + 1][MAXUNIT];
+int	cal_enable;		/* enable refclock calibrate */
 
 /*
  * Forward declarations
  */
 #ifdef QSORT_USES_VOID_P
-static int refclock_cmpl_fp P((const void *, const void *));
+static int refclock_cmpl_fp (const void *, const void *);
 #else
-static int refclock_cmpl_fp P((const double *, const double *));
+static int refclock_cmpl_fp (const double *, const double *);
 #endif /* QSORT_USES_VOID_P */
-static int refclock_sample P((struct refclockproc *));
+static int refclock_sample (struct refclockproc *);
+
 
 /*
  * refclock_report - note the occurance of an event
@@ -111,34 +106,35 @@ refclock_report(
 	pp = peer->procptr;
 	if (pp == NULL)
 		return;
-	if (code == CEVNT_BADREPLY)
-		pp->badformat++;
-	if (code == CEVNT_BADTIME)
-		pp->baddata++;
-	if (code == CEVNT_TIMEOUT)
+
+	switch (code) {
+
+	case CEVNT_TIMEOUT:
 		pp->noreply++;
+		break;
+
+	case CEVNT_BADREPLY:
+		pp->badformat++;
+		break;
+
+	case CEVNT_FAULT:
+		break;
+
+	case CEVNT_BADDATE:
+	case CEVNT_BADTIME:
+		pp->baddata++;
+		break;
+
+	default:
+		/* ignore others */
+		break;
+	}
+	if (pp->lastevent < 15)
+		pp->lastevent++;
 	if (pp->currentstatus != code) {
 		pp->currentstatus = (u_char)code;
-		pp->lastevent = (u_char)code;
-		if (code == CEVNT_FAULT)
-			msyslog(LOG_ERR,
-				"clock %s event '%s' (0x%02x)",
-				refnumtoa(&peer->srcadr),
-				ceventstr(code), code);
-		else {
-			NLOG(NLOG_CLOCKEVENT)
-				msyslog(LOG_INFO,
-				"clock %s event '%s' (0x%02x)",
-				refnumtoa(&peer->srcadr),
-				ceventstr(code), code);
-		}
+		report_event(PEVNT_CLOCK, peer, ceventstr(code));
 	}
-#ifdef DEBUG
-	if (debug)
-		printf("clock %s event '%s' (0x%02x)\n",
-			refnumtoa(&peer->srcadr),
-			ceventstr(code), code);
-#endif
 }
 
 
@@ -152,14 +148,11 @@ refclock_report(
 void
 init_refclock(void)
 {
-	int i, j;
+	int i;
 
-	for (i = 0; i < (int)num_refclock_conf; i++) {
+	for (i = 0; i < (int)num_refclock_conf; i++)
 		if (refclock_conf[i]->clock_init != noentry)
 			(refclock_conf[i]->clock_init)();
-		for (j = 0; j < MAXUNIT; j++)
-			typeunit[i][j] = 0;
-	}
 }
 
 
@@ -188,12 +181,6 @@ refclock_newpeer(
 	 * Check for valid clock address. If already running, shut it
 	 * down first.
 	 */
-	if (peer->srcadr.ss_family != AF_INET) {
-		msyslog(LOG_ERR,
-		       "refclock_newpeer: clock address %s invalid, address family not implemented for refclock",
-                        stoa(&peer->srcadr));
-                return (0);
-        }
 	if (!ISREFCLOCKADR(&peer->srcadr)) {
 		msyslog(LOG_ERR,
 			"refclock_newpeer: clock address %s invalid",
@@ -202,7 +189,7 @@ refclock_newpeer(
 	}
 	clktype = (u_char)REFCLOCKTYPE(&peer->srcadr);
 	unit = REFCLOCKUNIT(&peer->srcadr);
-	if (clktype >= num_refclock_conf || unit >= MAXUNIT ||
+	if (clktype >= num_refclock_conf ||
 		refclock_conf[clktype]->clock_start == noentry) {
 		msyslog(LOG_ERR,
 			"refclock_newpeer: clock type %d invalid\n",
@@ -213,11 +200,8 @@ refclock_newpeer(
 	/*
 	 * Allocate and initialize interface structure
 	 */
-	pp = (struct refclockproc *)emalloc(sizeof(struct refclockproc));
-	if (pp == NULL)
-		return (0);
-	memset((char *)pp, 0, sizeof(struct refclockproc));
-	typeunit[clktype][unit] = peer;
+	pp = emalloc(sizeof(*pp));
+	memset(pp, 0, sizeof(*pp));
 	peer->procptr = pp;
 
 	/*
@@ -226,8 +210,8 @@ refclock_newpeer(
 	peer->refclktype = clktype;
 	peer->refclkunit = (u_char)unit;
 	peer->flags |= FLAG_REFCLOCK;
+	peer->leap = LEAP_NOTINSYNC;
 	peer->stratum = STRATUM_REFCLOCK;
-	peer->hpoll = peer->minpoll;
 	peer->ppoll = peer->maxpoll;
 	pp->type = clktype;
 	pp->timestarted = current_time;
@@ -273,17 +257,36 @@ refclock_unpeer(
 	 * Wiggle the driver to release its resources, then give back
 	 * the interface structure.
 	 */
-	if (!peer->procptr)
+	if (NULL == peer->procptr)
 		return;
+
 	clktype = peer->refclktype;
 	unit = peer->refclkunit;
 	if (refclock_conf[clktype]->clock_shutdown != noentry)
 		(refclock_conf[clktype]->clock_shutdown)(unit, peer);
 	free(peer->procptr);
-	peer->procptr = 0;
+	peer->procptr = NULL;
 }
 
 
+/*
+ * refclock_timer - called once per second for housekeeping.
+ */
+void
+refclock_timer(
+	struct peer *peer	/* peer structure pointer */
+	)
+{
+	u_char clktype;
+	int unit;
+
+	clktype = peer->refclktype;
+	unit = peer->refclkunit;
+	if (refclock_conf[clktype]->clock_timer != noentry)
+		(refclock_conf[clktype]->clock_timer)(unit, peer);
+}
+	
+
 /*
  * refclock_transmit - simulate the transmit procedure
  *
@@ -299,18 +302,17 @@ refclock_transmit(
 {
 	u_char clktype;
 	int unit;
-	u_long next;
 
 	clktype = peer->refclktype;
 	unit = peer->refclkunit;
 	peer->sent++;
+	get_systime(&peer->xmt);
 
 	/*
 	 * This is a ripoff of the peer transmit routine, but
 	 * specialized for reference clocks. We do a little less
 	 * protocol here and call the driver-specific transmit routine.
 	 */
-	next = peer->outdate;
 	if (peer->burst == 0) {
 		u_char oreach;
 #ifdef DEBUG
@@ -323,30 +325,25 @@ refclock_transmit(
 		 * Update reachability and poll variables like the
 		 * network code.
 		 */
-		oreach = peer->reach;
+		oreach = peer->reach & 0xfe;
 		peer->reach <<= 1;
+		if (!(peer->reach & 0x0f))
+			clock_filter(peer, 0., 0., MAXDISPERSE);
+		peer->outdate = current_time;
 		if (!peer->reach) {
 			if (oreach) {
-				report_event(EVNT_UNREACH, peer);
+				report_event(PEVNT_UNREACH, peer, NULL);
 				peer->timereachable = current_time;
-				peer_clear(peer, "NONE");
 			}
 		} else {
-			if (!(oreach & 0x03)) {
-				clock_filter(peer, 0., 0., MAXDISPERSE);
-				clock_select();
-			}
 			if (peer->flags & FLAG_BURST)
 				peer->burst = NSTAGE;
 		}
-		next = current_time;
+	} else {
+		peer->burst--;
 	}
-	get_systime(&peer->xmt);
 	if (refclock_conf[clktype]->clock_poll != noentry)
 		(refclock_conf[clktype]->clock_poll)(unit, peer);
-	peer->outdate = next;
-	if (peer->burst > 0)
-		peer->burst--;
 	poll_update(peer, peer->hpoll);
 }
 
@@ -366,10 +363,13 @@ refclock_cmpl_fp(
 
 	if (*dp1 < *dp2)
 		return (-1);
+
 	if (*dp1 > *dp2)
 		return (1);
+
 	return (0);
 }
+
 #else
 static int
 refclock_cmpl_fp(
@@ -379,8 +379,10 @@ refclock_cmpl_fp(
 {
 	if (*dp1 < *dp2)
 		return (-1);
+
 	if (*dp1 > *dp2)
 		return (1);
+
 	return (0);
 }
 #endif /* QSORT_USES_VOID_P */
@@ -411,18 +413,27 @@ refclock_process_offset(
 	SAMPLE(doffset + fudge);
 }
 
+
 /*
  * refclock_process - process a sample from the clock
+ * refclock_process_f - refclock_process with other than time1 fudge
  *
  * This routine converts the timecode in the form days, hours, minutes,
  * seconds and milliseconds/microseconds to internal timestamp format,
  * then constructs a new entry in the median filter circular buffer.
  * Return success (1) if the data are correct and consistent with the
  * converntional calendar.
-*/
+ *
+ * Important for PPS users: Normally, the pp->lastrec is set to the
+ * system time when the on-time character is received and the pp->year,
+ * ..., pp->second decoded and the seconds fraction pp->nsec in
+ * nanoseconds). When a PPS offset is available, pp->nsec is forced to
+ * zero and the fraction for pp->lastrec is set to the PPS offset.
+ */
 int
-refclock_process(
-	struct refclockproc *pp		/* refclock structure pointer */
+refclock_process_f(
+	struct refclockproc *pp,	/* refclock structure pointer */
+	double fudge
 	)
 {
 	l_fp offset, ltemp;
@@ -438,20 +449,30 @@ refclock_process(
 	if (!clocktime(pp->day, pp->hour, pp->minute, pp->second, GMT,
 		pp->lastrec.l_ui, &pp->yearstart, &offset.l_ui))
 		return (0);
+
 	offset.l_uf = 0;
 	DTOLFP(pp->nsec / 1e9, &ltemp);
 	L_ADD(&offset, &ltemp);
-	refclock_process_offset(pp, offset, pp->lastrec,
-	    pp->fudgetime1);
+	refclock_process_offset(pp, offset, pp->lastrec, fudge);
 	return (1);
 }
 
+
+int
+refclock_process(
+	struct refclockproc *pp		/* refclock structure pointer */
+)
+{
+	return refclock_process_f(pp, pp->fudgetime1);
+}
+
+
 /*
  * refclock_sample - process a pile of samples from the clock
  *
  * This routine implements a recursive median filter to suppress spikes
  * in the data, as well as determine a performance statistic. It
- * calculates the mean offset and jitter (squares). A time adjustment
+ * calculates the mean offset and RMS jitter. A time adjustment
  * fudgetime1 can be added to the final offset to compensate for various
  * systematic errors. The routine returns the number of samples
  * processed, which could be zero.
@@ -461,9 +482,9 @@ refclock_sample(
 	struct refclockproc *pp		/* refclock structure pointer */
 	)
 {
-	int i, j, k, m, n;
-	double offset;
-	double off[MAXSTAGE];
+	int	i, j, k, m, n;
+	double	off[MAXSTAGE];
+	double	offset;
 
 	/*
 	 * Copy the raw offsets and sort into ascending order. Don't do
@@ -477,15 +498,22 @@ refclock_sample(
 	}
 	if (n == 0)
 		return (0);
+
 	if (n > 1)
-		qsort((char *)off, (size_t)n, sizeof(double), refclock_cmpl_fp);
+		qsort(
+#ifdef QSORT_USES_VOID_P
+		    (void *)
+#else
+		    (char *)
+#endif
+		    off, (size_t)n, sizeof(double), refclock_cmpl_fp);
 
 	/*
 	 * Reject the furthest from the median of the samples until
 	 * approximately 60 percent of the samples remain.
 	 */
 	i = 0; j = n;
-	m = n - (n * 2) / NSTAGE;
+	m = n - (n * 4) / 10;
 	while ((j - i) > m) {
 		offset = off[(j + i) / 2];
 		if (off[j - 1] - offset < offset - off[i])
@@ -497,19 +525,20 @@ refclock_sample(
 	/*
 	 * Determine the offset and jitter.
 	 */
-	offset = 0;
-	for (k = i; k < j; k++)
-		offset += off[k];
-	pp->offset = offset / m;
-	if (m > 1)
-		pp->jitter = SQUARE(off[i] - off[j - 1]);
-	else
-		pp->jitter = 0;
+	pp->offset = 0;
+	pp->jitter = 0;
+	for (k = i; k < j; k++) {
+		pp->offset += off[k];
+		if (k > i)
+			pp->jitter += SQUARE(off[k] - off[k - 1]);
+	}
+	pp->offset /= m;
+	pp->jitter = max(SQRT(pp->jitter / m), LOGTOD(sys_precision));
 #ifdef DEBUG
 	if (debug)
 		printf(
 		    "refclock_sample: n %d offset %.6f disp %.6f jitter %.6f\n",
-		    n, pp->offset, pp->disp, SQRT(pp->jitter));
+		    n, pp->offset, pp->disp, pp->jitter);
 #endif
 	return (n);
 }
@@ -543,61 +572,101 @@ refclock_receive(
 	 * filter.
 	 */
 	pp = peer->procptr;
+	peer->leap = pp->leap;
+	if (peer->leap == LEAP_NOTINSYNC)
+		return;
+
 	peer->received++;
 	peer->timereceived = current_time;
-	peer->leap = pp->leap;
-	if (peer->leap == LEAP_NOTINSYNC) {
-		refclock_report(peer, CEVNT_FAULT);
-		return;
+	if (!peer->reach) {
+		report_event(PEVNT_REACH, peer, NULL);
+		peer->timereachable = current_time;
 	}
-	if (!peer->reach)
-		report_event(EVNT_REACH, peer);
 	peer->reach |= 1;
 	peer->reftime = pp->lastref;
-	peer->org = pp->lastrec;
-	peer->rootdispersion = pp->disp;
-	get_systime(&peer->rec);
+	peer->aorg = pp->lastrec;
+	peer->rootdisp = pp->disp;
+	get_systime(&peer->dst);
 	if (!refclock_sample(pp))
 		return;
+
 	clock_filter(peer, pp->offset, 0., pp->jitter);
-	clock_select();
-	record_peer_stats(&peer->srcadr, ctlpeerstatus(peer),
-	    peer->offset, peer->delay, clock_phi * (current_time -
-	    peer->epoch), SQRT(peer->jitter));
-	if (cal_enable && last_offset < MINDISPERSE) {
-#ifdef KERNEL_PLL
-		if (peer != sys_peer || pll_status & STA_PPSTIME)
-#else
-		if (peer != sys_peer)
-#endif /* KERNEL_PLL */
+	if (cal_enable && fabs(last_offset) < sys_mindisp && sys_peer !=
+	    NULL) {
+		if (sys_peer->refclktype == REFCLK_ATOM_PPS &&
+		    peer->refclktype != REFCLK_ATOM_PPS)
 			pp->fudgetime1 -= pp->offset * FUDGEFAC;
-		else
-			pp->fudgetime1 -= pp->fudgetime1 * FUDGEFAC;
 	}
 }
 
+
 /*
  * refclock_gtlin - groom next input line and extract timestamp
  *
  * This routine processes the timecode received from the clock and
- * removes the parity bit and control characters. If a timestamp is
- * present in the timecode, as produced by the tty_clk STREAMS module,
- * it returns that as the timestamp; otherwise, it returns the buffer
- *  timestamp. The routine return code is the number of characters in
- * the line.
+ * strips the parity bit and control characters. It returns the number
+ * of characters in the line followed by a NULL character ('\0'), which
+ * is not included in the count. In case of an empty line, the previous
+ * line is preserved.
  */
 int
 refclock_gtlin(
 	struct recvbuf *rbufp,	/* receive buffer pointer */
-	char *lineptr,		/* current line pointer */
-	int bmax,		/* remaining characters in line */
-	l_fp *tsptr		/* pointer to timestamp returned */
+	char	*lineptr,	/* current line pointer */
+	int	bmax,		/* remaining characters in line */
+	l_fp	*tsptr		/* pointer to timestamp returned */
 	)
 {
-	char *dpt, *dpend, *dp;
-	int i;
-	l_fp trtmp, tstmp;
-	char c;
+	char	s[BMAX];
+	char	*dpt, *dpend, *dp;
+
+	dpt = s;
+	dpend = s + refclock_gtraw(rbufp, s, BMAX - 1, tsptr);
+	if (dpend - dpt > bmax - 1)
+		dpend = dpt + bmax - 1;
+	for (dp = lineptr; dpt < dpend; dpt++) {
+		char	c;
+
+		c = *dpt & 0x7f;
+		if (c >= 0x20 && c < 0x7f)
+			*dp++ = c;
+	}
+	if (dp == lineptr)
+		return (0);
+
+	*dp = '\0';
+	return (dp - lineptr);
+}
+
+
+/*
+ * refclock_gtraw - get next line/chunk of data
+ *
+ * This routine returns the raw data received from the clock in both
+ * canonical or raw modes. The terminal interface routines map CR to LF.
+ * In canonical mode this results in two lines, one containing data
+ * followed by LF and another containing only LF. In raw mode the
+ * interface routines can deliver arbitraty chunks of data from one
+ * character to a maximum specified by the calling routine. In either
+ * mode the routine returns the number of characters in the line
+ * followed by a NULL character ('\0'), which is not included in the
+ * count.
+ *
+ * If a timestamp is present in the timecode, as produced by the tty_clk
+ * STREAMS module, it returns that as the timestamp; otherwise, it
+ * returns the buffer timestamp.
+ */
+int
+refclock_gtraw(
+	struct recvbuf *rbufp,	/* receive buffer pointer */
+	char	*lineptr,	/* current line pointer */
+	int	bmax,		/* remaining characters in line */
+	l_fp	*tsptr		/* pointer to timestamp returned */
+	)
+{
+	char	*dpt, *dpend, *dp;
+	l_fp	trtmp, tstmp;
+	int	i;
 
 	/*
 	 * Check for the presence of a timestamp left by the tty_clock
@@ -609,7 +678,6 @@ refclock_gtlin(
 	dpt = (char *)rbufp->recv_buffer;
 	dpend = dpt + rbufp->recv_length;
 	trtmp = rbufp->recv_time;
-
 	if (dpend >= dpt + 8) {
 		if (buftvtots(dpend - 8, &tstmp)) {
 			L_SUB(&trtmp, &tstmp);
@@ -618,10 +686,12 @@ refclock_gtlin(
 				if (debug > 1) {
 					printf(
 					    "refclock_gtlin: fd %d ldisc %s",
-					    rbufp->fd, lfptoa(&trtmp, 6));
+					    rbufp->fd, lfptoa(&trtmp,
+					    6));
 					get_systime(&trtmp);
 					L_SUB(&trtmp, &tstmp);
-					printf(" sigio %s\n", lfptoa(&trtmp, 6));
+					printf(" sigio %s\n",
+					    lfptoa(&trtmp, 6));
 				}
 #endif
 				dpend -= 8;
@@ -632,34 +702,25 @@ refclock_gtlin(
 	}
 
 	/*
-	 * Edit timecode to remove control chars. Don't monkey with the
-	 * line buffer if the input buffer contains no ASCII printing
-	 * characters.
+	 * Copy the raw buffer to the user string. The string is padded
+	 * with a NULL, which is not included in the character count.
 	 */
 	if (dpend - dpt > bmax - 1)
 		dpend = dpt + bmax - 1;
-	for (dp = lineptr; dpt < dpend; dpt++) {
-		c = (char) (*dpt & 0x7f);
-		if (c >= ' ')
-			*dp++ = c;
-	}
+	for (dp = lineptr; dpt < dpend; dpt++)
+		*dp++ = *dpt;
+	*dp = '\0';
 	i = dp - lineptr;
-	if (i > 0)
-		*dp = '\0';
 #ifdef DEBUG
-	if (debug > 1) {
-		if (i > 0)
-			printf("refclock_gtlin: fd %d time %s timecode %d %s\n",
-			    rbufp->fd, ulfptoa(&trtmp, 6), i, lineptr);
-		else
-			printf("refclock_gtlin: fd %d time %s\n",
-			    rbufp->fd, ulfptoa(&trtmp, 6));
-	}
+	if (debug > 1)
+		printf("refclock_gtraw: fd %d time %s timecode %d %s\n",
+		    rbufp->fd, ulfptoa(&trtmp, 6), i, lineptr);
 #endif
 	*tsptr = trtmp;
 	return (i);
 }
 
+
 /*
  * The following code does not apply to WINNT & VMS ...
  */
@@ -674,24 +735,20 @@ refclock_gtlin(
  */
 int
 refclock_open(
-	char *dev,		/* device name pointer */
-	int speed,		/* serial port speed (code) */
-	int lflags		/* line discipline flags */
+	char	*dev,		/* device name pointer */
+	u_int	speed,		/* serial port speed (code) */
+	u_int	lflags		/* line discipline flags */
 	)
 {
-	int fd, i;
-	int flags;
-	TTY ttyb, *ttyp;
-#ifdef TIOCMGET
-	u_long ltemp;
-#endif /* TIOCMGET */
-	int omode;
+	int	fd;
+	int	omode;
+#ifdef O_NONBLOCK
+	char    trash[128];	/* litter bin for old input data */
+#endif
 
 	/*
 	 * Open serial port and set default options
 	 */
-	flags = lflags;
-
 	omode = O_RDWR;
 #ifdef O_NONBLOCK
 	omode |= O_NONBLOCK;
@@ -701,38 +758,66 @@ refclock_open(
 #endif
 
 	fd = open(dev, omode, 0777);
-
 	if (fd < 0) {
-		msyslog(LOG_ERR, "refclock_open: %s: %m", dev);
+		msyslog(LOG_ERR, "refclock_open %s: %m", dev);
 		return (0);
 	}
-
+	NTP_INSIST(fd != 0);
+	if (!refclock_setup(fd, speed, lflags)) {
+		close(fd);
+		return (0);
+	}
+	if (!refclock_ioctl(fd, lflags)) {
+		close(fd);
+		return (0);
+	}
+#ifdef O_NONBLOCK
 	/*
-	 * This little jewel lights up the PPS file descriptor if the
-	 * device name matches the name in the pps line in the
-	 * configuration file. This is so the atom driver can glom onto
-	 * the right device. Very silly.
+	 * We want to make sure there is no pending trash in the input
+	 * buffer. Since we have non-blocking IO available, this is a
+	 * good moment to read and dump all available outdated stuff
+	 * that might have become toxic for the driver.
 	 */
-	if (strcmp(dev, pps_device) == 0)
-		fdpps = fd;
+	while (read(fd, trash, sizeof(trash)) > 0 || errno == EINTR)
+		/*NOP*/;
+#endif
+	return (fd);
+}
+
+
+/*
+ * refclock_setup - initialize terminal interface structure
+ */
+int
+refclock_setup(
+	int	fd,		/* file descriptor */
+	u_int	speed,		/* serial port speed (code) */
+	u_int	lflags		/* line discipline flags */
+	)
+{
+	int	i;
+	TTY	ttyb, *ttyp;
+#ifdef PPS
+	fdpps = fd;		/* ppsclock legacy */
+#endif /* PPS */
 
 	/*
-	 * The following sections initialize the serial line port in
-	 * canonical (line-oriented) mode and set the specified line
-	 * speed, 8 bits and no parity. The modem control, break, erase
-	 * and kill functions are normally disabled. There is a
-	 * different section for each terminal interface, as selected at
-	 * compile time.
+	 * By default, the serial line port is initialized in canonical
+	 * (line-oriented) mode at specified line speed, 8 bits and no
+	 * parity. LF ends the line and CR is mapped to LF. The break,
+	 * erase and kill functions are disabled. There is a different
+	 * section for each terminal interface, as selected at compile
+	 * time. The flag bits can be used to set raw mode and echo.
 	 */
 	ttyp = &ttyb;
-
 #ifdef HAVE_TERMIOS
+
 	/*
 	 * POSIX serial line parameters (termios interface)
 	 */
 	if (tcgetattr(fd, ttyp) < 0) {
 		msyslog(LOG_ERR,
-			"refclock_open: fd %d tcgetattr: %m", fd);
+			"refclock_setup fd %d tcgetattr: %m", fd);
 		return (0);
 	}
 
@@ -740,53 +825,65 @@ refclock_open(
 	 * Set canonical mode and local connection; set specified speed,
 	 * 8 bits and no parity; map CR to NL; ignore break.
 	 */
-	ttyp->c_iflag = IGNBRK | IGNPAR | ICRNL;
-	ttyp->c_oflag = 0;
-	ttyp->c_cflag = CS8 | CLOCAL | CREAD;
-	(void)cfsetispeed(&ttyb, (u_int)speed);
-	(void)cfsetospeed(&ttyb, (u_int)speed);
-	ttyp->c_lflag = ICANON;
-	for (i = 0; i < NCCS; ++i)
-	{
-		ttyp->c_cc[i] = '\0';
+	if (speed) {
+		u_int	ltemp = 0;
+
+		ttyp->c_iflag = IGNBRK | IGNPAR | ICRNL;
+		ttyp->c_oflag = 0;
+		ttyp->c_cflag = CS8 | CLOCAL | CREAD;
+		if (lflags & LDISC_7O1) {
+			/* HP Z3801A needs 7-bit, odd parity */
+  			ttyp->c_cflag = CS7 | PARENB | PARODD | CLOCAL | CREAD;
+		}
+		cfsetispeed(&ttyb, speed);
+		cfsetospeed(&ttyb, speed);
+		for (i = 0; i < NCCS; ++i)
+			ttyp->c_cc[i] = '\0';
+
+#if defined(TIOCMGET) && !defined(SCO5_CLOCK)
+
+		/*
+		 * If we have modem control, check to see if modem leads
+		 * are active; if so, set remote connection. This is
+		 * necessary for the kernel pps mods to work.
+		 */
+		if (ioctl(fd, TIOCMGET, (char *)&ltemp) < 0)
+			msyslog(LOG_ERR,
+			    "refclock_setup fd %d TIOCMGET: %m", fd);
+#ifdef DEBUG
+		if (debug)
+			printf("refclock_setup fd %d modem status: 0x%x\n",
+			    fd, ltemp);
+#endif
+		if (ltemp & TIOCM_DSR && lflags & LDISC_REMOTE)
+			ttyp->c_cflag &= ~CLOCAL;
+#endif /* TIOCMGET */
 	}
 
 	/*
-	 * Some special cases
+	 * Set raw and echo modes. These can be changed on-fly.
 	 */
-	if (flags & LDISC_RAW) {
-		ttyp->c_iflag = 0;
+	ttyp->c_lflag = ICANON;
+	if (lflags & LDISC_RAW) {
 		ttyp->c_lflag = 0;
+		ttyp->c_iflag = 0;
 		ttyp->c_cc[VMIN] = 1;
 	}
-#if defined(TIOCMGET) && !defined(SCO5_CLOCK)
-	/*
-	 * If we have modem control, check to see if modem leads are
-	 * active; if so, set remote connection. This is necessary for
-	 * the kernel pps mods to work.
-	 */
-	ltemp = 0;
-	if (ioctl(fd, TIOCMGET, (char *)&ltemp) < 0)
-		msyslog(LOG_ERR,
-			"refclock_open: fd %d TIOCMGET failed: %m", fd);
-#ifdef DEBUG
-	if (debug)
-		printf("refclock_open: fd %d modem status 0x%lx\n",
-		    fd, ltemp);
-#endif
-	if (ltemp & TIOCM_DSR)
-		ttyp->c_cflag &= ~CLOCAL;
-#endif /* TIOCMGET */
+	if (lflags & LDISC_ECHO)
+		ttyp->c_lflag |= ECHO;
 	if (tcsetattr(fd, TCSANOW, ttyp) < 0) {
 		msyslog(LOG_ERR,
-		    "refclock_open: fd %d TCSANOW failed: %m", fd);
-		return (0);
-	}
-	if (tcflush(fd, TCIOFLUSH) < 0) {
-		msyslog(LOG_ERR,
-		    "refclock_open: fd %d TCIOFLUSH failed: %m", fd);
+		    "refclock_setup fd %d TCSANOW: %m", fd);
 		return (0);
 	}
+
+	/*
+	 * flush input and output buffers to discard any outdated stuff
+	 * that might have become toxic for the driver. Failing to do so
+	 * is logged, but we keep our fingers crossed otherwise.
+	 */
+	if (tcflush(fd, TCIOFLUSH) < 0)
+		msyslog(LOG_ERR, "refclock_setup fd %d tcflush(): %m", fd);
 #endif /* HAVE_TERMIOS */
 
 #ifdef HAVE_SYSV_TTYS
@@ -797,7 +894,7 @@ refclock_open(
 	 */
 	if (ioctl(fd, TCGETA, ttyp) < 0) {
 		msyslog(LOG_ERR,
-		    "refclock_open: fd %d TCGETA failed: %m", fd);
+		    "refclock_setup fd %d TCGETA: %m", fd);
 		return (0);
 	}
 
@@ -805,40 +902,47 @@ refclock_open(
 	 * Set canonical mode and local connection; set specified speed,
 	 * 8 bits and no parity; map CR to NL; ignore break.
 	 */
-	ttyp->c_iflag = IGNBRK | IGNPAR | ICRNL;
-	ttyp->c_oflag = 0;
-	ttyp->c_cflag = speed | CS8 | CLOCAL | CREAD;
-	ttyp->c_lflag = ICANON;
-	ttyp->c_cc[VERASE] = ttyp->c_cc[VKILL] = '\0';
+	if (speed) {
+		u_int	ltemp = 0;
+
+		ttyp->c_iflag = IGNBRK | IGNPAR | ICRNL;
+		ttyp->c_oflag = 0;
+		ttyp->c_cflag = speed | CS8 | CLOCAL | CREAD;
+		for (i = 0; i < NCCS; ++i)
+			ttyp->c_cc[i] = '\0';
+
+#if defined(TIOCMGET) && !defined(SCO5_CLOCK)
+
+		/*
+		 * If we have modem control, check to see if modem leads
+		 * are active; if so, set remote connection. This is
+		 * necessary for the kernel pps mods to work.
+		 */
+		if (ioctl(fd, TIOCMGET, (char *)&ltemp) < 0)
+			msyslog(LOG_ERR,
+			    "refclock_setup fd %d TIOCMGET: %m", fd);
+#ifdef DEBUG
+		if (debug)
+			printf("refclock_setup fd %d modem status: %x\n",
+			    fd, ltemp);
+#endif
+		if (ltemp & TIOCM_DSR)
+			ttyp->c_cflag &= ~CLOCAL;
+#endif /* TIOCMGET */
+	}
 
 	/*
-	 * Some special cases
+	 * Set raw and echo modes. These can be changed on-fly.
 	 */
-	if (flags & LDISC_RAW) {
-		ttyp->c_iflag = 0;
+	ttyp->c_lflag = ICANON;
+	if (lflags & LDISC_RAW) {
 		ttyp->c_lflag = 0;
+		ttyp->c_iflag = 0;
+		ttyp->c_cc[VMIN] = 1;
 	}
-#ifdef TIOCMGET
-	/*
-	 * If we have modem control, check to see if modem leads are
-	 * active; if so, set remote connection. This is necessary for
-	 * the kernel pps mods to work.
-	 */
-	ltemp = 0;
-	if (ioctl(fd, TIOCMGET, (char *)&ltemp) < 0)
-		msyslog(LOG_ERR,
-		    "refclock_open: fd %d TIOCMGET failed: %m", fd);
-#ifdef DEBUG
-	if (debug)
-		printf("refclock_open: fd %d modem status %lx\n",
-		    fd, ltemp);
-#endif
-	if (ltemp & TIOCM_DSR)
-		ttyp->c_cflag &= ~CLOCAL;
-#endif /* TIOCMGET */
 	if (ioctl(fd, TCSETA, ttyp) < 0) {
 		msyslog(LOG_ERR,
-		    "refclock_open: fd %d TCSETA failed: %m", fd);
+		    "refclock_setup fd %d TCSETA: %m", fd);
 		return (0);
 	}
 #endif /* HAVE_SYSV_TTYS */
@@ -850,28 +954,24 @@ refclock_open(
 	 */
 	if (ioctl(fd, TIOCGETP, (char *)ttyp) < 0) {
 		msyslog(LOG_ERR,
-		    "refclock_open: fd %d TIOCGETP %m", fd);
+		    "refclock_setup fd %d TIOCGETP: %m", fd);
 		return (0);
 	}
-	ttyp->sg_ispeed = ttyp->sg_ospeed = speed;
+	if (speed)
+		ttyp->sg_ispeed = ttyp->sg_ospeed = speed;
 	ttyp->sg_flags = EVENP | ODDP | CRMOD;
 	if (ioctl(fd, TIOCSETP, (char *)ttyp) < 0) {
 		msyslog(LOG_ERR,
-		    "refclock_open: TIOCSETP failed: %m");
+		    "refclock_setup TIOCSETP: %m");
 		return (0);
 	}
 #endif /* HAVE_BSD_TTYS */
-	if (!refclock_ioctl(fd, flags)) {
-		(void)close(fd);
-		msyslog(LOG_ERR,
-		    "refclock_open: fd %d ioctl failed: %m", fd);
-		return (0);
-	}
-	return (fd);
+	return(1);
 }
 #endif /* HAVE_TERMIOS || HAVE_SYSV_TTYS || HAVE_BSD_TTYS */
 #endif /* SYS_VXWORKS SYS_WINNT */
 
+
 /*
  * refclock_ioctl - set serial port control functions
  *
@@ -883,68 +983,51 @@ refclock_open(
  */
 int
 refclock_ioctl(
-	int fd, 		/* file descriptor */
-	int flags		/* line discipline flags */
+	int	fd, 		/* file descriptor */
+	u_int	lflags		/* line discipline flags */
 	)
 {
-	/* simply return 1 if no UNIX line discipline is supported */
+	/*
+	 * simply return 1 if no UNIX line discipline is supported
+	 */
 #if !defined SYS_VXWORKS && !defined SYS_WINNT
 #if defined(HAVE_TERMIOS) || defined(HAVE_SYSV_TTYS) || defined(HAVE_BSD_TTYS)
 
-#ifdef TTYCLK
-	TTY ttyb, *ttyp;
-#endif /* TTYCLK */
-
 #ifdef DEBUG
 	if (debug)
-		printf("refclock_ioctl: fd %d flags 0x%x\n", fd, flags);
+		printf("refclock_ioctl: fd %d flags 0x%x\n", fd,
+		    lflags);
 #endif
-	if (flags == 0)
-		return (1);
-#if !(defined(HAVE_TERMIOS) || defined(HAVE_BSD_TTYS))
-	if (flags & (LDISC_CLK | LDISC_PPS | LDISC_ACTS)) {
-		msyslog(LOG_ERR,
-			"refclock_ioctl: unsupported terminal interface");
-		return (0);
-	}
-#endif /* HAVE_TERMIOS HAVE_BSD_TTYS */
 #ifdef TTYCLK
-	ttyp = &ttyb;
-#endif /* TTYCLK */
 
-	/*
-	 * The following features may or may not require System V
-	 * STREAMS support, depending on the particular implementation.
-	 */
-#if defined(TTYCLK)
 	/*
 	 * The TTYCLK option provides timestamping at the driver level.
 	 * It requires the tty_clk streams module and System V STREAMS
 	 * support. If not available, don't complain.
 	 */
-	if (flags & (LDISC_CLK | LDISC_CLKPPS | LDISC_ACTS)) {
+	if (lflags & (LDISC_CLK | LDISC_CLKPPS | LDISC_ACTS)) {
 		int rval = 0;
 
 		if (ioctl(fd, I_PUSH, "clk") < 0) {
 			msyslog(LOG_NOTICE,
-			    "refclock_ioctl: I_PUSH clk failed: %m");
+			    "refclock_ioctl fd %d I_PUSH: %m", fd);
+			return (0);
+#ifdef CLK_SETSTR
 		} else {
 			char *str;
 
-			if (flags & LDISC_CLKPPS)
+			if (lflags & LDISC_CLKPPS)
 				str = "\377";
-			else if (flags & LDISC_ACTS)
+			else if (lflags & LDISC_ACTS)
 				str = "*";
 			else
 				str = "\n";
-#ifdef CLK_SETSTR
-			if ((rval = ioctl(fd, CLK_SETSTR, str)) < 0)
+			if (ioctl(fd, CLK_SETSTR, str) < 0) {
 				msyslog(LOG_ERR,
-				    "refclock_ioctl: CLK_SETSTR failed: %m");
-			if (debug)
-				printf("refclock_ioctl: fd %d CLK_SETSTR %d str %s\n",
-				    fd, rval, str);
-#endif
+				    "refclock_ioctl fd %d CLK_SETSTR: %m", fd);
+				return (0);
+			}
+#endif /*CLK_SETSTR */
 		}
 	}
 #endif /* TTYCLK */
@@ -953,6 +1036,7 @@ refclock_ioctl(
 	return (1);
 }
 
+
 /*
  * refclock_control - set and/or return clock values
  *
@@ -964,7 +1048,7 @@ refclock_ioctl(
  */
 void
 refclock_control(
-	struct sockaddr_storage *srcadr,
+	sockaddr_u *srcadr,
 	struct refclockstat *in,
 	struct refclockstat *out
 	)
@@ -977,19 +1061,17 @@ refclock_control(
 	/*
 	 * Check for valid address and running peer
 	 */
-	if (srcadr->ss_family != AF_INET)
-		return;
 	if (!ISREFCLOCKADR(srcadr))
 		return;
+
 	clktype = (u_char)REFCLOCKTYPE(srcadr);
 	unit = REFCLOCKUNIT(srcadr);
-	if (clktype >= num_refclock_conf || unit >= MAXUNIT)
-		return;
-	peer = typeunit[clktype][unit];
-	if (peer == NULL)
-		return;
-	if (peer->procptr == NULL)
+
+	peer = findexistingpeer(srcadr, NULL, -1);
+
+	if (NULL == peer || NULL == peer->procptr)
 		return;
+
 	pp = peer->procptr;
 
 	/*
@@ -1001,16 +1083,9 @@ refclock_control(
 		if (in->haveflags & CLK_HAVETIME2)
 			pp->fudgetime2 = in->fudgetime2;
 		if (in->haveflags & CLK_HAVEVAL1)
-			pp->stratum = (u_char) in->fudgeval1;
+			peer->stratum = pp->stratum = (u_char)in->fudgeval1;
 		if (in->haveflags & CLK_HAVEVAL2)
-			pp->refid = in->fudgeval2;
-		peer->stratum = pp->stratum;
-		if (peer->stratum == STRATUM_REFCLOCK || peer->stratum ==
-		    STRATUM_UNSPEC)
-			peer->refid = pp->refid;
-		else
-			peer->refid = ((struct
-			    sockaddr_in*)&peer->srcadr)->sin_addr.s_addr;
+			peer->refid = pp->refid = in->fudgeval2;
 		if (in->haveflags & CLK_HAVEFLAG1) {
 			pp->sloppyclockflag &= ~CLK_FLAG1;
 			pp->sloppyclockflag |= in->flags & CLK_FLAG1;
@@ -1051,7 +1126,7 @@ refclock_control(
 		out->currentstatus = pp->currentstatus;
 		out->type = pp->type;
 		out->clockdesc = pp->clockdesc;
-		out->lencode = pp->lencode;
+		out->lencode = (u_short)pp->lencode;
 		out->p_lastcode = pp->a_lastcode;
 	}
 
@@ -1072,30 +1147,30 @@ refclock_control(
  */
 void
 refclock_buginfo(
-	struct sockaddr_storage *srcadr, /* clock address */
+	sockaddr_u *srcadr,	/* clock address */
 	struct refclockbug *bug /* output structure */
 	)
 {
 	struct peer *peer;
 	struct refclockproc *pp;
-	u_char clktype;
+	int clktype;
 	int unit;
-	int i;
+	unsigned u;
 
 	/*
 	 * Check for valid address and peer structure
 	 */
-	if (srcadr->ss_family != AF_INET)
-		return;
 	if (!ISREFCLOCKADR(srcadr))
 		return;
+
 	clktype = (u_char) REFCLOCKTYPE(srcadr);
 	unit = REFCLOCKUNIT(srcadr);
-	if (clktype >= num_refclock_conf || unit >= MAXUNIT)
-		return;
-	peer = typeunit[clktype][unit];
-	if (peer == NULL)
+
+	peer = findexistingpeer(srcadr, NULL, -1);
+
+	if (NULL == peer || NULL == peer->procptr)
 		return;
+
 	pp = peer->procptr;
 
 	/*
@@ -1114,8 +1189,8 @@ refclock_buginfo(
 	bug->stimes = 0xfffffffc;
 	bug->times[0] = pp->lastref;
 	bug->times[1] = pp->lastrec;
-	for (i = 2; i < (int)bug->ntimes; i++)
-		DTOLFP(pp->filter[i - 2], &bug->times[i]);
+	for (u = 2; u < bug->ntimes; u++)
+		DTOLFP(pp->filter[u - 2], &bug->times[u]);
 
 	/*
 	 * Give the stuff to the clock
@@ -1124,4 +1199,154 @@ refclock_buginfo(
 		(refclock_conf[clktype]->clock_buginfo)(unit, bug, peer);
 }
 
+
+#ifdef HAVE_PPSAPI
+/*
+ * refclock_ppsapi - initialize/update ppsapi
+ *
+ * This routine is called after the fudge command to open the PPSAPI
+ * interface for later parameter setting after the fudge command.
+ */
+int
+refclock_ppsapi(
+	int	fddev,			/* fd device */
+	struct refclock_atom *ap	/* atom structure pointer */
+	)
+{
+	if (ap->handle == 0) {
+		if (time_pps_create(fddev, &ap->handle) < 0) {
+			msyslog(LOG_ERR,
+			    "refclock_ppsapi: time_pps_create: %m");
+			return (0);
+		}
+	}
+	return (1);
+}
+
+
+/*
+ * refclock_params - set ppsapi parameters
+ *
+ * This routine is called to set the PPSAPI parameters after the fudge
+ * command.
+ */
+int
+refclock_params(
+	int	mode,			/* mode bits */
+	struct refclock_atom *ap	/* atom structure pointer */
+	)
+{
+	memset(&ap->pps_params, 0, sizeof(pps_params_t));
+	ap->pps_params.api_version = PPS_API_VERS_1;
+
+	/*
+	 * Solaris serial ports provide PPS pulse capture only on the
+	 * assert edge. FreeBSD serial ports provide capture on the
+	 * clear edge, while FreeBSD parallel ports provide capture
+	 * on the assert edge. Your mileage may vary.
+	 */
+	if (mode & CLK_FLAG2)
+		ap->pps_params.mode = PPS_TSFMT_TSPEC | PPS_CAPTURECLEAR;
+	else
+		ap->pps_params.mode = PPS_TSFMT_TSPEC | PPS_CAPTUREASSERT;
+	if (time_pps_setparams(ap->handle, &ap->pps_params) < 0) {
+		msyslog(LOG_ERR,
+		    "refclock_params: time_pps_setparams: %m");
+		return (0);
+	}
+
+	/*
+	 * If flag3 is lit, select the kernel PPS.
+	 */
+	if (mode & CLK_FLAG3) {
+		if (time_pps_kcbind(ap->handle, PPS_KC_HARDPPS,
+		    ap->pps_params.mode & ~PPS_TSFMT_TSPEC,
+		    PPS_TSFMT_TSPEC) < 0) {
+			if (errno != EOPNOTSUPP) { 
+				msyslog(LOG_ERR,
+				    "refclock_params: time_pps_kcbind: %m");
+				return (0);
+			}
+		}
+		pps_enable = 1;
+	}
+	return (1);
+}
+
+
+/*
+ * refclock_pps - called once per second
+ *
+ * This routine is called once per second. It snatches the PPS
+ * timestamp from the kernel and saves the sign-extended fraction in
+ * a circular buffer for processing at the next poll event.
+ */
+int
+refclock_pps(
+	struct peer *peer,		/* peer structure pointer */
+	struct refclock_atom *ap,	/* atom structure pointer */
+	int	mode			/* mode bits */	
+	)
+{
+	struct refclockproc *pp;
+	pps_info_t pps_info;
+	struct timespec timeout;
+	double	dtemp;
+
+	/*
+	 * We require the clock to be synchronized before setting the
+	 * parameters. When the parameters have been set, fetch the
+	 * most recent PPS timestamp.
+	 */ 
+	pp = peer->procptr;
+	if (ap->handle == 0)
+		return (0);
+
+	if (ap->pps_params.mode == 0 && sys_leap != LEAP_NOTINSYNC) {
+		if (refclock_params(pp->sloppyclockflag, ap) < 1)
+			return (0);
+	}
+	timeout.tv_sec = 0;
+	timeout.tv_nsec = 0;
+	memset(&pps_info, 0, sizeof(pps_info_t));
+	if (time_pps_fetch(ap->handle, PPS_TSFMT_TSPEC, &pps_info,
+	    &timeout) < 0) {
+		refclock_report(peer, CEVNT_FAULT);
+		return (0);
+	}
+	timeout = ap->ts;
+	if (ap->pps_params.mode & PPS_CAPTUREASSERT)
+		ap->ts = pps_info.assert_timestamp;
+	else if (ap->pps_params.mode & PPS_CAPTURECLEAR)
+		ap->ts = pps_info.clear_timestamp;
+	else
+		return (0);
+	
+	/*
+	 * There can be zero, one or two PPS pulses between polls,
+	 * depending on the poll interval relative to the PPS interval.
+	 * The pulse must be newer and within the range gate relative
+	 * to the last pulse.
+	 */
+	if (ap->ts.tv_sec <= timeout.tv_sec || abs(ap->ts.tv_nsec -
+	    timeout.tv_nsec) > RANGEGATE)
+		return (0);
+
+	/*
+	 * Convert to signed fraction offset and stuff in median filter.
+	 */
+	pp->lastrec.l_ui = (u_int32)ap->ts.tv_sec + JAN_1970;
+	dtemp = ap->ts.tv_nsec / 1e9;
+	pp->lastrec.l_uf = (u_int32)(dtemp * FRAC);
+	if (dtemp > .5)
+		dtemp -= 1.;
+	SAMPLE(-dtemp + pp->fudgetime1);
+#ifdef DEBUG
+	if (debug > 1)
+		printf("refclock_pps: %lu %f %f\n", current_time,
+		    dtemp, pp->fudgetime1);
+#endif
+	return (1);
+}
+#endif /* HAVE_PPSAPI */
 #endif /* REFCLOCK */
