@@ -1,5 +1,5 @@
 /*
- * refclock_arc - clock driver for ARCRON MSF receivers
+ * refclock_arc - clock driver for ARCRON MSF/DCF/WWVB receivers
  */
 
 #ifdef HAVE_CONFIG_H
@@ -7,9 +7,12 @@
 #endif
 
 #if defined(REFCLOCK) && defined(CLOCK_ARCRON_MSF)
-static const char arc_version[] = { "V1.1 1997/06/23" };
 
-#undef ARCRON_DEBUG /* Define only while in development... */
+static const char arc_version[] = { "V1.3 2003/02/21" };
+
+/* define PRE_NTP420 for compatibility to previous versions of NTP (at least
+   to 4.1.0 */
+#undef PRE_NTP420
 
 #ifndef ARCRON_NOT_KEEN
 #define ARCRON_KEEN 1 /* Be keen, and trusting of the clock, if defined. */
@@ -28,6 +31,10 @@ static const char arc_version[] = { "V1.1 1997/06/23" };
 /*
 Code by Derek Mulcahy, <derek@toybox.demon.co.uk>, 1997.
 Modifications by Damon Hart-Davis, <d@hd.org>, 1997.
+Modifications by Paul Alfille, <palfille@partners.org>, 2003.
+Modifications by Christopher Price, <cprice@cs-home.com>, 2003.
+Modifications by Nigel Roles <nigel@9fs.org>, 2003.
+
 
 THIS CODE IS SUPPLIED AS IS, WITH NO WARRANTY OF ANY KIND.  USE AT
 YOUR OWN RISK.
@@ -42,6 +49,43 @@ reproduced.
 
 -------------------------------------------------------------------------------
 
+Nigel's notes:
+
+1) Called tcgetattr() before modifying, so that fields correctly initialised
+   for all operating systems
+
+2) Altered parsing of timestamp line so that it copes with fields which are
+   not always ASCII digits (e.g. status field when battery low)
+
+-------------------------------------------------------------------------------
+
+Christopher's notes:
+
+MAJOR CHANGES SINCE V1.2 
+========================
+ 1) Applied patch by Andrey Bray <abuse@madhouse.demon.co.uk>
+    2001-02-17 comp.protocols.time.ntp
+
+ 2) Added WWVB support via clock mode command, localtime/UTC time configured
+    via flag1=(0=UTC, 1=localtime)
+
+ 3) Added ignore resync request via flag2=(0=resync, 1=ignore resync)
+
+ 4) Added simplified conversion from localtime to UTC with dst/bst translation
+
+ 5) Added average signal quality poll
+
+ 6) Fixed a badformat error when no code is available due to stripping 
+    \n & \r's 
+
+ 7) Fixed a badformat error when clearing lencode & memset a_lastcode in poll
+    routine
+
+ 8) Lots of code cleanup, including standardized DEBUG macros and removal 
+    of unused code 
+
+-------------------------------------------------------------------------------
+
 Author's original note:
 
 I enclose my ntp driver for the Galleon Systems Arc MSF receiver.
@@ -94,27 +138,27 @@ GENERAL
  2) PRECISION should be -4/-5 (63ms/31ms) for the following reasons:
 
      a) The ARC documentation claims the internal clock is (only)
-        accurate to about 20ms relative to Rugby (plus there must be
-        noticable drift and delay in the ms range due to transmission
-        delays and changing atmospheric effects).  This clock is not
-        designed for ms accuracy as NTP has spoilt us all to expect.
+	accurate to about 20ms relative to Rugby (plus there must be
+	noticable drift and delay in the ms range due to transmission
+	delays and changing atmospheric effects).  This clock is not
+	designed for ms accuracy as NTP has spoilt us all to expect.
 
      b) The clock oscillator looks like a simple uncompensated quartz
-        crystal of the sort used in digital watches (ie 32768Hz) which
-        can have large temperature coefficients and drifts; it is not
-        clear if this oscillator is properly disciplined to the MSF
-        transmission, but as the default is to resync only once per
-        *day*, we can imagine that it is not, and is free-running.  We
-        can minimise drift by resyncing more often (at the cost of
-        reduced battery life), but drift/wander may still be
-        significant.
+	crystal of the sort used in digital watches (ie 32768Hz) which
+	can have large temperature coefficients and drifts; it is not
+	clear if this oscillator is properly disciplined to the MSF
+	transmission, but as the default is to resync only once per
+	*day*, we can imagine that it is not, and is free-running.  We
+	can minimise drift by resyncing more often (at the cost of
+	reduced battery life), but drift/wander may still be
+	significant.
 
      c) Note that the bit time of 3.3ms adds to the potential error in
-        the the clock timestamp, since the bit clock of the serial link
-        may effectively be free-running with respect to the host clock
-        and the MSF clock.  Actually, the error is probably 1/16th of
-        the above, since the input data is probably sampled at at least
-        16x the bit rate.
+	the the clock timestamp, since the bit clock of the serial link
+	may effectively be free-running with respect to the host clock
+	and the MSF clock.  Actually, the error is probably 1/16th of
+	the above, since the input data is probably sampled at at least
+	16x the bit rate.
 
     By keeping the clock marked as not very precise, it will have a
     fairly large dispersion, and thus will tend to be used as a
@@ -134,9 +178,9 @@ GENERAL
     IN ANY CASE, BE SURE TO SET AN APPROPRIATE FUDGE FACTOR TO REMOVE
     ANY RESIDUAL SKEW, eg:
 
-        server 127.127.27.0 # ARCRON MSF radio clock unit 0.
-        # Fudge timestamps by about 20ms.
-        fudge 127.127.27.0 time1 0.020
+	server 127.127.27.0 # ARCRON MSF radio clock unit 0.
+	# Fudge timestamps by about 20ms.
+	fudge 127.127.27.0 time1 0.020
 
     You will need to observe your system's behaviour, assuming you have
     some other NTP source to compare it with, to work out what the
@@ -273,28 +317,28 @@ You have to wait for character echo + 10ms before sending next character.
 /* 12. year tens */
 /* 13. year units */
 /* 14. BST/UTC status */
-/*      bit 7   parity */
-/*      bit 6   always 0 */
-/*      bit 5   always 1 */
-/*      bit 4   always 1 */
-/*      bit 3   always 0 */
-/*      bit 2   =1 if UTC is in effect, complementary to the BST bit */
-/*      bit 1   =1 if BST is in effect, according to the BST bit     */
-/*      bit 0   BST/UTC change impending bit=1 in case of change impending */
+/*	bit 7	parity */
+/*	bit 6	always 0 */
+/*	bit 5	always 1 */
+/*	bit 4	always 1 */
+/*	bit 3	always 0 */
+/*	bit 2	=1 if UTC is in effect, complementary to the BST bit */
+/*	bit 1	=1 if BST is in effect, according to the BST bit     */
+/*	bit 0	BST/UTC change impending bit=1 in case of change impending */
 /* 15. status */
-/*      bit 7   parity */
-/*      bit 6   always 0 */
-/*      bit 5   always 1 */
-/*      bit 4   always 1 */
-/*      bit 3   =1 if low battery is detected */
-/*      bit 2   =1 if the very last reception attempt failed and a valid */
-/*              time information already exists (bit0=1) */
-/*              =0 if the last reception attempt was successful */
-/*      bit 1   =1 if at least one reception since 2:30 am was successful */
-/*              =0 if no reception attempt since 2:30 am was successful */
-/*      bit 0   =1 if the RC Computer Clock contains valid time information */
-/*              This bit is zero after reset and one after the first */
-/*              successful reception attempt */
+/*	bit 7	parity */
+/*	bit 6	always 0 */
+/*	bit 5	always 1 */
+/*	bit 4	always 1 */
+/*	bit 3	=1 if low battery is detected */
+/*	bit 2	=1 if the very last reception attempt failed and a valid */
+/*		time information already exists (bit0=1) */
+/*		=0 if the last reception attempt was successful */
+/*	bit 1	=1 if at least one reception since 2:30 am was successful */
+/*		=0 if no reception attempt since 2:30 am was successful */
+/*	bit 0	=1 if the RC Computer Clock contains valid time information */
+/*		This bit is zero after reset and one after the first */
+/*		successful reception attempt */
 
 /* DHD note:
 Also note g<cr> command which confirms that a resync is in progress, and
@@ -306,6 +350,7 @@ Also note h<cr> command which starts a resync to MSF signal.
 #include "ntpd.h"
 #include "ntp_io.h"
 #include "ntp_refclock.h"
+#include "ntp_calendar.h"
 #include "ntp_stdlib.h"
 
 #include <stdio.h>
@@ -324,28 +369,35 @@ Also note h<cr> command which starts a resync to MSF signal.
 #endif
 
 /*
- * This driver supports the ARCRON MSF Radio Controlled Clock
+ * This driver supports the ARCRON MSF/DCF/WWVB Radio Controlled Clock
  */
 
 /*
  * Interface definitions
  */
-#define DEVICE          "/dev/arc%d"    /* Device name and unit. */
-#define SPEED           B300            /* UART speed (300 baud) */
-#define PRECISION       (-4)            /* Precision  (~63 ms). */
-#define HIGHPRECISION   (-5)            /* If things are going well... */
-#define REFID           "MSFa"          /* Reference ID. */
-#define DESCRIPTION     "ARCRON MSF Receiver"
+#define DEVICE		"/dev/arc%d"	/* Device name and unit. */
+#define SPEED		B300		/* UART speed (300 baud) */
+#define PRECISION	(-4)		/* Precision  (~63 ms). */
+#define HIGHPRECISION	(-5)		/* If things are going well... */
+#define REFID		"MSFa"		/* Reference ID. */
+#define REFID_MSF	"MSF"		/* Reference ID. */
+#define REFID_DCF77	"DCF"		/* Reference ID. */
+#define REFID_WWVB	"WWVB"		/* Reference ID. */
+#define DESCRIPTION	"ARCRON MSF/DCF/WWVB Receiver"
 
-#define NSAMPLESLONG    8               /* Stages of long filter. */
+#ifdef PRE_NTP420
+#define MODE ttlmax
+#else
+#define MODE ttl
+#endif
 
-#define LENARC          16              /* Format `o' timecode length. */
+#define LENARC		16		/* Format `o' timecode length. */
 
-#define BITSPERCHAR     11              /* Bits per character. */
-#define BITTIME         0x0DA740E       /* Time for 1 bit at 300bps. */
-#define CHARTIME10      0x8888888       /* Time for 10-bit char at 300bps. */
-#define CHARTIME11      0x962FC96       /* Time for 11-bit char at 300bps. */
-#define CHARTIME                        /* Time for char at 300bps. */ \
+#define BITSPERCHAR	11		/* Bits per character. */
+#define BITTIME		0x0DA740E	/* Time for 1 bit at 300bps. */
+#define CHARTIME10	0x8888888	/* Time for 10-bit char at 300bps. */
+#define CHARTIME11	0x962FC96	/* Time for 11-bit char at 300bps. */
+#define CHARTIME			/* Time for char at 300bps. */ \
 ( (BITSPERCHAR == 11) ? CHARTIME11 : ( (BITSPERCHAR == 10) ? CHARTIME10 : \
 				       (BITSPERCHAR * BITTIME) ) )
 
@@ -405,28 +457,12 @@ Also note h<cr> command which starts a resync to MSF signal.
 #endif
      };
 
-/* Chose filter length dependent on fudge flag 4. */
-#define CHOSENSAMPLES(pp) \
-(((pp)->sloppyclockflag & CLK_FLAG4) ? NSAMPLESLONG : NSAMPLES)
-     /*
-Chose how many filter samples to keep.  Several factors are in play.
-
- 1) Discard at least one sample to allow a spike value to be
-    discarded.
-
- 2) Discard about 1-in-8 to 1-in-30 samples to handle spikes.
-
- 3) Keep an odd number of samples to avoid median value being biased
-    high or low.
-*/
-#define NKEEP(pp) ((CHOSENSAMPLES(pp) - 1 - (CHOSENSAMPLES(pp)>>3)) | 1)
-
-#define DEFAULT_RESYNC_TIME  (57*60)    /* Gap between resync attempts (s). */
-#define RETRY_RESYNC_TIME    (27*60)    /* Gap to emergency resync attempt. */
+#define DEFAULT_RESYNC_TIME  (57*60)	/* Gap between resync attempts (s). */
+#define RETRY_RESYNC_TIME    (27*60)	/* Gap to emergency resync attempt. */
 #ifdef ARCRON_KEEN
-#define INITIAL_RESYNC_DELAY 500        /* Delay before first resync. */
+#define INITIAL_RESYNC_DELAY 500	/* Delay before first resync. */
 #else
-#define INITIAL_RESYNC_DELAY 50         /* Delay before first resync. */
+#define INITIAL_RESYNC_DELAY 50		/* Delay before first resync. */
 #endif
 
      static const int moff[12] =
@@ -434,35 +470,37 @@ Chose how many filter samples to keep.  Several factors are in play.
 /* Flags for a raw open() of the clock serial device. */
 #ifdef O_NOCTTY /* Good, we can avoid tty becoming controlling tty. */
 #define OPEN_FLAGS (O_RDWR | O_NOCTTY)
-#else           /* Oh well, it may not matter... */
+#else		/* Oh well, it may not matter... */
 #define OPEN_FLAGS (O_RDWR)
 #endif
 
 
 /* Length of queue of command bytes to be sent. */
-#define CMDQUEUELEN 4                   /* Enough for two cmds + each \r. */
+#define CMDQUEUELEN 4			/* Enough for two cmds + each \r. */
 /* Queue tick time; interval in seconds between chars taken off queue. */
 /* Must be >= 2 to allow o\r response to come back uninterrupted. */
-#define QUEUETICK   2                   /* Allow o\r reply to finish. */
+#define QUEUETICK   2			/* Allow o\r reply to finish. */
 
 /*
  * ARC unit control structure
  */
 struct arcunit {
-	l_fp lastrec;       /* Time tag for the receive time (system). */
-	int status;         /* Clock status. */
+	l_fp lastrec;	    /* Time tag for the receive time (system). */
+	int status;	    /* Clock status. */
 
-	int quality;        /* Quality of reception 0--5 for unit. */
+	int quality;	    /* Quality of reception 0--5 for unit. */
 	/* We may also use the values -1 or 6 internally. */
+	u_long quality_stamp; /* Next time to reset quality average. */
 
 	u_long next_resync; /* Next resync time (s) compared to current_time. */
-	int resyncing;      /* Resync in progress if true. */
+	int resyncing;	    /* Resync in progress if true. */
 
 	/* In the outgoing queue, cmdqueue[0] is next to be sent. */
 	char cmdqueue[CMDQUEUELEN+1]; /* Queue of outgoing commands + \0. */
 
 	u_long saved_flags; /* Saved fudge flags. */
 };
+
 #ifdef ARCRON_LEAPSECOND_KEEN
 /* The flag `possible_leap' is set non-zero when any MSF unit
        thinks a leap-second may have happened.
@@ -489,11 +527,11 @@ static int possible_leap = 0;       /* No resync required by default. */
 #endif
 
 #if 0
-static void dummy_event_handler P((struct peer *));
-static void   arc_event_handler P((struct peer *));
+static void dummy_event_handler (struct peer *);
+static void   arc_event_handler (struct peer *);
 #endif /* 0 */
 
-#define QUALITY_UNKNOWN     -1 /* Indicates unknown clock quality. */
+#define QUALITY_UNKNOWN	    -1 /* Indicates unknown clock quality. */
 #define MIN_CLOCK_QUALITY    0 /* Min quality clock will return. */
 #define MIN_CLOCK_QUALITY_OK 3 /* Min quality for OK reception. */
 #define MAX_CLOCK_QUALITY    5 /* Max quality clock will return. */
@@ -501,39 +539,37 @@ static void   arc_event_handler P((struct peer *));
 /*
  * Function prototypes
  */
-static  int     arc_start       P((int, struct peer *));
-static  void    arc_shutdown    P((int, struct peer *));
-static  void    arc_receive     P((struct recvbuf *));
-static  void    arc_poll        P((int, struct peer *));
+static	int	arc_start	(int, struct peer *);
+static	void	arc_shutdown	(int, struct peer *);
+static	void	arc_receive	(struct recvbuf *);
+static	void	arc_poll	(int, struct peer *);
 
 /*
  * Transfer vector
  */
 struct  refclock refclock_arc = {
-	arc_start,              /* start up driver */
-	arc_shutdown,           /* shut down driver */
-	arc_poll,               /* transmit poll message */
-	noentry,                /* not used (old arc_control) */
-	noentry,                /* initialize driver (not used) */
-	noentry,                /* not used (old arc_buginfo) */
-	NOFLAGS                 /* not used */
+	arc_start,		/* start up driver */
+	arc_shutdown,		/* shut down driver */
+	arc_poll,		/* transmit poll message */
+	noentry,		/* not used (old arc_control) */
+	noentry,		/* initialize driver (not used) */
+	noentry,		/* not used (old arc_buginfo) */
+	NOFLAGS			/* not used */
 };
 
 /* Queue us up for the next tick. */
 #define ENQUEUE(up) \
 	do { \
-	     if((up)->ev.next != 0) { break; } /* WHOOPS! */ \
-	     peer->nextdate = current_time + QUEUETICK; \
+	     peer->nextaction = current_time + QUEUETICK; \
 	} while(0)
 
-#if 0
-/* Placeholder event handler---does nothing safely---soaks up lose tick. */
+/* Placeholder event handler---does nothing safely---soaks up loose tick. */
 static void
 dummy_event_handler(
 	struct peer *peer
 	)
 {
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 	if(debug) { printf("arc: dummy_event_handler() called.\n"); }
 #endif
 }
@@ -555,10 +591,10 @@ arc_event_handler(
 	)
 {
 	struct refclockproc *pp = peer->procptr;
-	register struct arcunit *up = (struct arcunit *)pp->unitptr;
+	register struct arcunit *up = pp->unitptr;
 	int i;
 	char c;
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 	if(debug > 2) { printf("arc: arc_event_handler() called.\n"); }
 #endif
 
@@ -572,12 +608,13 @@ arc_event_handler(
 		if(write(pp->io.fd, &c, 1) != 1) {
 			msyslog(LOG_NOTICE, "ARCRON: write to fd %d failed", pp->io.fd);
 		}
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 		else if(debug) { printf("arc: sent `%2.2x', fd %d.\n", c, pp->io.fd); }
 #endif
 	}
+
+	ENQUEUE(up);
 }
-#endif /* 0 */
 
 /*
  * arc_start - open the devices and initialize data for processing
@@ -590,47 +627,50 @@ arc_start(
 {
 	register struct arcunit *up;
 	struct refclockproc *pp;
+	int temp_fd;
 	int fd;
 	char device[20];
 #ifdef HAVE_TERMIOS
 	struct termios arg;
 #endif
 
-	msyslog(LOG_NOTICE, "ARCRON: %s: opening unit %d", arc_version, unit);
-#ifdef ARCRON_DEBUG
-	if(debug) {
-		printf("arc: %s: attempt to open unit %d.\n", arc_version, unit);
-	}
-#endif
-
-	/* Prevent a ridiculous device number causing overflow of device[]. */
-	if((unit < 0) || (unit > 255)) { return(0); }
+	msyslog(LOG_NOTICE, "MSF_ARCRON %s: opening unit %d",
+		arc_version, unit);
+	DPRINTF(1, ("arc: %s: attempt to open unit %d.\n", arc_version,
+		unit));
 
 	/*
 	 * Open serial port. Use CLK line discipline, if available.
 	 */
-	(void)sprintf(device, DEVICE, unit);
-	if (!(fd = refclock_open(device, SPEED, LDISC_CLK)))
-		return(0);
-#ifdef ARCRON_DEBUG
-	if(debug) { printf("arc: unit %d using open().\n", unit); }
-#endif
-	fd = open(device, OPEN_FLAGS);
-	if(fd < 0) {
-#ifdef DEBUG
-		if(debug) { printf("arc: failed [open()] to open %s.\n", device); }
-#endif
-		return(0);
+	snprintf(device, sizeof(device), DEVICE, unit);
+	temp_fd = refclock_open(device, SPEED, LDISC_CLK);
+	if (temp_fd <= 0)
+		return 0;
+	DPRINTF(1, ("arc: unit %d using tty_open().\n", unit));
+	fd = tty_open(device, OPEN_FLAGS, 0777);
+	if (fd < 0) {
+		msyslog(LOG_ERR, "MSF_ARCRON(%d): failed second open(%s, 0777): %m.\n",
+			unit, device);
+		close(temp_fd);
+		return 0;
 	}
+	close(temp_fd);
+	temp_fd = -1;
 
+#ifndef SYS_WINNT
 	fcntl(fd, F_SETFL, 0); /* clear the descriptor flags */
-#ifdef ARCRON_DEBUG
-	if(debug)
-	{ printf("Opened RS232 port with file descriptor %d.\n", fd); }
 #endif
+	DPRINTF(1, ("arc: opened RS232 port with file descriptor %d.\n", fd));
 
 #ifdef HAVE_TERMIOS
 
+	if (tcgetattr(fd, &arg) < 0) {
+		msyslog(LOG_ERR, "MSF_ARCRON(%d): tcgetattr(%s): %m.\n",
+			unit, device);
+		close(fd);
+		return 0;
+	}
+
 	arg.c_iflag = IGNBRK | ISTRIP;
 	arg.c_oflag = 0;
 	arg.c_cflag = B300 | CS8 | CREAD | CLOCAL | CSTOPB;
@@ -638,28 +678,36 @@ arc_start(
 	arg.c_cc[VMIN] = 1;
 	arg.c_cc[VTIME] = 0;
 
-	tcsetattr(fd, TCSANOW, &arg);
+	if (tcsetattr(fd, TCSANOW, &arg) < 0) {
+		msyslog(LOG_ERR, "MSF_ARCRON(%d): tcsetattr(%s): %m.\n",
+			unit, device);
+		close(fd);
+		return 0;
+	}
 
 #else
 
-	msyslog(LOG_ERR, "ARCRON: termios not supported in this driver");
+	msyslog(LOG_ERR, "ARCRON: termios required by this driver");
 	(void)close(fd);
 
 	return 0;
 
 #endif
 
-	up = (struct arcunit *) emalloc(sizeof(struct arcunit));
-	if(!up) { (void) close(fd); return(0); }
 	/* Set structure to all zeros... */
-	memset((char *)up, 0, sizeof(struct arcunit));
+	up = emalloc_zero(sizeof(*up));
 	pp = peer->procptr;
 	pp->io.clock_recv = arc_receive;
 	pp->io.srcclock = (caddr_t)peer;
 	pp->io.datalen = 0;
 	pp->io.fd = fd;
-	if(!io_addclock(&pp->io)) { (void) close(fd); free(up); return(0); }
-	pp->unitptr = (caddr_t)up;
+	if (!io_addclock(&pp->io)) {
+		close(fd);
+		pp->io.fd = -1;
+		free(up); 
+		return(0); 
+	}
+	pp->unitptr = up;
 
 	/*
 	 * Initialize miscellaneous variables
@@ -667,7 +715,27 @@ arc_start(
 	peer->precision = PRECISION;
 	peer->stratum = 2;              /* Default to stratum 2 not 0. */
 	pp->clockdesc = DESCRIPTION;
-	memcpy((char *)&pp->refid, REFID, 4);
+	if (peer->MODE > 3) {
+		msyslog(LOG_NOTICE, "ARCRON: Invalid mode %d", peer->MODE);
+		return 0;
+	}
+#ifdef DEBUG
+	if(debug) { printf("arc: mode = %d.\n", peer->MODE); }
+#endif
+	switch (peer->MODE) {
+	    case 1:
+		memcpy((char *)&pp->refid, REFID_MSF, 4);
+		break;
+	    case 2:
+		memcpy((char *)&pp->refid, REFID_DCF77, 4);
+		break;
+	    case 3:
+		memcpy((char *)&pp->refid, REFID_WWVB, 4);
+		break;
+	    default:
+		memcpy((char *)&pp->refid, REFID, 4);
+		break;
+	}
 	/* Spread out resyncs so that they should remain separated. */
 	up->next_resync = current_time + INITIAL_RESYNC_DELAY + (67*unit)%1009;
 
@@ -686,6 +754,11 @@ arc_start(
 #else
 	up->quality = MIN_CLOCK_QUALITY;/* Don't trust the clock yet. */
 #endif
+
+	peer->action = arc_event_handler;
+
+	ENQUEUE(up);
+
 	return(1);
 }
 
@@ -702,10 +775,14 @@ arc_shutdown(
 	register struct arcunit *up;
 	struct refclockproc *pp;
 
+	peer->action = dummy_event_handler;
+
 	pp = peer->procptr;
-	up = (struct arcunit *)pp->unitptr;
-	io_closeclock(&pp->io);
-	free(up);
+	up = pp->unitptr;
+	if (-1 != pp->io.fd)
+		io_closeclock(&pp->io);
+	if (NULL != up)
+		free(up);
 }
 
 /*
@@ -740,34 +817,50 @@ send_slow(
 	int sl = strlen(s);
 	int spaceleft = space_left(up);
 
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 	if(debug > 1) { printf("arc: spaceleft = %d.\n", spaceleft); }
 #endif
 	if(spaceleft < sl) { /* Should not normally happen... */
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 		msyslog(LOG_NOTICE, "ARCRON: send-buffer overrun (%d/%d)",
-		       sl, spaceleft);
+			sl, spaceleft);
 #endif
-		return(0);                      /* FAILED! */
+		return(0);			/* FAILED! */
 	}
 
 	/* Copy in the command to be sent. */
-	while(*s) { up->cmdqueue[CMDQUEUELEN - spaceleft--] = *s++; }
+	while(*s && spaceleft > 0) { up->cmdqueue[CMDQUEUELEN - spaceleft--] = *s++; }
 
 	return(1);
 }
 
 
+static int
+get2(char *p, int *val)
+{
+  if (!isdigit((int)p[0]) || !isdigit((int)p[1])) return 0;
+  *val = (p[0] - '0') * 10 + p[1] - '0';
+  return 1;
+}
+
+static int
+get1(char *p, int *val)
+{
+  if (!isdigit((int)p[0])) return 0;
+  *val = p[0] - '0';
+  return 1;
+}
+
 /* Macro indicating action we will take for different quality values. */
 #define quality_action(q) \
 (((q) == QUALITY_UNKNOWN) ?         "UNKNOWN, will use clock anyway" : \
  (((q) < MIN_CLOCK_QUALITY_OK) ? "TOO POOR, will not use clock" : \
   "OK, will use clock"))
 
-     /*
+/*
  * arc_receive - receive data from the serial interface
  */
-     static void
+static void
 arc_receive(
 	struct recvbuf *rbufp
 	)
@@ -776,15 +869,18 @@ arc_receive(
 	struct refclockproc *pp;
 	struct peer *peer;
 	char c;
-	int i, n, wday, month, bst, status;
+	int i, n, wday, month, flags, status;
 	int arc_last_offset;
+	static int quality_average = 0;
+	static int quality_sum = 0;
+	static int quality_polls = 0;
 
 	/*
 	 * Initialize pointers and read the timecode and timestamp
 	 */
 	peer = (struct peer *)rbufp->recv_srcclock;
 	pp = peer->procptr;
-	up = (struct arcunit *)pp->unitptr;
+	up = pp->unitptr;
 
 
 	/*
@@ -857,7 +953,7 @@ arc_receive(
 		  handle for tty_clk or somesuch kernel timestamper.
 		*/
 		if(arc_last_offset > LENARC) {
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 			if(debug) {
 				printf("arc: input code too long (%d cf %d); rejected.\n",
 				       arc_last_offset, LENARC);
@@ -869,7 +965,7 @@ arc_receive(
 		}
 
 		L_SUBUF(&timestamp, charoffsets[arc_last_offset]);
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 		if(debug > 1) {
 			printf(
 				"arc: %s%d char(s) rcvd, the last for lastcode[%d]; -%sms offset applied.\n",
@@ -898,7 +994,7 @@ arc_receive(
 		   L_ISGEQ(&(up->lastrec), &timestamp))
 #endif
 		{
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 			if(debug > 1) {
 				printf("arc: system timestamp captured.\n");
 #ifdef ARCRON_MULTIPLE_SAMPLES
@@ -920,9 +1016,9 @@ arc_receive(
 	/* Just in case we still have lots of rubbish in the buffer... */
 	/* ...and to avoid the same timestamp being reused by mistake, */
 	/* eg on receipt of the \r coming in on its own after the      */
-	/* timecode.                                                   */
+	/* timecode.						       */
 	if(pp->lencode >= LENARC) {
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 		if(debug && (rbufp->recv_buffer[0] != '\r'))
 		{ printf("arc: rubbish in pp->a_lastcode[].\n"); }
 #endif
@@ -947,11 +1043,12 @@ arc_receive(
 		*/
 		if((c == 'o') && (pp->lencode == 1)) {
 			L_CLR(&(up->lastrec));
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 			if(debug > 1) { printf("arc: clearing timestamp.\n"); }
 #endif
 		}
 	}
+	if (pp->lencode == 0) return;
 
 	/* Handle a quality message. */
 	if(pp->a_lastcode[0] == 'g') {
@@ -963,17 +1060,31 @@ arc_receive(
 		if(((q & 0x70) != 0x30) || ((q & 0xf) > MAX_CLOCK_QUALITY) ||
 		   ((r & 0x70) != 0x30)) {
 			/* Badly formatted response. */
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 			if(debug) { printf("arc: bad `g' response %2x %2x.\n", r, q); }
 #endif
 			return;
 		}
 		if(r == '3') { /* Only use quality value whilst sync in progress. */
-			up->quality = (q & 0xf);
+			if (up->quality_stamp < current_time) {
+				struct calendar cal;
+				l_fp new_stamp;
+			
+				get_systime (&new_stamp);
+				caljulian (new_stamp.l_ui, &cal);
+				up->quality_stamp = 
+					current_time + 60 - cal.second + 5;
+				quality_sum = 0;
+				quality_polls = 0;
+			}
+			quality_sum += (q & 0xf);
+			quality_polls++;
+			quality_average = (quality_sum / quality_polls);
 #ifdef DEBUG
-			if(debug) { printf("arc: signal quality %d.\n", up->quality); }
+			if(debug) { printf("arc: signal quality %d (%d).\n", quality_average, (q & 0xf)); }
 #endif
 		} else if( /* (r == '2') && */ up->resyncing) {
+			up->quality = quality_average;
 #ifdef DEBUG
 			if(debug)
 			{
@@ -983,10 +1094,13 @@ arc_receive(
 			}
 #endif
 			msyslog(LOG_NOTICE,
-			       "ARCRON: sync finished, signal quality %d: %s",
-			       up->quality,
-			       quality_action(up->quality));
+				"ARCRON: sync finished, signal quality %d: %s",
+				up->quality,
+				quality_action(up->quality));
 			up->resyncing = 0; /* Resync is over. */
+			quality_average = 0;
+			quality_sum = 0;
+			quality_polls = 0;
 
 #ifdef ARCRON_KEEN
 			/* Clock quality dubious; resync earlier than usual. */
@@ -1011,13 +1125,13 @@ arc_receive(
 
 
 	/* WE HAVE NOW COLLECTED ONE TIMESTAMP (phew)... */
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 	if(debug > 1) { printf("arc: NOW HAVE TIMESTAMP...\n"); }
 #endif
 
 	/* But check that we actually captured a system timestamp on it. */
 	if(L_ISZERO(&(up->lastrec))) {
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 		if(debug) { printf("arc: FAILED TO GET SYSTEM TIMESTAMP\n"); }
 #endif
 		pp->lencode = 0;
@@ -1033,20 +1147,39 @@ arc_receive(
 	pp->a_lastcode[pp->lencode] = ((up->quality == QUALITY_UNKNOWN) ?
 				       '6' : ('0' + up->quality));
 	pp->a_lastcode[pp->lencode + 1] = '\0'; /* Terminate for printf(). */
-	record_clock_stats(&peer->srcadr, pp->a_lastcode);
-	n = sscanf(pp->a_lastcode, "o%2d%2d%2d%1d%2d%2d%2d%1d%1d",
-		   &pp->hour, &pp->minute, &pp->second,
-		   &wday, &pp->day, &month, &pp->year, &bst, &status);
 
+#ifdef PRE_NTP420
+	/* We don't use the micro-/milli- second part... */
+	pp->usec = 0;
+	pp->msec = 0;
+#else
+	/* We don't use the nano-second part... */
+	pp->nsec = 0;
+#endif	
 	/* Validate format and numbers. */
-	if(n != 9) {
-#ifdef ARCRON_DEBUG
+	if (pp->a_lastcode[0] != 'o'
+		|| !get2(pp->a_lastcode + 1, &pp->hour)
+		|| !get2(pp->a_lastcode + 3, &pp->minute)
+		|| !get2(pp->a_lastcode + 5, &pp->second)
+		|| !get1(pp->a_lastcode + 7, &wday)
+		|| !get2(pp->a_lastcode + 8, &pp->day)
+		|| !get2(pp->a_lastcode + 10, &month)
+		|| !get2(pp->a_lastcode + 12, &pp->year)) {
+#ifdef DEBUG
 		/* Would expect to have caught major problems already... */
 		if(debug) { printf("arc: badly formatted data.\n"); }
 #endif
+		pp->lencode = 0;
 		refclock_report(peer, CEVNT_BADREPLY);
 		return;
 	}
+	flags = pp->a_lastcode[14];
+	status = pp->a_lastcode[15];
+#ifdef DEBUG
+	if(debug) { printf("arc: status 0x%.2x flags 0x%.2x\n", flags, status); }
+#endif
+	n = 9;
+
 	/*
 	  Validate received values at least enough to prevent internal
 	  array-bounds problems, etc.
@@ -1059,15 +1192,21 @@ arc_receive(
 	   (month < 1) || (month > 12) ||
 	   (pp->year < 0) || (pp->year > 99)) {
 		/* Data out of range. */
-		refclock_report(peer, CEVNT_BADREPLY);
-		return;
-	}
-	/* Check that BST/UTC bits are the complement of one another. */
-	if(!(bst & 2) == !(bst & 4)) {
+		pp->lencode = 0;
 		refclock_report(peer, CEVNT_BADREPLY);
 		return;
 	}
 
+
+	if(peer->MODE == 0) { /* compatiblity to original version */
+		int bst = flags;
+		/* Check that BST/UTC bits are the complement of one another. */
+		if(!(bst & 2) == !(bst & 4)) {
+			pp->lencode = 0;
+			refclock_report(peer, CEVNT_BADREPLY);
+			return;
+		}
+	}
 	if(status & 0x8) { msyslog(LOG_NOTICE, "ARCRON: battery low"); }
 
 	/* Year-2000 alert! */
@@ -1082,14 +1221,14 @@ arc_receive(
 	if(pp->year >= YEAR_PIVOT+2000-2 ) {  			/* Y2KFixes */
 		/*This should get attention B^> */
 		msyslog(LOG_NOTICE,
-		       "ARCRON: fix me!  EITHER YOUR DATE IS BADLY WRONG or else I will break soon!");
+			"ARCRON: fix me!  EITHER YOUR DATE IS BADLY WRONG or else I will break soon!");
 	}
 #ifdef DEBUG
 	if(debug) {
 		printf("arc: n=%d %02d:%02d:%02d %02d/%02d/%04d %1d %1d\n",
 		       n,
 		       pp->hour, pp->minute, pp->second,
-		       pp->day, month, pp->year, bst, status);
+		       pp->day, month, pp->year, flags, status);
 	}
 #endif
 
@@ -1109,50 +1248,168 @@ arc_receive(
 			msyslog(LOG_NOTICE, "ARCRON: signal lost");
 			pp->leap = LEAP_NOTINSYNC; /* MSF clock is free-running. */
 			up->status = status;
+			pp->lencode = 0;
 			refclock_report(peer, CEVNT_FAULT);
 			return;
 		}
 	}
 	up->status = status;
 
-	pp->day += moff[month - 1];
+	if (peer->MODE == 0) { /* compatiblity to original version */
+		int bst = flags;
 
-	if(isleap_4(pp->year) && month > 2) { pp->day++; }	/* Y2KFixes */
+		pp->day += moff[month - 1];
 
-	/* Convert to UTC if required */
-	if(bst & 2) {
-		pp->hour--;
-		if (pp->hour < 0) {
-			pp->hour = 23;
-			pp->day--;
-			/* If we try to wrap round the year (BST on 1st Jan), reject.*/
-			if(pp->day < 0) {
-				refclock_report(peer, CEVNT_BADTIME);
-				return;
+		if(isleap_4(pp->year) && month > 2) { pp->day++; }/* Y2KFixes */
+
+		/* Convert to UTC if required */
+		if(bst & 2) {
+			pp->hour--;
+			if (pp->hour < 0) {
+				pp->hour = 23;
+				pp->day--;
+				/* If we try to wrap round the year
+				 * (BST on 1st Jan), reject.*/
+				if(pp->day < 0) {
+					pp->lencode = 0;
+					refclock_report(peer, CEVNT_BADTIME);
+					return;
+				}
 			}
 		}
 	}
 
-	/* If clock signal quality is unknown, revert to default PRECISION...*/
-	if(up->quality == QUALITY_UNKNOWN) { peer->precision = PRECISION; }
-	/* ...else improve precision if flag3 is set... */
-	else {
-		peer->precision = ((pp->sloppyclockflag & CLK_FLAG3) ?
-				   HIGHPRECISION : PRECISION);
+	if(peer->MODE > 0) {
+		if(pp->sloppyclockflag & CLK_FLAG1) {
+			struct tm  local;
+			struct tm *gmtp;
+			time_t	   unixtime;
+
+			/*
+			 * Convert to GMT for sites that distribute localtime.
+			 * This means we have to do Y2K conversion on the
+			 * 2-digit year; otherwise, we get the time wrong.
+			 */
+
+			memset(&local, 0, sizeof(local));
+
+			local.tm_year  = pp->year-1900;
+			local.tm_mon   = month-1;
+			local.tm_mday  = pp->day;
+			local.tm_hour  = pp->hour;
+			local.tm_min   = pp->minute;
+			local.tm_sec   = pp->second;
+			switch (peer->MODE) {
+			    case 1:
+				local.tm_isdst = (flags & 2);
+				break;
+			    case 2:
+				local.tm_isdst = (flags & 2);
+				break;
+			    case 3:
+				switch (flags & 3) {
+				    case 0: /* It is unclear exactly when the 
+					       Arcron changes from DST->ST and 
+					       ST->DST. Testing has shown this
+					       to be irregular. For the time 
+					       being, let the OS decide. */
+					local.tm_isdst = 0;
+#ifdef DEBUG
+					if (debug)
+					    printf ("arc: DST = 00 (0)\n"); 
+#endif
+					break;
+				    case 1: /* dst->st time */
+					local.tm_isdst = -1;
+#ifdef DEBUG
+					if (debug) 
+					    printf ("arc: DST = 01 (1)\n"); 
+#endif
+					break;
+				    case 2: /* st->dst time */
+					local.tm_isdst = -1;
+#ifdef DEBUG
+					if (debug) 
+					    printf ("arc: DST = 10 (2)\n"); 
+#endif
+					break;
+				    case 3: /* dst time */
+				        local.tm_isdst = 1;
+#ifdef DEBUG
+					if (debug) 
+					    printf ("arc: DST = 11 (3)\n"); 
+#endif
+					break;
+				}
+				break;
+			    default:
+				msyslog(LOG_NOTICE, "ARCRON: Invalid mode %d",
+					peer->MODE);
+				return;
+				break;
+			}
+			unixtime = mktime (&local);
+			if ((gmtp = gmtime (&unixtime)) == NULL)
+			{
+				pp->lencode = 0;
+				refclock_report (peer, CEVNT_FAULT);
+				return;
+			}
+			pp->year = gmtp->tm_year+1900;
+			month = gmtp->tm_mon+1;
+			pp->day = ymd2yd(pp->year,month,gmtp->tm_mday);
+			/* pp->day = gmtp->tm_yday; */
+			pp->hour = gmtp->tm_hour;
+			pp->minute = gmtp->tm_min;
+			pp->second = gmtp->tm_sec;
+#ifdef DEBUG
+			if (debug)
+			{
+				printf ("arc: time is %04d/%02d/%02d %02d:%02d:%02d UTC\n",
+					pp->year,month,gmtp->tm_mday,pp->hour,pp->minute,
+					pp->second);
+			}
+#endif
+		} else 
+		{
+			/*
+			* For more rational sites distributing UTC
+			*/
+			pp->day    = ymd2yd(pp->year,month,pp->day);
+		}
+	}
+
+	if (peer->MODE == 0) { /* compatiblity to original version */
+				/* If clock signal quality is 
+				 * unknown, revert to default PRECISION...*/
+		if(up->quality == QUALITY_UNKNOWN) { 
+			peer->precision = PRECISION; 
+		} else { /* ...else improve precision if flag3 is set... */
+			peer->precision = ((pp->sloppyclockflag & CLK_FLAG3) ?
+					   HIGHPRECISION : PRECISION);
+		}
+	} else {
+		if ((status == 0x3) && (pp->sloppyclockflag & CLK_FLAG2)) {
+			peer->precision = ((pp->sloppyclockflag & CLK_FLAG3) ?
+					   HIGHPRECISION : PRECISION);
+		} else if (up->quality == QUALITY_UNKNOWN) {
+			peer->precision = PRECISION;
+		} else {
+			peer->precision = ((pp->sloppyclockflag & CLK_FLAG3) ?
+					   HIGHPRECISION : PRECISION);
+		}
 	}
 
 	/* Notice and log any change (eg from initial defaults) for flags. */
 	if(up->saved_flags != pp->sloppyclockflag) {
-#ifdef ARCRON_DEBUG
+#ifdef DEBUG
 		msyslog(LOG_NOTICE, "ARCRON: flags enabled: %s%s%s%s",
-		       ((pp->sloppyclockflag & CLK_FLAG1) ? "1" : "."),
-		       ((pp->sloppyclockflag & CLK_FLAG2) ? "2" : "."),
-		       ((pp->sloppyclockflag & CLK_FLAG3) ? "3" : "."),
-		       ((pp->sloppyclockflag & CLK_FLAG4) ? "4" : "."));
+			((pp->sloppyclockflag & CLK_FLAG1) ? "1" : "."),
+			((pp->sloppyclockflag & CLK_FLAG2) ? "2" : "."),
+			((pp->sloppyclockflag & CLK_FLAG3) ? "3" : "."),
+			((pp->sloppyclockflag & CLK_FLAG4) ? "4" : "."));
 		/* Note effects of flags changing... */
 		if(debug) {
-			printf("arc: CHOSENSAMPLES(pp) = %d.\n", CHOSENSAMPLES(pp));
-			printf("arc: NKEEP(pp) = %d.\n", NKEEP(pp));
 			printf("arc: PRECISION = %d.\n", peer->precision);
 		}
 #endif
@@ -1180,10 +1437,11 @@ arc_receive(
 #endif
 
 	if (!refclock_process(pp)) {
+		pp->lencode = 0;
 		refclock_report(peer, CEVNT_BADTIME);
 		return;
 	}
-	pp->lastref = pp->lastrec;
+	record_clock_stats(&peer->srcadr, pp->a_lastcode);
 	refclock_receive(peer);
 }
 
@@ -1191,7 +1449,7 @@ arc_receive(
 /* request_time() sends a time request to the clock with given peer. */
 /* This automatically reports a fault if necessary. */
 /* No data should be sent after this until arc_poll() returns. */
-static  void    request_time    P((int, struct peer *));
+static  void    request_time    (int, struct peer *);
 static void
 request_time(
 	int unit,
@@ -1199,14 +1457,17 @@ request_time(
 	)
 {
 	struct refclockproc *pp = peer->procptr;
-	register struct arcunit *up = (struct arcunit *)pp->unitptr;
+	register struct arcunit *up = pp->unitptr;
 #ifdef DEBUG
 	if(debug) { printf("arc: unit %d: requesting time.\n", unit); }
 #endif
 	if (!send_slow(up, pp->io.fd, "o\r")) {
-#ifdef ARCRON_DEBUG
-		msyslog(LOG_NOTICE, "ARCRON: unit %d: problem sending", unit);
+#ifdef DEBUG
+		if (debug) {
+			printf("arc: unit %d: problem sending", unit);
+		}
 #endif
+		pp->lencode = 0;
 		refclock_report(peer, CEVNT_FAULT);
 		return;
 	}
@@ -1227,9 +1488,11 @@ arc_poll(
 	int resync_needed;              /* Should we start a resync? */
 
 	pp = peer->procptr;
-	up = (struct arcunit *)pp->unitptr;
+	up = pp->unitptr;
+#if 0
 	pp->lencode = 0;
 	memset(pp->a_lastcode, 0, sizeof(pp->a_lastcode));
+#endif
 
 #if 0
 	/* Flush input. */
@@ -1237,7 +1500,8 @@ arc_poll(
 #endif
 
 	/* Resync if our next scheduled resync time is here or has passed. */
-	resync_needed = (up->next_resync <= current_time);
+	resync_needed = ( !(pp->sloppyclockflag & CLK_FLAG2) &&
+			  (up->next_resync <= current_time) );
 
 #ifdef ARCRON_LEAPSECOND_KEEN
 	/*
@@ -1305,6 +1569,7 @@ arc_poll(
 			printf("arc: clock quality %d too poor.\n", up->quality);
 		}
 #endif
+		pp->lencode = 0;
 		refclock_report(peer, CEVNT_FAULT);
 		return;
 	}
