@@ -33,11 +33,11 @@
 /*
  * Function prototypes
  */
-static	int 	fg_init 		P((int));
-static	int 	fg_start 		P((int, struct peer *));
-static	void	fg_shutdown		P((int, struct peer *));
-static	void	fg_poll		P((int, struct peer *));
-static  void    fg_receive     P((struct recvbuf *));
+static	int 	fg_init 	(int);
+static	int 	fg_start 	(int, struct peer *);
+static	void	fg_shutdown	(int, struct peer *);
+static	void	fg_poll		(int, struct peer *);
+static  void    fg_receive	(struct recvbuf *);
 
 /* 
  * Forum Graphic unit control structure
@@ -323,11 +323,8 @@ fg_receive(
                 printf ("fg: time is %04d/%03d %02d:%02d:%02d UTC\n",
                          pp->year, pp->day, pp->hour, pp->minute, pp->second);
 #endif
-
-        if (peer->stratum <= 1)
-                peer->refid = pp->refid;
         pp->disp =  (10e-6);
-	pp->lastrec = rbufp->recv_time; /* Is it better then get_systime()? */
+	pp->lastrec = rbufp->recv_time; /* Is it better than get_systime()? */
 	/* pp->leap = LEAP_NOWARNING; */
 
         /*
