@@ -3,18 +3,11 @@
  * Refclock_neoclock4x.c
  * - NeoClock4X driver for DCF77 or FIA Timecode
  *
- * Date: 2003-01-10 v1.12
+ * Date: 2009-12-04 v1.16
  *
  * see http://www.linum.com/redir/jump/id=neoclock4x&action=redir
  * for details about the NeoClock4X device
  *
- * Copyright (C) 2002-2003 by Linum Software GmbH <neoclock4x@linum.com>
- *                                    
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
- *
- *
  */
 
 #ifdef HAVE_CONFIG_H
@@ -39,9 +32,11 @@
 
 #if defined HAVE_SYS_MODEM_H
 # include <sys/modem.h>
-# define TIOCMSET MCSETA
-# define TIOCMGET MCGETA
-# define TIOCM_RTS MRTS
+# ifndef __QNXNTO__
+#  define TIOCMSET MCSETA
+#  define TIOCMGET MCGETA
+#  define TIOCM_RTS MRTS
+# endif
 #endif
 
 #ifdef HAVE_TERMIOS_H
@@ -58,22 +53,34 @@
 # include <sys/ioctl.h>
 #endif
 
+/*
+ * NTP version 4.20 change the pp->msec field to pp->nsec.
+ * To allow to support older ntp versions with this sourcefile
+ * you can define NTP_PRE_420 to allow this driver to compile
+ * with ntp version back to 4.1.2.
+ *
+ */
+#if 0
+#define NTP_PRE_420
+#endif
+
 /*
  * If you want the driver for whatever reason to not use
  * the TX line to send anything to your NeoClock4X
  * device you must tell the NTP refclock driver which
- * firmware you NeoClock4X device uses. 
+ * firmware you NeoClock4X device uses.
  *
  * If you want to enable this feature change the "#if 0"
  * line to "#if 1" and make sure that the defined firmware
  * matches the firmware off your NeoClock4X receiver!
- * 
+ *
  */
 
 #if 0
 #define NEOCLOCK4X_FIRMWARE                NEOCLOCK4X_FIRMWARE_VERSION_A
 #endif
 
+/* at this time only firmware version A is known */
 #define NEOCLOCK4X_FIRMWARE_VERSION_A      'A'
 
 #define NEOCLOCK4X_TIMECODELEN 37
@@ -95,7 +102,9 @@
 #define NEOCLOCK4X_OFFSET_ANTENNA2         33
 #define NEOCLOCK4X_OFFSET_CRC              35
 
-#define NEOCLOCK4X_DRIVER_VERSION          "1.12 (2003-01-10)"
+#define NEOCLOCK4X_DRIVER_VERSION          "1.16 (2009-12-04)"
+
+#define NSEC_TO_MILLI                      1000000
 
 struct neoclock4x_unit {
   l_fp	laststamp;	/* last receive timestamp */
@@ -103,7 +112,7 @@ struct neoclock4x_unit {
   u_long polled;	/* flag to detect noreplies */
   char	leap_status;	/* leap second flag */
   int	recvnow;
-  
+
   char  firmware[80];
   char  firmwaretag;
   char  serial[7];
@@ -122,23 +131,20 @@ struct neoclock4x_unit {
   int   utc_msec;
 };
 
-static	int	neoclock4x_start        P((int, struct peer *));
-static	void	neoclock4x_shutdown	P((int, struct peer *));
-static	void	neoclock4x_receive	P((struct recvbuf *));
-static	void	neoclock4x_poll		P((int, struct peer *));
-static	void	neoclock4x_control      P((int, struct refclockstat *, struct refclockstat *, struct peer *));
+static	int	neoclock4x_start        (int, struct peer *);
+static	void	neoclock4x_shutdown	(int, struct peer *);
+static	void	neoclock4x_receive	(struct recvbuf *);
+static	void	neoclock4x_poll		(int, struct peer *);
+static	void	neoclock4x_control      (int, struct refclockstat *, struct refclockstat *, struct peer *);
 
-static int      neol_atoi_len           P((const char str[], int *, int));
-static int      neol_hexatoi_len        P((const char str[], int *, int));
-static void     neol_jdn_to_ymd         P((unsigned long, int *, int *, int *));
-static void     neol_localtime          P((unsigned long, int* , int*, int*, int*, int*, int*));
-static unsigned long neol_mktime        P((int, int, int, int, int, int));
-#if 0
-static void     neol_mdelay             P((int));
-#endif
+static int      neol_atoi_len           (const char str[], int *, int);
+static int      neol_hexatoi_len        (const char str[], int *, int);
+static void     neol_jdn_to_ymd         (unsigned long, int *, int *, int *);
+static void     neol_localtime          (unsigned long, int* , int*, int*, int*, int*, int*);
+static unsigned long neol_mktime        (int, int, int, int, int, int);
 #if !defined(NEOCLOCK4X_FIRMWARE)
-static int      neol_query_firmware     P((int, int, char *, int));
-static int      neol_check_firmware     P((int, const char*, char *));
+static int      neol_query_firmware     (int, int, char *, int);
+static int      neol_check_firmware     (int, const char*, char *);
 #endif
 
 struct refclock refclock_neoclock4x = {
@@ -168,17 +174,41 @@ neoclock4x_start(int unit,
 #endif
 
   (void) snprintf(dev, sizeof(dev)-1, "/dev/neoclock4x-%d", unit);
-  
+
   /* LDISC_STD, LDISC_RAW
    * Open serial port. Use CLK line discipline, if available.
    */
-  fd = refclock_open(dev, B2400, LDISC_CLK);
+  fd = refclock_open(dev, B2400, LDISC_STD);
   if(fd <= 0)
     {
       return (0);
     }
 
 #if defined(HAVE_TERMIOS)
+
+#if 1
+  if(tcgetattr(fd, &termsettings) < 0)
+    {
+      msyslog(LOG_CRIT, "NeoClock4X(%d): (tcgetattr) can't query serial port settings: %m", unit);
+      (void) close(fd);
+      return (0);
+    }
+
+  /* 2400 Baud 8N2 */
+  termsettings.c_iflag = IGNBRK | IGNPAR | ICRNL;
+  termsettings.c_oflag = 0;
+  termsettings.c_cflag = CS8 | CSTOPB | CLOCAL | CREAD;
+  (void)cfsetispeed(&termsettings, (u_int)B2400);
+  (void)cfsetospeed(&termsettings, (u_int)B2400);
+
+  if(tcsetattr(fd, TCSANOW, &termsettings) < 0)
+    {
+      msyslog(LOG_CRIT, "NeoClock4X(%d): (tcsetattr) can't set serial port 2400 8N2: %m", unit);
+      (void) close(fd);
+      return (0);
+    }
+
+#else
   if(tcgetattr(fd, &termsettings) < 0)
     {
       msyslog(LOG_CRIT, "NeoClock4X(%d): (tcgetattr) can't query serial port settings: %m", unit);
@@ -198,6 +228,8 @@ neoclock4x_start(int unit,
       (void) close(fd);
       return (0);
     }
+#endif
+
 #elif defined(HAVE_SYSV_TTYS)
   if(ioctl(fd, TCGETA, &termsettings) < 0)
     {
@@ -205,13 +237,13 @@ neoclock4x_start(int unit,
       (void) close(fd);
       return (0);
     }
-  
+
   /* 2400 Baud 8N2 */
   termsettings.c_cflag &= ~PARENB;
   termsettings.c_cflag |= CSTOPB;
   termsettings.c_cflag &= ~CSIZE;
   termsettings.c_cflag |= CS8;
-  
+
   if(ioctl(fd, TCSETA, &termsettings) < 0)
     {
       msyslog(LOG_CRIT, "NeoClock4X(%d): (TSGETA) can't set serial port 2400 8N2: %m", unit);
@@ -223,7 +255,7 @@ neoclock4x_start(int unit,
   (void) close(fd);
   return (0);
 #endif
-  
+
 #if defined(TIOCMSET) && (defined(TIOCM_RTS) || defined(CIOCM_RTS))
   /* turn on RTS, and DTR for power supply */
   /* NeoClock4x is powered from serial line */
@@ -275,14 +307,14 @@ neoclock4x_start(int unit,
    *  (NeoClock4X timecode len)
    */
   pp->fudgetime1 = (NEOCLOCK4X_TIMECODELEN * 11) / 2400.0;
-  
+
   /*
    * Initialize miscellaneous variables
    */
   peer->precision = -10;
   peer->burst = NSTAGE;
   memcpy((char *)&pp->refid, "neol", 4);
-  
+
   up->leap_status = 0;
   up->unit = unit;
   strcpy(up->firmware, "?");
@@ -307,7 +339,7 @@ neoclock4x_start(int unit,
   strcpy(up->firmware, "(c) 2002 NEOL S.A. FRANCE / L0.01 NDF:A:* (compile time)");
   up->firmwaretag = 'A';
 #else
-  msyslog(LOG_EMERG, "NeoClock4X(%d): Unkown firmware defined at compile time for NeoClock4X",
+  msyslog(LOG_EMERG, "NeoClock4X(%d): unknown firmware defined at compile time for NeoClock4X",
 	  unit);
   (void) close(fd);
   pp->io.fd = -1;
@@ -348,7 +380,7 @@ neoclock4x_start(int unit,
       pp->unitptr = NULL;
       return (0);
     }
-  
+
   NLOG(NLOG_CLOCKINFO)
     msyslog(LOG_INFO, "NeoClock4X(%d): receiver setup successful done", unit);
 
@@ -383,7 +415,7 @@ neoclock4x_shutdown(int unit,
                     }
 #ifdef TIOCM_RTS
                   /* turn on RTS, and DTR for power supply */
-                  sl232 &= ~(TIOCM_DTR | TIOCM_RTS);	
+                  sl232 &= ~(TIOCM_DTR | TIOCM_RTS);
 #else
                   /* turn on RTS, and DTR for power supply */
                   sl232 &= ~(CIOCM_DTR | CIOCM_RTS);
@@ -418,23 +450,25 @@ neoclock4x_receive(struct recvbuf *rbufp)
   int day;
   int month;	/* ddd conversion */
   int c;
+  int dsec;
   unsigned char calc_chksum;
   int recv_chksum;
-  
+
   peer = (struct peer *)rbufp->recv_srcclock;
   pp = peer->procptr;
   up = (struct neoclock4x_unit *)pp->unitptr;
-  
+
   /* wait till poll interval is reached */
   if(0 == up->recvnow)
     return;
-  
+
   /* reset poll interval flag */
   up->recvnow = 0;
 
   /* read last received timecode */
   pp->lencode = refclock_gtlin(rbufp, pp->a_lastcode, BMAX, &pp->lastrec);
-  
+  pp->leap = LEAP_NOWARNING;
+
   if(NEOCLOCK4X_TIMECODELEN != pp->lencode)
     {
       NLOG(NLOG_CLOCKEVENT)
@@ -464,7 +498,7 @@ neoclock4x_receive(struct recvbuf *rbufp)
   /* Allow synchronization even is quartz clock is
    * never initialized.
    * WARNING: This is dangerous!
-   */  
+   */
   up->quarzstatus = pp->a_lastcode[NEOCLOCK4X_OFFSET_QUARZSTATUS];
   if(0==(pp->sloppyclockflag & CLK_FLAG2))
     {
@@ -484,9 +518,9 @@ neoclock4x_receive(struct recvbuf *rbufp)
 	msyslog(LOG_NOTICE, "NeoClock4X(%d): using uninitialized quartz clock for time synchronization: %s",
 		up->unit, pp->a_lastcode);
     }
-  
+
   /*
-   * If NeoClock4X is not synchronized to a radio clock 
+   * If NeoClock4X is not synchronized to a radio clock
    * check if we're allowed to synchronize with the quartz
    * clock.
    */
@@ -521,9 +555,13 @@ neoclock4x_receive(struct recvbuf *rbufp)
   neol_atoi_len(&pp->a_lastcode[NEOCLOCK4X_OFFSET_HOUR], &pp->hour, 2);
   neol_atoi_len(&pp->a_lastcode[NEOCLOCK4X_OFFSET_MINUTE], &pp->minute, 2);
   neol_atoi_len(&pp->a_lastcode[NEOCLOCK4X_OFFSET_SECOND], &pp->second, 2);
-  neol_atoi_len(&pp->a_lastcode[NEOCLOCK4X_OFFSET_HSEC], &pp->nsec, 2);
-  pp->nsec *= 10000; /* convert 1/100s from neoclock to real nanoseconds */
-  
+  neol_atoi_len(&pp->a_lastcode[NEOCLOCK4X_OFFSET_HSEC], &dsec, 2);
+#if defined(NTP_PRE_420)
+  pp->msec = dsec * 10; /* convert 1/100s from neoclock to real miliseconds */
+#else
+  pp->nsec = dsec * 10 * NSEC_TO_MILLI; /* convert 1/100s from neoclock to nanoseconds */
+#endif
+
   memcpy(up->radiosignal, &pp->a_lastcode[NEOCLOCK4X_OFFSET_RADIOSIGNAL], 3);
   up->radiosignal[3] = 0;
   memcpy(up->serial, &pp->a_lastcode[NEOCLOCK4X_OFFSET_SERIAL], 6);
@@ -549,7 +587,7 @@ neoclock4x_receive(struct recvbuf *rbufp)
     refclock_report(peer, CEVNT_BADDATE);
     return;
   }
-  
+
   /* Year-2000 check not needed anymore. Same problem
    * will arise at 2099 but what should we do...?
    *
@@ -562,10 +600,10 @@ neoclock4x_receive(struct recvbuf *rbufp)
   */
   pp->year += 2000;
 
-  /* adjust NeoClock4X local time to UTC */  
+  /* adjust NeoClock4X local time to UTC */
   calc_utc = neol_mktime(pp->year, month, day, pp->hour, pp->minute, pp->second);
   calc_utc -= 3600;
-  /* adjust NeoClock4X daylight saving time if needed */  
+  /* adjust NeoClock4X daylight saving time if needed */
   if('S' == up->dststatus)
     calc_utc -= 3600;
   neol_localtime(calc_utc, &pp->year, &month, &day, &pp->hour, &pp->minute, &pp->second);
@@ -575,13 +613,19 @@ neoclock4x_receive(struct recvbuf *rbufp)
   */
   pp->day = ymd2yd(pp->year, month, day);
   pp->leap = 0;
-  
+
   if(pp->sloppyclockflag & CLK_FLAG4)
     {
-      msyslog(LOG_DEBUG, "NeoClock4X(%d): calculated UTC date/time: %04d-%02d-%02d %02d:%02d:%02d.%03d",
+      msyslog(LOG_DEBUG, "NeoClock4X(%d): calculated UTC date/time: %04d-%02d-%02d %02d:%02d:%02d.%03ld",
 	      up->unit,
 	      pp->year, month, day,
-	      pp->hour, pp->minute, pp->second, pp->nsec/1000);
+	      pp->hour, pp->minute, pp->second,
+#if defined(NTP_PRE_420)
+              pp->msec
+#else
+              pp->nsec/NSEC_TO_MILLI
+#endif
+              );
     }
 
   up->utc_year   = pp->year;
@@ -590,8 +634,12 @@ neoclock4x_receive(struct recvbuf *rbufp)
   up->utc_hour   = pp->hour;
   up->utc_minute = pp->minute;
   up->utc_second = pp->second;
-  up->utc_msec   = pp->nsec/1000;
-  
+#if defined(NTP_PRE_420)
+  up->utc_msec   = pp->msec;
+#else
+  up->utc_msec   = pp->nsec/NSEC_TO_MILLI;
+#endif
+
   if(!refclock_process(pp))
     {
       NLOG(NLOG_CLOCKEVENT)
@@ -600,7 +648,10 @@ neoclock4x_receive(struct recvbuf *rbufp)
       return;
     }
   refclock_receive(peer);
-  
+
+  /* report good status */
+  refclock_report(peer, CEVNT_NOMINAL);
+
   record_clock_stats(&peer->srcadr, pp->a_lastcode);
 }
 
@@ -626,7 +677,7 @@ neoclock4x_control(int unit,
 {
   struct neoclock4x_unit *up;
   struct refclockproc *pp;
-  
+
   if(NULL == peer)
     {
       msyslog(LOG_ERR, "NeoClock4X(%d): control: unit invalid/inactive", unit);
@@ -657,7 +708,7 @@ neoclock4x_control(int unit,
 	    msyslog(LOG_NOTICE, "NeoClock4X(%d): using fudgetime1 with %0.5fs from ntp.conf.",
 		    unit, pp->fudgetime1);
 	}
-      
+
       /* notify */
       if(pp->sloppyclockflag & CLK_FLAG1)
 	{
@@ -673,15 +724,13 @@ neoclock4x_control(int unit,
 
   if(NULL != out)
     {
-      static char outstatus[800];	/* status output buffer */
       char *tt;
       char tmpbuf[80];
-      
-      outstatus[0] = '\0';
+
       out->kv_list = (struct ctl_var *)0;
       out->type    = REFCLK_NEOCLOCK4X;
-      
-      snprintf(tmpbuf, sizeof(tmpbuf)-1, 
+
+      snprintf(tmpbuf, sizeof(tmpbuf)-1,
 	       "%04d-%02d-%02d %02d:%02d:%02d.%03d",
 	       up->utc_year, up->utc_month, up->utc_day,
 	       up->utc_hour, up->utc_minute, up->utc_second,
@@ -727,32 +776,34 @@ neoclock4x_control(int unit,
     }
 }
 
-static int neol_hexatoi_len(const char str[],
-			    int *result,
-			    int maxlen)
+static int
+neol_hexatoi_len(const char str[],
+		 int *result,
+		 int maxlen)
 {
   int hexdigit;
   int i;
   int n = 0;
-  
-  for(i=0; isxdigit(str[i]) && i < maxlen; i++)
+
+  for(i=0; isxdigit((int)str[i]) && i < maxlen; i++)
     {
-      hexdigit = isdigit(str[i]) ? toupper(str[i]) - '0' : toupper(str[i]) - 'A' + 10;
+      hexdigit = isdigit((int)str[i]) ? toupper(str[i]) - '0' : toupper(str[i]) - 'A' + 10;
       n = 16 * n + hexdigit;
     }
   *result = n;
   return (n);
 }
 
-int neol_atoi_len(const char str[],
+static int
+neol_atoi_len(const char str[],
 		  int *result,
 		  int maxlen)
 {
   int digit;
   int i;
   int n = 0;
-  
-  for(i=0; isdigit(str[i]) && i < maxlen; i++)
+
+  for(i=0; isdigit((int)str[i]) && i < maxlen; i++)
     {
       digit = str[i] - '0';
       n = 10 * n + digit;
@@ -776,12 +827,13 @@ int neol_atoi_len(const char str[],
  * machines were long is 32-bit! (However, as time_t is signed, we
  * will already get problems at other places on 2038-01-19 03:14:08)
  */
-static unsigned long neol_mktime(int year,
-				 int mon,
-				 int day, 
-				 int hour,
-				 int min,
-				 int sec)
+static unsigned long
+neol_mktime(int year,
+	    int mon,
+	    int day,
+	    int hour,
+	    int min,
+	    int sec)
 {
   if (0 >= (int) (mon -= 2)) {    /* 1..12 . 11,12,1..10 */
     mon += 12;      /* Puts Feb last since it has leap day */
@@ -795,13 +847,14 @@ static unsigned long neol_mktime(int year,
           )*60 + sec; /* finally seconds */
 }
 
-static void neol_localtime(unsigned long utc,
-			   int* year,
-			   int* month,
-			   int* day,
-			   int* hour,
-			   int* min,
-			   int* sec)
+static void
+neol_localtime(unsigned long utc,
+	       int* year,
+	       int* month,
+	       int* day,
+	       int* hour,
+	       int* min,
+	       int* sec)
 {
   *sec = utc % 60;
   utc /= 60;
@@ -809,20 +862,21 @@ static void neol_localtime(unsigned long utc,
   utc /= 60;
   *hour = utc % 24;
   utc /= 24;
-  
+
   /*             JDN Date 1/1/1970 */
   neol_jdn_to_ymd(utc + 2440588L, year, month, day);
 }
 
-static void neol_jdn_to_ymd(unsigned long jdn, 
-			    int *yy,
-			    int *mm, 
-			    int *dd)
+static void
+neol_jdn_to_ymd(unsigned long jdn,
+		int *yy,
+		int *mm,
+		int *dd)
 {
   unsigned long x, z, m, d, y;
   unsigned long daysPer400Years = 146097UL;
   unsigned long fudgedDaysPer4000Years = 1460970UL + 31UL;
-  
+
   x = jdn + 68569UL;
   z = 4UL * x / daysPer400Years;
   x = x - (daysPer400Years * z + 3UL) / 4UL;
@@ -833,30 +887,12 @@ static void neol_jdn_to_ymd(unsigned long jdn,
   x = m / 11UL;
   m = m + 2UL - 12UL * x;
   y = 100UL * (z - 49UL) + y + x;
-  
+
   *yy = (int)y;
   *mm = (int)m;
   *dd = (int)d;
 }
 
-#if 0
-/* 
- *  delay in milliseconds 
- */ 
-static void 
-neol_mdelay(int milliseconds) 
-{ 
-  struct timeval tv;
-  
-  if(milliseconds)
-    { 
-      tv.tv_sec  = 0; 
-      tv.tv_usec = milliseconds * 1000; 
-      select(1, NULL, NULL, NULL, &tv); 
-    }
-}
-#endif
-
 #if !defined(NEOCLOCK4X_FIRMWARE)
 static int
 neol_query_firmware(int fd,
@@ -864,7 +900,7 @@ neol_query_firmware(int fd,
 		    char *firmware,
 		    int maxlen)
 {
-  unsigned char tmpbuf[256];
+  char tmpbuf[256];
   int len;
   int lastsearch;
   unsigned char c;
@@ -882,7 +918,7 @@ neol_query_firmware(int fd,
       /* wait a little bit */
       sleep(1);
       memset(tmpbuf, 0x00, sizeof(tmpbuf));
-      
+
       len = 0;
       lastsearch = 0;
       last_c_was_crlf = 0;
@@ -926,14 +962,17 @@ neol_query_firmware(int fd,
 	    {
 	      if(0xA9 != c) /* wait for (c) char in input stream */
 		continue;
-	      
+
 	      strcpy(tmpbuf, "(c)");
 	      len = 3;
 	      init = 0;
 	      continue;
 	    }
-	  
-	  //msyslog(LOG_NOTICE, "NeoClock4X(%d): firmware %c = %02Xh", unit, c, c);
+
+#if 0
+	  msyslog(LOG_NOTICE, "NeoClock4X(%d): firmware %c = %02Xh", unit, c, c);
+#endif
+
 	  if(0x0A == c || 0x0D == c)
 	    {
 	      if(last_c_was_crlf)
@@ -959,7 +998,7 @@ neol_query_firmware(int fd,
 	    {
 	      last_c_was_crlf = 0;
 	      if(0x00 != c)
-		tmpbuf[len++] = c;
+		tmpbuf[len++] = (char) c;
 	    }
 	  tmpbuf[len] = '\0';
 	  if(len > sizeof(tmpbuf)-5)
@@ -978,6 +1017,12 @@ neol_query_firmware(int fd,
     {
       NLOG(NLOG_CLOCKINFO)
 	msyslog(LOG_INFO, "NeoClock4X(%d): firmware version: %s", unit, firmware);
+
+      if(strstr(firmware, "/R2"))
+	{
+	  msyslog(LOG_INFO, "NeoClock4X(%d): Your NeoClock4X uses the new R2 firmware release. Please note the changed LED behaviour.", unit);
+	}
+
     }
 
   return (flag);
@@ -995,7 +1040,7 @@ neol_check_firmware(int unit,
   if(NULL != ptr)
     {
       if((strlen(firmware) - strlen(ptr)) >= 7)
-        { 
+        {
           if(':' == *(ptr+5) && '*' == *(ptr+6))
             *firmwaretag = *(ptr+4);
         }
@@ -1010,7 +1055,7 @@ neol_check_firmware(int unit,
   return (1);
 }
 #endif
-  
+
 #else
 int refclock_neoclock4x_bs;
 #endif /* REFCLOCK */
@@ -1030,7 +1075,7 @@ int refclock_neoclock4x_bs;
  * - don't assume sprintf returns an int anymore
  * - change the way the firmware version is read
  * - some customers would like to put a device called
- *   data diode to the NeoClock4X device to disable 
+ *   data diode to the NeoClock4X device to disable
  *   the write line. We need to now the firmware
  *   version even in this case. We made a compile time
  *   definition in this case. The code was previously
@@ -1047,4 +1092,33 @@ int refclock_neoclock4x_bs;
  * - replaced ldiv
  * - add code to support FreeBSD
  *
+ * 2003/07/07 cjh
+ * Revision 1.13
+ * - fix reporting of clock status
+ *   changes. previously a bad clock
+ *   status was never reset.
+ *
+ * 2004/04/07 cjh
+ * Revision 1.14
+ * - open serial port in a way
+ *   AIX and some other OS can
+ *   handle much better
+ *
+ * 2006/01/11 cjh
+ * Revision 1.15
+ * - remove some unsued #ifdefs
+ * - fix nsec calculation, closes #499
+ *
+ * 2009/12/04 cjh
+ * Revision 1.16
+ * - change license to ntp COPYRIGHT notice. This should allow Debian
+ *   to add this refclock driver in further releases.
+ * - detect R2 hardware
+ *
  */
+
+
+
+
+
+
