@@ -167,15 +167,15 @@ struct true_unit {
 /*
  * Function prototypes
  */
-static	int	true_start	P((int, struct peer *));
-static	void	true_shutdown	P((int, struct peer *));
-static	void	true_receive	P((struct recvbuf *));
-static	void	true_poll	P((int, struct peer *));
-static	void	true_send	P((struct peer *, const char *));
-static	void	true_doevent	P((struct peer *, enum true_event));
+static	int	true_start	(int, struct peer *);
+static	void	true_shutdown	(int, struct peer *);
+static	void	true_receive	(struct recvbuf *);
+static	void	true_poll	(int, struct peer *);
+static	void	true_send	(struct peer *, const char *);
+static	void	true_doevent	(struct peer *, enum true_event);
 
 #ifdef CLOCK_PPS720
-static	u_long	true_sample720	P((void));
+static	u_long	true_sample720	(void);
 #endif
 
 /*
@@ -235,6 +235,7 @@ true_debug(struct peer *peer, const char *fmt, ...)
 		fprintf(up->debug, "true%d: ", up->unit);
 		vfprintf(up->debug, fmt, ap);
 	}
+	va_end(ap);
 }
 #endif /*STDC*/
 
@@ -480,7 +481,8 @@ true_receive(
 		 * Adjust the synchronize indicator according to timecode
 		 * say were OK, and then say not if we really are not OK
 		 */
-		if (synced == '>' || synced == '#' || synced == '?')
+		if (synced == '>' || synced == '#' || synced == '?'
+		    || synced == 'X')
 		    pp->leap = LEAP_NOTINSYNC;
 		else
                     pp->leap = LEAP_NOWARNING;
@@ -503,6 +505,7 @@ true_receive(
 				refclock_report(peer, CEVNT_BADTIME);
 				return;
 			}
+			off.l_uf = 0;
 #endif
 
 			pp->usec = true_sample720();
@@ -545,8 +548,9 @@ true_receive(
 		 * If clock is good we send a NOMINAL message so that
 		 * any previous BAD messages are nullified
 		 */
-                refclock_report(peer, CEVNT_NOMINAL);
+                pp->lastref = pp->lastrec;
 		refclock_receive(peer);
+		refclock_report(peer, CEVNT_NOMINAL);
 
 		/*
 		 * We have succedded in answering the poll.
