@@ -20,6 +20,11 @@
 #include "ntp_syslog.h"
 #include "ntp_select.h"
 #include "ntp_stdlib.h"
+#include <ssl_applink.c>
+
+#include "isc/net.h"
+#include "isc/result.h"
+#include "isc/sockaddr.h"
 
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
@@ -31,27 +36,18 @@
 #ifdef HAVE_POLL_H
 # include <poll.h>
 #endif
-#ifndef SYS_WINNT
-# include <netdb.h>
-# ifdef HAVE_SYS_SIGNAL_H
-#  include <sys/signal.h>
-# else
-#  include <signal.h>
-# endif
-# ifdef HAVE_SYS_IOCTL_H
-#  include <sys/ioctl.h>
-# endif
-#endif /* SYS_WINNT */
+#ifdef HAVE_SYS_SIGNAL_H
+# include <sys/signal.h>
+#endif
+#ifdef HAVE_SYS_IOCTL_H
+# include <sys/ioctl.h>
+#endif
 #ifdef HAVE_SYS_RESOURCE_H
 # include <sys/resource.h>
-#endif /* HAVE_SYS_RESOURCE_H */
+#endif
 
 #include <arpa/inet.h>
 
-#ifdef __QNXNTO__
-# include "adjtime.h"
-#endif
-
 #ifdef SYS_VXWORKS
 # include "ioLib.h"
 # include "sockLib.h"
@@ -59,6 +55,12 @@
 
 /* select wants a zero structure ... */
 struct timeval timeout = {0,0};
+#elif defined(SYS_WINNT)
+/*
+ * Windows does not abort a select select call if SIGALRM goes off
+ * so a 200 ms timeout is needed
+ */
+struct sock_timeval timeout = {0,1000000/TIMER_HZ};
 #else
 struct timeval timeout = {60,0};
 #endif
@@ -70,12 +72,11 @@ struct timeval timeout = {60,0};
 #include "recvbuff.h"
 
 #ifdef SYS_WINNT
-#define EPROTONOSUPPORT WSAEPROTONOSUPPORT
-#define EAFNOSUPPORT	WSAEAFNOSUPPORT
-#define EPFNOSUPPORT	WSAEPFNOSUPPORT
 #define TARGET_RESOLUTION 1  /* Try for 1-millisecond accuracy
 				on Windows NT timers. */
 #pragma comment(lib, "winmm")
+isc_boolean_t ntp_port_inuse(int af, u_short port);
+UINT wTimerRes;
 #endif /* SYS_WINNT */
 
 /*
@@ -96,13 +97,18 @@ static timer_t ntpdate_timerid;
  * Compatibility stuff for Version 2
  */
 #define NTP_MAXSKW	0x28f	/* 0.01 sec in fp format */
-#define NTP_MINDIST 0x51f	/* 0.02 sec in fp format */
+#define NTP_MINDIST	0x51f	/* 0.02 sec in fp format */
 #define PEER_MAXDISP	(64*FP_SECOND)	/* maximum dispersion (fp 64) */
 #define NTP_INFIN	15	/* max stratum, infinity a la Bellman-Ford */
 #define NTP_MAXWGT	(8*FP_SECOND)	/* maximum select weight 8 seconds */
-#define NTP_MAXLIST 5	/* maximum select list size */
+#define NTP_MAXLIST	5	/* maximum select list size */
 #define PEER_SHIFT	8	/* 8 suitable for crystal time base */
 
+/*
+ * for get_systime()
+ */
+s_char	sys_precision;		/* local clock precision (log2 s) */
+
 /*
  * Debugging flag
  */
@@ -113,14 +119,14 @@ volatile int debug = 0;
  */
 
 int ai_fam_templ;
-int nbsock;
-SOCKET fd[MAX_AF];	/* support up to 2 sockets */
+int nbsock;             /* the number of sockets used */
+SOCKET fd[MAX_AF];
 int fd_family[MAX_AF];	/* to remember the socket family */
 #ifdef HAVE_POLL_H
 struct pollfd fdmask[MAX_AF];
 #else
 fd_set fdmask;
-int maxfd;
+SOCKET maxfd;
 #endif
 int polltest = 0;
 
@@ -141,15 +147,10 @@ volatile int alarm_flag = 0;
 int simple_query = 0;
 
 /*
- * Unpriviledged port flag.
+ * Unprivileged port flag.
  */
 int unpriv_port = 0;
 
-/*
- * Time to spend measuring drift rate
- */
-int rate = 0;
-
 /*
  * Program name.
  */
@@ -202,35 +203,34 @@ int verbose = 0;
 int always_step = 0;
 int never_step = 0;
 
-int 	ntpdatemain P((int, char **));
+int 	ntpdatemain (int, char **);
 
-static	void	transmit	P((struct server *));
-static	void	receive 	P((struct recvbuf *));
-static	void	server_data P((struct server *, s_fp, l_fp *, u_fp));
-static	void	clock_filter	P((struct server *));
-static	struct server *clock_select P((void));
-static	int clock_adjust	P((void));
-static	void	addserver	P((char *));
-static	struct server *findserver P((struct sockaddr_storage *));
-		void	timer		P((void));
-static	void	init_alarm	P((void));
+static	void	transmit	(struct server *);
+static	void	receive 	(struct recvbuf *);
+static	void	server_data (struct server *, s_fp, l_fp *, u_fp);
+static	void	clock_filter	(struct server *);
+static	struct server *clock_select (void);
+static	int clock_adjust	(void);
+static	void	addserver	(char *);
+static	struct server *findserver (sockaddr_u *);
+		void	timer		(void);
+static	void	init_alarm	(void);
 #ifndef SYS_WINNT
-static	RETSIGTYPE alarming P((int));
+static	RETSIGTYPE alarming (int);
 #endif /* SYS_WINNT */
-static	void	init_io 	P((void));
-static	void	sendpkt 	P((struct sockaddr_storage *, struct pkt *, int));
-void	input_handler	P((void));
+static	void	init_io 	(void);
+static	void	sendpkt 	(sockaddr_u *, struct pkt *, int);
+void	input_handler	(void);
 
-static	int l_adj_systime	P((l_fp *));
-static	int l_step_systime	P((l_fp *));
+static	int l_adj_systime	(l_fp *);
+static	int l_step_systime	(l_fp *);
 
-static	void	printserver P((struct server *, FILE *));
+static	void	printserver (struct server *, FILE *);
 
 #ifdef SYS_WINNT
 int 	on = 1;
 WORD	wVersionRequested;
-WSADATA wsaData;
-HANDLE	TimerThreadHandle = NULL;
+WSADATA	wsaData;
 #endif /* SYS_WINNT */
 
 #ifdef NO_MAIN_ALLOWED
@@ -261,14 +261,10 @@ void clear_globals()
   simple_query = 0;
 
   /*
-   * Unpriviledged port flag.
+   * Unprivileged port flag.
    */
   unpriv_port = 0;
 
-  /*
-   * Time to spend measuring drift rate
-   */
-  rate = 0;
   /*
    * Systemwide parameters and flags
    */
@@ -294,7 +290,7 @@ void clear_globals()
 #endif
 
 #ifdef HAVE_NETINFO
-static ni_namelist *getnetinfoservers P((void));
+static ni_namelist *getnetinfoservers (void);
 #endif
 
 /*
@@ -319,7 +315,7 @@ ntpdatemain (
 	)
 {
 	int was_alarmed;
-	struct recvbuf *rbuflist;
+	int tot_recvbufs;
 	struct recvbuf *rbuf;
 	l_fp tmp;
 	int errflg;
@@ -330,26 +326,24 @@ ntpdatemain (
 	ni_namelist *netinfoservers;
 #endif
 #ifdef SYS_WINNT
-	HANDLE process_handle;
-
-	wVersionRequested = MAKEWORD(1,1);
-	if (WSAStartup(wVersionRequested, &wsaData)) {
-		netsyslog(LOG_ERR, "No useable winsock.dll: %m");
-		exit(1);
-	}
-
 	key_file = key_file_storage;
 
 	if (!ExpandEnvironmentStrings(KEYFILE, key_file, MAX_PATH))
-	{
 		msyslog(LOG_ERR, "ExpandEnvironmentStrings(KEYFILE) failed: %m\n");
-	}
+
+	ssl_applink();
 #endif /* SYS_WINNT */
 
 #ifdef NO_MAIN_ALLOWED
 	clear_globals();
 #endif
 
+	init_lib();	/* sets up ipv4_works, ipv6_works */
+
+	/* Check to see if we have IPv6. Otherwise default to IPv4 */
+	if (!ipv6_works)
+		ai_fam_templ = AF_INET;
+
 	errflg = 0;
 	progname = argv[0];
 	syslogit = 0;
@@ -357,7 +351,7 @@ ntpdatemain (
 	/*
 	 * Decode argument list
 	 */
-	while ((c = ntp_getopt(argc, argv, "46a:bBde:k:o:p:qr:st:uv")) != EOF)
+	while ((c = ntp_getopt(argc, argv, "46a:bBde:k:o:p:qst:uv")) != EOF)
 		switch (c)
 		{
 		case '4':
@@ -413,17 +407,6 @@ ntpdatemain (
 		case 'q':
 			simple_query = 1;
 			break;
-		case 'r':
-			c = atoi(ntp_optarg);
-			if (c <= 0 || c > (60 * 60)) {
-				(void) fprintf(stderr,
-					   "%s: rate (%d) is invalid: 0 - %d\n",
-					   progname, c, (60 * 60));
-				errflg++;
-			} else {
-				rate = c;
-			}
-			break;
 		case 's':
 			syslogit = 1;
 			break;
@@ -455,7 +438,7 @@ ntpdatemain (
 	
 	if (errflg) {
 		(void) fprintf(stderr,
-		    "usage: %s [-46bBdqsuv] [-a key#] [-e delay] [-k file] [-p samples] [-o version#] [-r rate] [-t timeo] server ...\n",
+		    "usage: %s [-46bBdqsuv] [-a key#] [-e delay] [-k file] [-p samples] [-o version#] [-t timeo] server ...\n",
 		    progname);
 		exit(2);
 	}
@@ -536,10 +519,8 @@ ntpdatemain (
 		}
 		authtrust(sys_authkey, 1);
 		if (!authistrusted(sys_authkey)) {
-			char buf[10];
-
-			(void) sprintf(buf, "%lu", (unsigned long)sys_authkey);
-			msyslog(LOG_ERR, "authentication key %s unknown", buf);
+			msyslog(LOG_ERR, "authentication key %lu unknown",
+				(unsigned long) sys_authkey);
 			exit(1);
 		}
 	}
@@ -558,18 +539,10 @@ ntpdatemain (
 #if defined(HAVE_BSD_NICE)
 	(void) setpriority(PRIO_PROCESS, 0, NTPDATE_PRIO);
 #endif
-#ifdef SYS_WINNT
-	process_handle = GetCurrentProcess();
-	if (!SetPriorityClass(process_handle, (DWORD) REALTIME_PRIORITY_CLASS)) {
-		msyslog(LOG_ERR, "SetPriorityClass failed: %m");
-	}
-#endif /* SYS_WINNT */
-
 
 
 	initializing = 0;
 	was_alarmed = 0;
-	rbuflist = (struct recvbuf *)0;
 
 	while (complete_servers < sys_numservers) {
 #ifdef HAVE_POLL_H
@@ -584,9 +557,9 @@ ntpdatemain (
 			was_alarmed = 1;
 			alarm_flag = 0;
 		}
-		rbuflist = getrecvbufs();	/* get received buffers */
+		tot_recvbufs = full_recvbuffs();	/* get received buffers */
 
-		if (!was_alarmed && rbuflist == (struct recvbuf *)0) {
+		if (!was_alarmed && tot_recvbufs == 0) {
 			/*
 			 * Nothing to do.	 Wait for something.
 			 */
@@ -599,26 +572,23 @@ ntpdatemain (
 #endif
 			if (nfound > 0)
 				input_handler();
-			else if (
-#ifndef SYS_WINNT
-				nfound == -1
-#else
-				nfound == SOCKET_ERROR
-#endif /* SYS_WINNT */
-				) {
+			else if (nfound == SOCKET_ERROR)
+			{
 #ifndef SYS_WINNT
 				if (errno != EINTR)
+#else
+				if (WSAGetLastError() != WSAEINTR)
 #endif
-					netsyslog(LOG_ERR,
+					msyslog(LOG_ERR,
 #ifdef HAVE_POLL_H
 						"poll() error: %m"
 #else
 						"select() error: %m"
 #endif
 						);
-			} else {
+			} else if (errno != 0) {
 #ifndef SYS_VXWORKS
-				netsyslog(LOG_DEBUG,
+				msyslog(LOG_DEBUG,
 #ifdef HAVE_POLL_H
 					"poll(): nfound = %d, error: %m",
 #else
@@ -631,18 +601,19 @@ ntpdatemain (
 				was_alarmed = 1;
 				alarm_flag = 0;
 			}
-			rbuflist = getrecvbufs();	/* get received buffers */
+			tot_recvbufs = full_recvbuffs();	/* get received buffers */
 		}
 
 		/*
 		 * Out here, signals are unblocked.  Call receive
 		 * procedure for each incoming packet.
 		 */
-		while (rbuflist != (struct recvbuf *)0) {
-			rbuf = rbuflist;
-			rbuflist = rbuf->next;
+		rbuf = get_full_recv_buffer();
+		while (rbuf != NULL)
+		{
 			receive(rbuf);
 			freerecvbuf(rbuf);
+			rbuf = get_full_recv_buffer();
 		}
 
 		/*
@@ -687,7 +658,7 @@ transmit(
 	struct pkt xpkt;
 
 	if (debug)
-		printf("transmit(%s)\n", stoa(&(server->srcadr)));
+		printf("transmit(%s)\n", stoa(&server->srcadr));
 
 	if (server->filter_nextpt < server->xmtcnt) {
 		l_fp ts;
@@ -719,7 +690,7 @@ transmit(
 	xpkt.ppoll = NTP_MINPOLL;
 	xpkt.precision = NTPDATE_PRECISION;
 	xpkt.rootdelay = htonl(NTPDATE_DISTANCE);
-	xpkt.rootdispersion = htonl(NTPDATE_DISP);
+	xpkt.rootdisp = htonl(NTPDATE_DISP);
 	xpkt.refid = htonl(NTPDATE_REFID);
 	L_CLR(&xpkt.reftime);
 	L_CLR(&xpkt.org);
@@ -738,18 +709,18 @@ transmit(
 		L_ADDUF(&server->xmt, sys_authdelay);
 		HTONL_FP(&server->xmt, &xpkt.xmt);
 		len = authencrypt(sys_authkey, (u_int32 *)&xpkt, LEN_PKT_NOMAC);
-		sendpkt(&(server->srcadr), &xpkt, (int)(LEN_PKT_NOMAC + len));
+		sendpkt(&server->srcadr, &xpkt, (int)(LEN_PKT_NOMAC + len));
 
 		if (debug > 1)
 			printf("transmit auth to %s\n",
-			   stoa(&(server->srcadr)));
+			   stoa(&server->srcadr));
 	} else {
 		get_systime(&(server->xmt));
 		HTONL_FP(&server->xmt, &xpkt.xmt);
-		sendpkt(&(server->srcadr), &xpkt, LEN_PKT_NOMAC);
+		sendpkt(&server->srcadr, &xpkt, LEN_PKT_NOMAC);
 
 		if (debug > 1)
-			printf("transmit to %s\n", stoa(&(server->srcadr)));
+			printf("transmit to %s\n", stoa(&server->srcadr));
 	}
 
 	/*
@@ -864,7 +835,7 @@ receive(
 	server->stratum = PKT_TO_STRATUM(rpkt->stratum);
 	server->precision = rpkt->precision;
 	server->rootdelay = ntohl(rpkt->rootdelay);
-	server->rootdispersion = ntohl(rpkt->rootdispersion);
+	server->rootdisp = ntohl(rpkt->rootdisp);
 	server->refid = rpkt->refid;
 	NTOHL_FP(&rpkt->reftime, &server->reftime);
 	NTOHL_FP(&rpkt->rec, &rec);
@@ -1283,7 +1254,7 @@ clock_adjust(void)
 	}
 
 	if (dostep) {
-		if (simple_query || l_step_systime(&server->offset)) {
+		if (simple_query || debug || l_step_systime(&server->offset)){
 			msyslog(LOG_NOTICE, "step time server %s offset %s sec",
 				stoa(&server->srcadr),
 				lfptoa(&server->offset, 6));
@@ -1313,6 +1284,30 @@ clock_adjust(void)
 }
 
 
+/*
+ * is_unreachable - check to see if we have a route to given destination
+ *		    (non-blocking).
+ */
+static int
+is_reachable (sockaddr_u *dst)
+{
+	SOCKET sockfd;
+
+	sockfd = socket(AF(dst), SOCK_DGRAM, 0);
+	if (sockfd == -1) {
+		return 0;
+	}
+
+	if (connect(sockfd, &dst->sa, SOCKLEN(dst))) {
+		closesocket(sockfd);
+		return 0;
+	}
+	closesocket(sockfd);
+	return 1;
+}
+
+
+
 /* XXX ELIMINATE: merge BIG slew into adj_systime in lib/systime.c */
 /*
  * addserver - determine a server's address and allocate a new structure
@@ -1325,7 +1320,7 @@ addserver(
 {
 	register struct server *server;
 	/* Address infos structure to store result of getaddrinfo */
-	struct addrinfo *addrResult;
+	struct addrinfo *addrResult, *ptr;
 	/* Address infos structure to store hints for getaddrinfo */
 	struct addrinfo hints;
 	/* Error variable for getaddrinfo */
@@ -1346,32 +1341,46 @@ addserver(
 
 	error = getaddrinfo(serv, service, &hints, &addrResult);
 	if (error != 0) {
+		/* Conduct more refined error analysis */
+		if (error == EAI_FAIL || error == EAI_AGAIN){
+			/* Name server is unusable. Exit after failing on the
+			   first server, in order to shorten the timeout caused
+			   by waiting for resolution of several servers */
+			fprintf(stderr, "Name server cannot be used, exiting");
+			msyslog(LOG_ERR, "name server cannot be used, reason: %s\n", gai_strerror(error));
+			exit(1);
+		}
 		fprintf(stderr, "Error : %s\n", gai_strerror(error));
 		msyslog(LOG_ERR, "can't find host %s\n", serv);
 		return;
 	}
 #ifdef DEBUG
 	else if (debug) {
-		fprintf(stderr, "host found : %s\n", stohost((struct sockaddr_storage*)addrResult->ai_addr));
+		fprintf(stderr, "host found : %s\n", stohost((sockaddr_u *)addrResult->ai_addr));
 	}
 #endif
 
-	server = (struct server *)emalloc(sizeof(struct server));
-	memset((char *)server, 0, sizeof(struct server));
+	/* We must get all returned server in case the first one fails */
+	for (ptr = addrResult; ptr != NULL; ptr = ptr->ai_next) {
+		if (is_reachable ((sockaddr_u *)ptr->ai_addr)) {
+			server = emalloc(sizeof(*server));
+			memset(server, 0, sizeof(*server));
 
-	/* For now we only get the first returned server of the addrinfo list */
-	memset(&(server->srcadr), 0, sizeof(struct sockaddr_storage));
-	memcpy(&(server->srcadr), addrResult->ai_addr, addrResult->ai_addrlen);
-	server->event_time = ++sys_numservers;
-	if (sys_servers == NULL)
-		sys_servers = server;
-	else {
-		struct server *sp;
+			memcpy(&server->srcadr, ptr->ai_addr, ptr->ai_addrlen);
+			server->event_time = ++sys_numservers;
+			if (sys_servers == NULL)
+				sys_servers = server;
+			else {
+				struct server *sp;
 
-		for (sp = sys_servers; sp->next_server != NULL;
-		     sp = sp->next_server) ;
-		sp->next_server = server;
+				for (sp = sys_servers; sp->next_server != NULL;
+				     sp = sp->next_server) ;
+				sp->next_server = server;
+			}
+		}
 	}
+
+	freeaddrinfo(addrResult);
 }
 
 
@@ -1381,34 +1390,25 @@ addserver(
  */
 static struct server *
 findserver(
-	struct sockaddr_storage *addr
+	sockaddr_u *addr
 	)
 {
 	struct server *server;
 	struct server *mc_server;
 
 	mc_server = NULL;
-	if (htons(((struct sockaddr_in*)addr)->sin_port) != NTP_PORT)
+	if (SRCPORT(addr) != NTP_PORT)
 		return 0;
 
 	for (server = sys_servers; server != NULL; 
 	     server = server->next_server) {
-		
-		if (memcmp(addr, &server->srcadr, SOCKLEN(addr))==0)
+		if (SOCK_EQ(addr, &server->srcadr))
 			return server;
-		/* Multicast compatibility to verify here... I'm not sure it's working */
-		if(addr->ss_family == AF_INET) {
-			if (IN_MULTICAST(ntohl(((struct sockaddr_in*)addr)->sin_addr.s_addr)))
+
+		if (AF(addr) == AF(&server->srcadr)) {
+			if (IS_MCAST(&server->srcadr))
 				mc_server = server;
 		}
-		else {
-#ifdef AF_INET6
-			if (IN6_IS_ADDR_MULTICAST(&((struct sockaddr_in6*)(&server->srcadr))->sin6_addr))
-			mc_server = server;
-#else
-			return 0;
-#endif
-		}
 	}
 
 	if (mc_server != NULL) {	
@@ -1420,10 +1420,10 @@ findserver(
 			complete_servers++;
 		}
 
-		server = (struct server *)emalloc(sizeof(struct server));
-		memset((char *)server, 0, sizeof(struct server));
+		server = emalloc(sizeof(*server));
+		memset(server, 0, sizeof(*server));
 
-		memcpy(&server->srcadr, &addr, sizeof(struct sockaddr_storage));
+		server->srcadr = *addr;
 
 		server->event_time = ++sys_numservers;
 
@@ -1483,8 +1483,18 @@ alarming(
 void CALLBACK 
 alarming(UINT uTimerID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
 {
+	UNUSED_ARG(uTimerID); UNUSED_ARG(uMsg); UNUSED_ARG(dwUser);
+	UNUSED_ARG(dw1); UNUSED_ARG(dw2);
+
 	alarm_flag++;
 }
+
+static void
+callTimeEndPeriod(void)
+{
+	timeEndPeriod( wTimerRes );
+	wTimerRes = 0;
+}
 #endif /* SYS_WINNT */
 
 
@@ -1502,7 +1512,7 @@ init_alarm(void)
 # endif
 #else
 	TIMECAPS tc;
-	UINT wTimerRes, wTimerID;
+	UINT wTimerID;
 # endif /* SYS_WINNT */
 #if defined SYS_CYGWIN32 || defined SYS_WINNT
 	HANDLE hToken;
@@ -1554,7 +1564,7 @@ init_alarm(void)
 # endif
 #if defined SYS_CYGWIN32
 	/*
-	 * Get previleges needed for fiddling with the clock
+	 * Get privileges needed for fiddling with the clock
 	 */
 
 	/* get the current process token handle */
@@ -1576,7 +1586,7 @@ init_alarm(void)
 	_tzset();
 
 	/*
-	 * Get previleges needed for fiddling with the clock
+	 * Get privileges needed for fiddling with the clock
 	 */
 
 	/* get the current process token handle */
@@ -1609,6 +1619,7 @@ init_alarm(void)
 	wTimerRes = min(max(tc.wPeriodMin, TARGET_RESOLUTION), tc.wPeriodMax);
 	/* establish the minimum timer resolution that we'll use */
 	timeBeginPeriod(wTimerRes);
+	atexit(callTimeEndPeriod);
 
 	/* start the timer event */
 	wTimerID = timeSetEvent(
@@ -1651,6 +1662,7 @@ init_io(void)
 	struct addrinfo hints;
 	char service[5];
 	int optval = 1;
+	int check_ntp_port_in_use = !debug && !simple_query && !unpriv_port;
 
 	/*
 	 * Init buffer free list and stat counters
@@ -1667,6 +1679,7 @@ init_io(void)
 	 * Init hints addrinfo structure
 	 */
 	memset(&hints, 0, sizeof(hints));
+	hints.ai_family = ai_fam_templ;
 	hints.ai_flags = AI_PASSIVE;
 	hints.ai_socktype = SOCK_DGRAM;
 
@@ -1676,6 +1689,13 @@ init_io(void)
 	       /*NOTREACHED*/
 	}
 
+#ifdef SYS_WINNT
+	if (check_ntp_port_in_use && ntp_port_inuse(AF_INET, NTP_PORT)){
+		msyslog(LOG_ERR, "the NTP socket is in use, exiting: %m");
+		exit(1);
+	}
+#endif
+
 	/* Remember the address of the addrinfo structure chain */
 	ressave = res;
 
@@ -1684,102 +1704,109 @@ init_io(void)
 	 */
 	for(nbsock = 0; (nbsock < MAX_AF) && res ; res = res->ai_next) {
 	/* create a datagram (UDP) socket */
-	   if ((fd[nbsock] = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0) {
+		fd[nbsock] = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
+		if (fd[nbsock] == SOCKET_ERROR) {
+#ifndef SYS_WINNT
 		if (errno == EPROTONOSUPPORT || errno == EAFNOSUPPORT ||
 		    errno == EPFNOSUPPORT)
+#else
+		int err = WSAGetLastError();
+		if (err == WSAEPROTONOSUPPORT || err == WSAEAFNOSUPPORT ||
+		    err == WSAEPFNOSUPPORT)
+#endif
 			continue;
-		netsyslog(LOG_ERR, "socket() failed: %m");
+		msyslog(LOG_ERR, "socket() failed: %m");
 		exit(1);
 		/*NOTREACHED*/
-	   }
-	   /* set socket to reuse address */
-	   if (setsockopt(fd[nbsock], SOL_SOCKET, SO_REUSEADDR, (void*) &optval, sizeof(optval)) < 0) {
-		   netsyslog(LOG_ERR, "setsockopt() SO_REUSEADDR failed: %m");
-		   exit(1);
-		   /*NOTREACHED*/
-	   }
+		}
+		/* set socket to reuse address */
+		if (setsockopt(fd[nbsock], SOL_SOCKET, SO_REUSEADDR, (void*) &optval, sizeof(optval)) < 0) {
+				msyslog(LOG_ERR, "setsockopt() SO_REUSEADDR failed: %m");
+				exit(1);
+				/*NOTREACHED*/
+		}
 #ifdef IPV6_V6ONLY
-	   /* Restricts AF_INET6 socket to IPv6 communications (see RFC 2553bis-03) */
-	   if (res->ai_family == AF_INET6)
-		if (setsockopt(fd[nbsock], IPPROTO_IPV6, IPV6_V6ONLY, (void*) &optval, sizeof(optval)) < 0) {
-			   netsyslog(LOG_ERR, "setsockopt() IPV6_V6ONLY failed: %m");
-			   exit(1);
-			   /*NOTREACHED*/
+		/* Restricts AF_INET6 socket to IPv6 communications (see RFC 2553bis-03) */
+		if (res->ai_family == AF_INET6)
+			if (setsockopt(fd[nbsock], IPPROTO_IPV6, IPV6_V6ONLY, (void*) &optval, sizeof(optval)) < 0) {
+				   msyslog(LOG_ERR, "setsockopt() IPV6_V6ONLY failed: %m");
+					exit(1);
+					/*NOTREACHED*/
 		}
 #endif
 
-	   /* Remember the socket family in fd_family structure */
-	   fd_family[nbsock] = res->ai_family;
+		/* Remember the socket family in fd_family structure */
+		fd_family[nbsock] = res->ai_family;
 
-	/*
-	 * bind the socket to the NTP port
-	 */
-	if (!debug && !simple_query && !unpriv_port) {
-		if (bind(fd[nbsock], res->ai_addr, SOCKLEN(res->ai_addr)) < 0) {
+		/*
+		 * bind the socket to the NTP port
+		 */
+		if (check_ntp_port_in_use) {
+			if (bind(fd[nbsock], res->ai_addr, 
+				 SOCKLEN((sockaddr_u *)res->ai_addr)) < 0) {
 #ifndef SYS_WINNT
-			if (errno == EADDRINUSE)
+				if (errno == EADDRINUSE)
 #else
 				if (WSAGetLastError() == WSAEADDRINUSE)
 #endif /* SYS_WINNT */
-				netsyslog(LOG_ERR,
-					"the NTP socket is in use, exiting");
+					msyslog(LOG_ERR, "the NTP socket is in use, exiting");
 				else
-				netsyslog(LOG_ERR, "bind() fails: %m");
-			exit(1);
+					msyslog(LOG_ERR, "bind() fails: %m");
+				exit(1);
+			}
 		}
-	}
 
 #ifdef HAVE_POLL_H
-	    fdmask[nbsock].fd = fd[nbsock];
-	    fdmask[nbsock].events = POLLIN;
+		fdmask[nbsock].fd = fd[nbsock];
+		fdmask[nbsock].events = POLLIN;
 #else
-	    FD_SET(fd[nbsock], &fdmask);
-	    if ((SOCKET) maxfd < fd[nbsock]+1) {
-		maxfd = fd[nbsock]+1;
-	    }
+		FD_SET(fd[nbsock], &fdmask);
+		if (maxfd < fd[nbsock]+1) {
+			maxfd = fd[nbsock]+1;
+		}
 #endif
 
-	/*
-	 * set non-blocking,
-	 */
+		/*
+		 * set non-blocking,
+		 */
 #ifndef SYS_WINNT
 # ifdef SYS_VXWORKS
-  {
-	int on = TRUE;
+		{
+			int on = TRUE;
 
-	if (ioctl(fd[nbsock],FIONBIO, &on) == ERROR) {
-	  netsyslog(LOG_ERR, "ioctl(FIONBIO) fails: %m");
-	  exit(1);
-	}
-  }
+			if (ioctl(fd[nbsock],FIONBIO, &on) == ERROR) {
+				msyslog(LOG_ERR, "ioctl(FIONBIO) fails: %m");
+				exit(1);
+			}
+		}
 # else /* not SYS_VXWORKS */
 #  if defined(O_NONBLOCK)
-	   if (fcntl(fd[nbsock], F_SETFL, O_NONBLOCK) < 0) {
-		netsyslog(LOG_ERR, "fcntl(FNDELAY|FASYNC) fails: %m");
-		exit(1);
-		/*NOTREACHED*/
-	}
+		if (fcntl(fd[nbsock], F_SETFL, O_NONBLOCK) < 0) {
+			msyslog(LOG_ERR, "fcntl(FNDELAY|FASYNC) fails: %m");
+			exit(1);
+			/*NOTREACHED*/
+		}
 #  else /* not O_NONBLOCK */
 #	if defined(FNDELAY)
-	   if (fcntl(fd[nbsock], F_SETFL, FNDELAY) < 0) {
-		netsyslog(LOG_ERR, "fcntl(FNDELAY|FASYNC) fails: %m");
-		exit(1);
-		/*NOTREACHED*/
-	}
+		if (fcntl(fd[nbsock], F_SETFL, FNDELAY) < 0) {
+			msyslog(LOG_ERR, "fcntl(FNDELAY|FASYNC) fails: %m");
+			exit(1);
+			/*NOTREACHED*/
+		}
 #	else /* FNDELAY */
 #	 include "Bletch: Need non blocking I/O"
 #	endif /* FNDELAY */
 #  endif /* not O_NONBLOCK */
 # endif /* SYS_VXWORKS */
 #else /* SYS_WINNT */
-	if (ioctlsocket(fd[nbsock], FIONBIO, (u_long *) &on) == SOCKET_ERROR) {
-		netsyslog(LOG_ERR, "ioctlsocket(FIONBIO) fails: %m");
-		exit(1);
-	}
+		if (ioctlsocket(fd[nbsock], FIONBIO, (u_long *) &on) == SOCKET_ERROR) {
+			msyslog(LOG_ERR, "ioctlsocket(FIONBIO) fails: %m");
+			exit(1);
+		}
 #endif /* SYS_WINNT */
-	nbsock++;
-    }
-    freeaddrinfo(ressave);
+		nbsock++;
+	}
+	freeaddrinfo(ressave);
 }
 
 /*
@@ -1787,14 +1814,14 @@ init_io(void)
  */
 static void
 sendpkt(
-	struct sockaddr_storage *dest,
+	sockaddr_u *dest,
 	struct pkt *pkt,
 	int len
 	)
 {
 	int i;
 	int cc;
-	SOCKET sock = 0;
+	SOCKET sock = INVALID_SOCKET;
 
 #ifdef SYS_WINNT
 	DWORD err;
@@ -1802,14 +1829,14 @@ sendpkt(
 
 	/* Find a local family compatible socket to send ntp packet to ntp server */
 	for(i = 0; (i < MAX_AF); i++) {
-		if(dest->ss_family == fd_family[i]) {
+		if(AF(dest) == fd_family[i]) {
 			sock = fd[i];
 		break;
 		}
 	}
 
-	if ( sock == 0 ) {
-		netsyslog(LOG_ERR, "cannot find family compatible socket to send ntp packet");
+	if (INVALID_SOCKET == sock) {
+		msyslog(LOG_ERR, "cannot find family compatible socket to send ntp packet");
 		exit(1);
 		/*NOTREACHED*/
 	}
@@ -1817,15 +1844,14 @@ sendpkt(
 	cc = sendto(sock, (char *)pkt, len, 0, (struct sockaddr *)dest,
 			SOCKLEN(dest));
 
+	if (SOCKET_ERROR == cc) {
 #ifndef SYS_WINNT
-	if (cc == -1) {
 		if (errno != EWOULDBLOCK && errno != ENOBUFS)
 #else
-	if (cc == SOCKET_ERROR) {
 		err = WSAGetLastError();
 		if (err != WSAEWOULDBLOCK && err != WSAENOBUFS)
 #endif /* SYS_WINNT */
-			netsyslog(LOG_ERR, "sendto(%s): %m", stohost(dest));
+			msyslog(LOG_ERR, "sendto(%s): %m", stohost(dest));
 	}
 }
 
@@ -1838,7 +1864,7 @@ input_handler(void)
 {
 	register int n;
 	register struct recvbuf *rb;
-	struct timeval tvzero;
+	struct sock_timeval tvzero;
 	int fromlen;
 	l_fp ts;
 	int i;
@@ -1877,7 +1903,7 @@ input_handler(void)
 		 * Determine which socket received data
 		 */
 
-		for(i=0; i < maxfd; i++) {
+		for(i=0; i < nbsock; i++) {
 			if(FD_ISSET(fd[i], &fds)) {
 				 fdc = fd[i];
 				 break;
@@ -1895,7 +1921,7 @@ input_handler(void)
 			return;
 		else if (n == -1) {
 			if (errno != EINTR)
-				netsyslog(LOG_ERR,
+				msyslog(LOG_ERR,
 #ifdef HAVE_POLL_H
 					"poll() error: %m"
 #else
@@ -1930,7 +1956,7 @@ input_handler(void)
 
 		rb = get_free_recv_buffer();
 
-		fromlen = sizeof(struct sockaddr_storage);
+		fromlen = sizeof(rb->recv_srcadr);
 		rb->recv_length = recvfrom(fdc, (char *)&rb->recv_pkt,
 		   sizeof(rb->recv_pkt), 0,
 		   (struct sockaddr *)&rb->recv_srcadr, &fromlen);
@@ -2200,7 +2226,7 @@ getnetinfoservers(void)
 	ni_status status;
 	void *domain;
 	ni_id confdir;
-	ni_namelist *namelist = (ni_namelist*)malloc(sizeof(ni_namelist));
+	ni_namelist *namelist = emalloc(sizeof(ni_namelist));
 
 	/* Find a time server in NetInfo */
 	if ((status = ni_open(NULL, ".", &domain)) != NI_OK) return NULL;
@@ -2223,3 +2249,34 @@ getnetinfoservers(void)
 	return(namelist);
 }
 #endif
+
+#ifdef SYS_WINNT
+isc_boolean_t ntp_port_inuse(int af, u_short port)
+{
+	/*
+	 * Check if NTP socket is already in use on this system
+	 * This is only for Windows Systems, as they tend not to fail on the real bind() below
+	 */
+	
+	SOCKET checksocket;
+	struct sockaddr_in checkservice;
+	checksocket = socket(af, SOCK_DGRAM, 0);
+	if (checksocket == INVALID_SOCKET) {
+		return (ISC_TRUE);
+	}
+
+	checkservice.sin_family = (short) AF_INET;
+	checkservice.sin_addr.s_addr = INADDR_LOOPBACK;
+	checkservice.sin_port = htons(port);
+
+	if (bind(checksocket, (struct sockaddr *)&checkservice,
+		sizeof(checkservice)) == SOCKET_ERROR) {
+		if ( WSAGetLastError() == WSAEADDRINUSE ){
+			closesocket(checksocket);
+			return (ISC_TRUE);
+		}
+	}
+	closesocket(checksocket);
+	return (ISC_FALSE);
+}
+#endif
