@@ -3,36 +3,37 @@
  */
 
 #include <stdio.h>
+#include <stddef.h>
+#include <ctype.h>
+#include <signal.h>
+#include <setjmp.h>
 
 #include "ntpdc.h"
 #include "ntp_select.h"
 #include "ntp_io.h"
 #include "ntp_stdlib.h"
-/* Don't include ISC's version of IPv6 variables and structures */
-#define ISC_IPV6_H 1
+#include "ntp_assert.h"
+#include "ntp_lineedit.h"
 #include "isc/net.h"
 #include "isc/result.h"
+#include <ssl_applink.c>
 
-#include <ctype.h>
-#include <signal.h>
-#include <setjmp.h>
-#include <netdb.h>
+#include "ntpdc-opts.h"
 
 #ifdef SYS_WINNT
+# include <Mswsock.h>
 # include <io.h>
-#else
-# define closesocket close
 #endif /* SYS_WINNT */
 
-#if defined(HAVE_LIBREADLINE) || defined (HAVE_LIBEDIT)
-# include <readline/readline.h>
-# include <readline/history.h>
-#endif /* HAVE_LIBREADLINE || HAVE_LIBEDIT */
-
 #ifdef SYS_VXWORKS
-/* vxWorks needs mode flag -casey*/
-#define open(name, flags)   open(name, flags, 0777)
-#define SERVER_PORT_NUM     123
+				/* vxWorks needs mode flag -casey*/
+# define open(name, flags)   open(name, flags, 0777)
+# define SERVER_PORT_NUM     123
+#endif
+
+/* We use COMMAND as an autogen keyword */
+#ifdef COMMAND
+# undef COMMAND
 #endif
 
 /*
@@ -47,51 +48,53 @@ static	const char *	prompt = "ntpdc> ";	/* prompt to ask him about */
  * Keyid used for authenticated requests.  Obtained on the fly.
  */
 static	u_long	info_auth_keyid;
+static int keyid_entered = 0;
 
-/*
- * Type of key md5
- */
-#define	KEY_TYPE_MD5	4
-
-static	int info_auth_keytype = KEY_TYPE_MD5;	/* MD5 */
+static	int	info_auth_keytype = NID_md5;	/* MD5 */
+static	size_t	info_auth_hashlen = 16;		/* MD5 */
 u_long	current_time;		/* needed by authkeys; not used */
 
-int		ntpdcmain	P((int,	char **));
+/*
+ * for get_systime()
+ */
+s_char	sys_precision;		/* local clock precision (log2 s) */
+
+int		ntpdcmain	(int,	char **);
 /*
  * Built in command handler declarations
  */
-static	int	openhost	P((const char *));
-static	int	sendpkt		P((char *, int));
-static	void	growpktdata	P((void));
-static	int	getresponse	P((int, int, int *, int *, char **, int));
-static	int	sendrequest	P((int, int, int, int, int, char *));
-static	void	getcmds		P((void));
-static	RETSIGTYPE abortcmd	P((int));
-static	void	docmd		P((const char *));
-static	void	tokenize	P((const char *, char **, int *));
-static	int	findcmd		P((char *, struct xcmd *, struct xcmd *, struct xcmd **));
-static	int	getarg		P((char *, int, arg_v *));
-static	int	getnetnum	P((const char *, struct sockaddr_storage *, char *, int));
-static	void	help		P((struct parse *, FILE *));
+static	int	openhost	(const char *);
+static	int	sendpkt		(void *, size_t);
+static	void	growpktdata	(void);
+static	int	getresponse	(int, int, int *, int *, char **, int);
+static	int	sendrequest	(int, int, int, u_int, size_t, char *);
+static	void	getcmds		(void);
+static	RETSIGTYPE abortcmd	(int);
+static	void	docmd		(const char *);
+static	void	tokenize	(const char *, char **, int *);
+static	int	findcmd		(char *, struct xcmd *, struct xcmd *, struct xcmd **);
+static	int	getarg		(char *, int, arg_v *);
+static	int	getnetnum	(const char *, sockaddr_u *, char *, int);
+static	void	help		(struct parse *, FILE *);
 #ifdef QSORT_USES_VOID_P
-static	int	helpsort	P((const void *, const void *));
+static	int	helpsort	(const void *, const void *);
 #else
-static	int	helpsort	P((char **, char **));
+static	int	helpsort	(char **, char **);
 #endif
-static	void	printusage	P((struct xcmd *, FILE *));
-static	void	timeout		P((struct parse *, FILE *));
-static	void	my_delay	P((struct parse *, FILE *));
-static	void	host		P((struct parse *, FILE *));
-static	void	keyid		P((struct parse *, FILE *));
-static	void	keytype		P((struct parse *, FILE *));
-static	void	passwd		P((struct parse *, FILE *));
-static	void	hostnames	P((struct parse *, FILE *));
-static	void	setdebug	P((struct parse *, FILE *));
-static	void	quit		P((struct parse *, FILE *));
-static	void	version		P((struct parse *, FILE *));
-static	void	warning		P((const char *, const char *, const char *));
-static	void	error		P((const char *, const char *, const char *));
-static	u_long	getkeyid	P((const char *));
+static	void	printusage	(struct xcmd *, FILE *);
+static	void	timeout		(struct parse *, FILE *);
+static	void	my_delay	(struct parse *, FILE *);
+static	void	host		(struct parse *, FILE *);
+static	void	keyid		(struct parse *, FILE *);
+static	void	keytype		(struct parse *, FILE *);
+static	void	passwd		(struct parse *, FILE *);
+static	void	hostnames	(struct parse *, FILE *);
+static	void	setdebug	(struct parse *, FILE *);
+static	void	quit		(struct parse *, FILE *);
+static	void	version		(struct parse *, FILE *);
+static	void	warning		(const char *, const char *, const char *);
+static	void	error		(const char *, const char *, const char *);
+static	u_long	getkeyid	(const char *);
 
 
 
@@ -105,10 +108,10 @@ static	struct xcmd builtins[] = {
 	{ "help",	help,		{  OPT|NTP_STR, NO, NO, NO },
 	  { "command", "", "", "" },
 	  "tell the use and syntax of commands" },
-	{ "timeout",	timeout,	{ OPT|UINT, NO, NO, NO },
+	{ "timeout",	timeout,	{ OPT|NTP_UINT, NO, NO, NO },
 	  { "msec", "", "", "" },
 	  "set the primary receive time out" },
-	{ "delay",	my_delay,	{ OPT|INT, NO, NO, NO },
+	{ "delay",	my_delay,	{ OPT|NTP_INT, NO, NO, NO },
 	  { "msec", "", "", "" },
 	  "set the delay added to encryption time stamps" },
 	{ "host",	host,		{ OPT|NTP_STR, OPT|NTP_STR, NO, NO },
@@ -129,7 +132,7 @@ static	struct xcmd builtins[] = {
 	{ "exit",	quit,		{ NO, NO, NO, NO },
 	  { "", "", "", "" },
 	  "exit ntpdc" },
-	{ "keyid",	keyid,		{ OPT|UINT, NO, NO, NO },
+	{ "keyid",	keyid,		{ OPT|NTP_UINT, NO, NO, NO },
 	  { "key#", "", "", "" },
 	  "set/show keyid to use for authenticated requests" },
 	{ "keytype",	keytype,	{ OPT|NTP_STR, NO, NO, NO },
@@ -146,23 +149,22 @@ static	struct xcmd builtins[] = {
 /*
  * Default values we use.
  */
+#define	DEFHOST		"localhost"	/* default host name */
 #define	DEFTIMEOUT	(5)		/* 5 second time out */
 #define	DEFSTIMEOUT	(2)		/* 2 second time out after first */
 #define	DEFDELAY	0x51EB852	/* 20 milliseconds, l_fp fraction */
-#define	DEFHOST		"localhost"	/* default host name */
 #define	LENHOSTNAME	256		/* host name is 256 characters long */
 #define	MAXCMDS		100		/* maximum commands on cmd line */
 #define	MAXHOSTS	200		/* maximum hosts on cmd line */
 #define	MAXLINE		512		/* maximum line length */
-#define	MAXTOKENS	(1+1+MAXARGS+2)	/* maximum number of usable tokens */
-					/* command + -4|-6 + MAXARGS + */
-					/* redirection */
+#define	MAXTOKENS	(1+1+MAXARGS+MOREARGS+2)	/* maximum number of usable tokens */
+#define	SCREENWIDTH  	78		/* nominal screen width in columns */
 
 /*
  * Some variables used and manipulated locally
  */
-static	struct timeval tvout = { DEFTIMEOUT, 0 };	/* time out for reads */
-static	struct timeval tvsout = { DEFSTIMEOUT, 0 };	/* secondary time out */
+static	struct sock_timeval tvout = { DEFTIMEOUT, 0 };	/* time out for reads */
+static	struct sock_timeval tvsout = { DEFSTIMEOUT, 0 };/* secondary time out */
 static	l_fp delay_time;				/* delay time */
 static	char currenthost[LENHOSTNAME];			/* current host name */
 int showhostnames = 1;					/* show host names by default */
@@ -173,18 +175,6 @@ static	SOCKET sockfd;					/* fd socket is opened on */
 static	int havehost = 0;				/* set to 1 when host open */
 int s_port = 0;
 
-#if defined (SYS_WINNT) || defined (SYS_VXWORKS)
-char password[9];
-#endif /* SYS_WINNT || SYS_VXWORKS */
-
-#ifdef SYS_WINNT
-DWORD NumberOfBytesWritten;
-
-HANDLE	TimerThreadHandle = NULL;	/* 1998/06/03 - Used in ntplib/machines.c */
-void timer(void)	{  ; };	/* 1998/06/03 - Used in ntplib/machines.c */
-
-#endif /* SYS_WINNT */
-
 /*
  * Holds data returned from queries.  We allocate INITDATASIZE
  * octets to begin with, increasing this as we need to.
@@ -268,11 +258,7 @@ main(
 #ifdef SYS_VXWORKS
 void clear_globals(void)
 {
-    extern int ntp_optind;
-    extern char *ntp_optarg;
     showhostnames = 0;              /* show host names by default */
-    ntp_optind = 0;
-    ntp_optarg = 0;
     havehost = 0;                   /* set to 1 when host open */
     numcmds = 0;
     numhosts = 0;
@@ -288,10 +274,7 @@ ntpdcmain(
 	char *argv[]
 	)
 {
-	int c;
-	int errflg = 0;
 	extern int ntp_optind;
-	extern char *ntp_optarg;
 
 	delay_time.l_ui = 0;
 	delay_time.l_uf = DEFDELAY;
@@ -301,20 +284,72 @@ ntpdcmain(
 	taskPrioritySet(taskIdSelf(), 100 );
 #endif
 
-#ifdef SYS_WINNT
-	if (!Win32InitSockets())
-	{
-		fprintf(stderr, "No useable winsock.dll:");
-		exit(1);
-	}
-#endif /* SYS_WINNT */
+	init_lib();	/* sets up ipv4_works, ipv6_works */
+	ssl_applink();
 
-	/* Check to see if we have IPv6. Otherwise force the -4 flag */
-	if (isc_net_probeipv6() != ISC_R_SUCCESS) {
+	/* Check to see if we have IPv6. Otherwise default to IPv4 */
+	if (!ipv6_works)
 		ai_fam_default = AF_INET;
-	}
 
 	progname = argv[0];
+
+	{
+		int optct = optionProcess(&ntpdcOptions, argc, argv);
+		argc -= optct;
+		argv += optct;
+	}
+
+	if (HAVE_OPT(IPV4))
+		ai_fam_templ = AF_INET;
+	else if (HAVE_OPT(IPV6))
+		ai_fam_templ = AF_INET6;
+	else
+		ai_fam_templ = ai_fam_default;
+
+	if (HAVE_OPT(COMMAND)) {
+		int		cmdct = STACKCT_OPT( COMMAND );
+		const char**	cmds  = STACKLST_OPT( COMMAND );
+
+		while (cmdct-- > 0) {
+			ADDCMD(*cmds++);
+		}
+	}
+
+	debug = DESC(DEBUG_LEVEL).optOccCt;
+
+	if (HAVE_OPT(INTERACTIVE)) {
+		interactive = 1;
+	}
+
+	if (HAVE_OPT(NUMERIC)) {
+		showhostnames = 0;
+	}
+
+	if (HAVE_OPT(LISTPEERS)) {
+		ADDCMD("listpeers");
+	}
+
+	if (HAVE_OPT(PEERS)) {
+		ADDCMD("peers");
+	}
+
+	if (HAVE_OPT(SHOWPEERS)) {
+		ADDCMD("dmpeers");
+	}
+
+	if (ntp_optind == argc) {
+		ADDHOST(DEFHOST);
+	} else {
+		for (; ntp_optind < argc; ntp_optind++)
+		    ADDHOST(argv[ntp_optind]);
+	}
+
+	if (numcmds == 0 && interactive == 0
+	    && isatty(fileno(stdin)) && isatty(fileno(stderr))) {
+		interactive = 1;
+	}
+
+#if 0
 	ai_fam_templ = ai_fam_default;
 	while ((c = ntp_getopt(argc, argv, "46c:dilnps")) != EOF)
 	    switch (c) {
@@ -349,12 +384,14 @@ ntpdcmain(
 		    errflg++;
 		    break;
 	    }
+
 	if (errflg) {
 		(void) fprintf(stderr,
 			       "usage: %s [-46dilnps] [-c cmd] host ...\n",
 			       progname);
 		exit(2);
 	}
+
 	if (ntp_optind == argc) {
 		ADDHOST(DEFHOST);
 	} else {
@@ -366,6 +403,7 @@ ntpdcmain(
 	    && isatty(fileno(stdin)) && isatty(fileno(stderr))) {
 		interactive = 1;
 	}
+#endif
 
 #ifndef SYS_WINNT /* Under NT cannot handle SIGINT, WIN32 spawns a handler */
 	if (interactive)
@@ -375,12 +413,8 @@ ntpdcmain(
 	/*
 	 * Initialize the packet data buffer
 	 */
-	pktdata = (char *)malloc(INITDATASIZE);
-	if (pktdata == NULL) {
-		(void) fprintf(stderr, "%s: malloc() failed!\n", progname);
-		exit(1);
-	}
 	pktdatasize = INITDATASIZE;
+	pktdata = emalloc(INITDATASIZE);
 
 	if (numcmds == 0) {
 		(void) openhost(chosts[0]);
@@ -428,10 +462,14 @@ openhost(
 	
 	if (*cp == '[') {
 		cp++;	
-		for(i = 0; *cp != ']'; cp++, i++)
-			name[i] = *cp;	
-		name[i] = '\0';
-		hname = name;
+		for (i = 0; *cp && *cp != ']'; cp++, i++)
+			name[i] = *cp;
+		if (*cp == ']') {
+			name[i] = '\0';
+			hname = name;
+		} else {
+			return 0;
+		}
 	}	
 
 	/*
@@ -449,7 +487,11 @@ openhost(
 	hints.ai_flags = AI_NUMERICHOST;
 
 	a_info = getaddrinfo(hname, service, &hints, &ai);
-	if (a_info == EAI_NONAME || a_info == EAI_NODATA) {
+	if (a_info == EAI_NONAME
+#ifdef EAI_NODATA
+	    || a_info == EAI_NODATA
+#endif
+	   ) {
 		hints.ai_flags = AI_CANONNAME;
 #ifdef AI_ADDRCONFIG
 		hints.ai_flags |= AI_ADDRCONFIG;
@@ -463,11 +505,19 @@ openhost(
 	}
 	if (a_info != 0) {
 		(void) fprintf(stderr, "%s\n", gai_strerror(a_info));
+		if (ai != NULL)
+			freeaddrinfo(ai);
 		return 0;
 	}
 
+	/* 
+	 * getaddrinfo() has returned without error so ai should not 
+	 * be NULL.
+	 */
+	NTP_INSIST(ai != NULL);
+
 	if (ai->ai_canonname == NULL) {
-		strncpy(temphost, stoa((struct sockaddr_storage *)ai->ai_addr),
+		strncpy(temphost, stoa((sockaddr_u *)ai->ai_addr),
 		    LENHOSTNAME);
 		temphost[LENHOSTNAME-1] = '\0';
 	} else {
@@ -542,8 +592,8 @@ openhost(
 		    ai->ai_addrlen) == -1)
 #endif /* SYS_VXWORKS */
 	    error("connect", "", "");
-	if (a_info)
-		freeaddrinfo(ai);
+
+	freeaddrinfo(ai);
 	havehost = 1;
 	req_pkt_size = REQ_LEN_NOMAC;
 	impl_ver = IMPL_XNTPD;
@@ -557,11 +607,11 @@ openhost(
  */
 static int
 sendpkt(
-	char *xdata,
-	int xdatalen
+	void *	xdata,
+	size_t	xdatalen
 	)
 {
-	if (send(sockfd, xdata, (size_t)xdatalen, 0) == -1) {
+	if (send(sockfd, xdata, xdatalen, 0) == -1) {
 		warning("write to %s failed", currenthost, "");
 		return -1;
 	}
@@ -577,11 +627,7 @@ static void
 growpktdata(void)
 {
 	pktdatasize += INCDATASIZE;
-	pktdata = (char *)realloc(pktdata, (unsigned)pktdatasize);
-	if (pktdata == 0) {
-		(void) fprintf(stderr, "%s: realloc() failed!\n", progname);
-		exit(1);
-	}
+	pktdata = erealloc(pktdata, (size_t)pktdatasize);
 }
 
 
@@ -599,7 +645,7 @@ getresponse(
 	)
 {
 	struct resp_pkt rpkt;
-	struct timeval tvo;
+	struct sock_timeval tvo;
 	int items;
 	int i;
 	int size;
@@ -633,9 +679,9 @@ getresponse(
 
     again:
 	if (firstpkt)
-	    tvo = tvout;
+		tvo = tvout;
 	else
-	    tvo = tvsout;
+		tvo = tvsout;
 	
 	FD_SET(sockfd, &fds);
 	n = select(sockfd+1, &fds, (fd_set *)0, (fd_set *)0, &tvo);
@@ -746,11 +792,12 @@ getresponse(
 		pad = esize - size;
 	else 
 		pad = 0;
-	if ((datasize = items*size) > (n-RESP_HEADER_SIZE)) {
+	datasize = items * size;
+	if ((size_t)datasize > (n-RESP_HEADER_SIZE)) {
 		if (debug)
 		    printf(
-			    "Received items %d, size %d (total %d), data in packet is %d\n",
-			    items, size, datasize, n-RESP_HEADER_SIZE);
+			    "Received items %d, size %d (total %d), data in packet is %lu\n",
+			    items, size, datasize, (u_long)(n-RESP_HEADER_SIZE));
 		goto again;
 	}
 
@@ -792,7 +839,7 @@ getresponse(
 	if ((datap + datasize + (pad * items)) > (pktdata + pktdatasize)) {
 		int offset = datap - pktdata;
 		growpktdata();
-	        *rdata = pktdata; /* might have been realloced ! */
+		*rdata = pktdata; /* might have been realloced ! */
 		datap = pktdata + offset;
 	}
 	/* 
@@ -800,9 +847,9 @@ getresponse(
 	 * items.  This is so we can play nice with older implementations
 	 */
 
-	tmp_data = (char *)rpkt.data;
-	for(i = 0; i <items; i++){
-		memmove(datap, tmp_data, (unsigned)size);
+	tmp_data = rpkt.data;
+	for (i = 0; i < items; i++) {
+		memcpy(datap, tmp_data, (unsigned)size);
 		tmp_data += size;
 		memset(datap + size, 0, pad);
 		datap += size + pad;
@@ -820,88 +867,126 @@ getresponse(
 	 */
 	++numrecv;
 	if (numrecv <= lastseq)
-	    goto again;
+		goto again;
 	return INFO_OKAY;
 }
 
 
 /*
  * sendrequest - format and send a request packet
+ *
+ * Historically, ntpdc has used a fixed-size request packet regardless
+ * of the actual payload size.  When authenticating, the timestamp, key
+ * ID, and digest have been placed just before the end of the packet.
+ * With the introduction in late 2009 of support for authenticated
+ * ntpdc requests using larger 20-octet digests (vs. 16 for MD5), we
+ * come up four bytes short.
+ *
+ * To maintain interop while allowing for larger digests, the behavior
+ * is unchanged when using 16-octet digests.  For larger digests, the
+ * timestamp, key ID, and digest are placed immediately following the
+ * request payload, with the overall packet size variable.  ntpd can
+ * distinguish 16-octet digests by the overall request size being
+ * REQ_LEN_NOMAC + 4 + 16 with the auth bit enabled.  When using a
+ * longer digest, that request size should be avoided.
+ *
+ * With the form used with 20-octet and larger digests, the timestamp,
+ * key ID, and digest are located by ntpd relative to the start of the
+ * packet, and the size of the digest is then implied by the packet
+ * size.
  */
 static int
 sendrequest(
 	int implcode,
 	int reqcode,
 	int auth,
-	int qitems,
-	int qsize,
+	u_int qitems,
+	size_t qsize,
 	char *qdata
 	)
 {
 	struct req_pkt qpkt;
-	int datasize;
+	size_t	datasize;
+	size_t	reqsize;
+	u_long	key_id;
+	l_fp	ts;
+	l_fp *	ptstamp;
+	int	maclen;
+	char *	pass;
 
-	memset((char *)&qpkt, 0, sizeof qpkt);
+	memset(&qpkt, 0, sizeof(qpkt));
 
 	qpkt.rm_vn_mode = RM_VN_MODE(0, 0, 0);
 	qpkt.implementation = (u_char)implcode;
 	qpkt.request = (u_char)reqcode;
 
 	datasize = qitems * qsize;
-	if (datasize != 0 && qdata != NULL) {
-		memmove((char *)qpkt.data, qdata, (unsigned)datasize);
+	if (datasize && qdata != NULL) {
+		memcpy(qpkt.data, qdata, datasize);
 		qpkt.err_nitems = ERR_NITEMS(0, qitems);
 		qpkt.mbz_itemsize = MBZ_ITEMSIZE(qsize);
 	} else {
 		qpkt.err_nitems = ERR_NITEMS(0, 0);
-		qpkt.mbz_itemsize = MBZ_ITEMSIZE(0);
+		qpkt.mbz_itemsize = MBZ_ITEMSIZE(qsize);  /* allow for optional first item */
 	}
 
-	if (!auth) {
+	if (!auth || (keyid_entered && info_auth_keyid == 0)) {
 		qpkt.auth_seq = AUTH_SEQ(0, 0);
-		return sendpkt((char *)&qpkt, req_pkt_size);
-	} else {
-		l_fp ts;
-		int maclen = 0;
-		const char *pass = "\0";
-		struct req_pkt_tail *qpktail;
-
-		qpktail = (struct req_pkt_tail *)((char *)&qpkt + req_pkt_size
-		    + MAX_MAC_LEN - sizeof(struct req_pkt_tail));
-
-		if (info_auth_keyid == 0) {
-			maclen = getkeyid("Keyid: ");
-			if (maclen == 0) {
-				(void) fprintf(stderr,
-				    "Invalid key identifier\n");
-				return 1;
-			}
-			info_auth_keyid = maclen;
-		}
-		if (!authistrusted(info_auth_keyid)) {
-			pass = getpass("MD5 Password: ");
-			if (*pass == '\0') {
-				(void) fprintf(stderr,
-				    "Invalid password\n");
-				return (1);
-			}
-		}
-		authusekey(info_auth_keyid, info_auth_keytype, (const u_char *)pass);
-		authtrust(info_auth_keyid, 1);
-		qpkt.auth_seq = AUTH_SEQ(1, 0);
-		qpktail->keyid = htonl(info_auth_keyid);
-		get_systime(&ts);
-		L_ADD(&ts, &delay_time);
-		HTONL_FP(&ts, &qpktail->tstamp);
-		maclen = authencrypt(info_auth_keyid, (u_int32 *)&qpkt,
-		    req_pkt_size);
-		if (maclen == 0) {  
-			(void) fprintf(stderr, "Key not found\n");
-			return (1);
-		}
-		return sendpkt((char *)&qpkt, (int)(req_pkt_size + maclen));
+		return sendpkt(&qpkt, req_pkt_size);
 	}
-	/*NOTREACHED*/
+
+	if (info_auth_keyid == 0) {
+		key_id = getkeyid("Keyid: ");
+		if (!key_id) {
+			fprintf(stderr, "Invalid key identifier\n");
+			return 1;
+		}
+		info_auth_keyid = key_id;
+	}
+	if (!authistrusted(info_auth_keyid)) {
+		pass = getpass_keytype(info_auth_keytype);
+		if ('\0' == pass[0]) {
+			fprintf(stderr, "Invalid password\n");
+			return 1;
+		}
+		authusekey(info_auth_keyid, info_auth_keytype,
+			   (u_char *)pass);
+		authtrust(info_auth_keyid, 1);
+	}
+	qpkt.auth_seq = AUTH_SEQ(1, 0);
+	if (info_auth_hashlen > 16) {
+		/*
+		 * Only ntpd which expects REQ_LEN_NOMAC plus maclen
+		 * octets in an authenticated request using a 16 octet
+		 * digest (that is, a newer ntpd) will handle digests
+		 * larger than 16 octets, so for longer digests, do
+		 * not attempt to shorten the requests for downlevel
+		 * ntpd compatibility.
+		 */
+		if (REQ_LEN_NOMAC != req_pkt_size)
+			return 1;
+		reqsize = REQ_LEN_HDR + datasize + sizeof(*ptstamp);
+		/* align to 32 bits */
+		reqsize = (reqsize + 3) & ~3;
+	} else
+		reqsize = req_pkt_size;
+	ptstamp = (void *)((char *)&qpkt + reqsize);
+	ptstamp--;
+	get_systime(&ts);
+	L_ADD(&ts, &delay_time);
+	HTONL_FP(&ts, ptstamp);
+	maclen = authencrypt(info_auth_keyid, (void *)&qpkt, reqsize);
+	if (!maclen) {  
+		fprintf(stderr, "Key not found\n");
+		return 1;
+	} else if (maclen != (info_auth_hashlen + sizeof(keyid_t))) {
+		fprintf(stderr,
+			"%d octet MAC, %lu expected with %lu octet digest\n",
+			maclen, (u_long)(info_auth_hashlen + sizeof(keyid_t)),
+			(u_long)info_auth_hashlen);
+		return 1;
+	}
+	return sendpkt(&qpkt, reqsize + maclen);
 }
 
 
@@ -926,7 +1011,7 @@ doquery(
 	int res;
 	char junk[512];
 	fd_set fds;
-	struct timeval tvzero;
+	struct sock_timeval tvzero;
 
 	/*
 	 * Check to make sure host is open
@@ -960,7 +1045,7 @@ again:
 	 */
 	res = sendrequest(implcode, reqcode, auth, qitems, qsize, qdata);
 	if (res != 0)
-	    return res;
+		return res;
 	
 	/*
 	 * Get the response.  If we got a standard error, print a message
@@ -998,36 +1083,36 @@ again:
  	/* log error message if not told to be quiet */
  	if ((res > 0) && (((1 << res) & quiet_mask) == 0)) {
 		switch(res) {
-		    case INFO_ERR_IMPL:
+		case INFO_ERR_IMPL:
 			/* Give us a chance to try the older implementation. */
 			if (implcode == IMPL_XNTPD)
 				break;
 			(void) fprintf(stderr,
 				       "***Server implementation incompatable with our own\n");
 			break;
-		    case INFO_ERR_REQ:
+		case INFO_ERR_REQ:
 			(void) fprintf(stderr,
 				       "***Server doesn't implement this request\n");
 			break;
-		    case INFO_ERR_FMT:
+		case INFO_ERR_FMT:
 			(void) fprintf(stderr,
 				       "***Server reports a format error in the received packet (shouldn't happen)\n");
 			break;
-		    case INFO_ERR_NODATA:
+		case INFO_ERR_NODATA:
 			(void) fprintf(stderr,
 				       "***Server reports data not found\n");
 			break;
-		    case INFO_ERR_AUTH:
+		case INFO_ERR_AUTH:
 			(void) fprintf(stderr, "***Permission denied\n");
 			break;
-		    case ERR_TIMEOUT:
+		case ERR_TIMEOUT:
 			(void) fprintf(stderr, "***Request timed out\n");
 			break;
-		    case ERR_INCOMPLETE:
+		case ERR_INCOMPLETE:
 			(void) fprintf(stderr,
 				       "***Response from server was incomplete\n");
 			break;
-		    default:
+		default:
 			(void) fprintf(stderr,
 				       "***Server returns unknown error code %d\n", res);
 			break;
@@ -1043,33 +1128,20 @@ again:
 static void
 getcmds(void)
 {
-#if defined(HAVE_LIBREADLINE) || defined(HAVE_LIBEDIT)
-	char *line;
+	char *	line;
+	int	count;
+
+	ntp_readline_init(interactive ? prompt : NULL);
 
 	for (;;) {
-		if ((line = readline(interactive?prompt:"")) == NULL) return;
-		if (*line) add_history(line);
+		line = ntp_readline(&count);
+		if (NULL == line)
+			break;
 		docmd(line);
 		free(line);
 	}
-#else /* not (HAVE_LIBREADLINE || HAVE_LIBEDIT) */
-	char line[MAXLINE];
 
-	for (;;) {
-		if (interactive) {
-#ifdef VMS	/* work around a problem with mixing stdout & stderr */
-			fputs("",stdout);
-#endif
-			(void) fputs(prompt, stderr);
-			(void) fflush(stderr);
-		}
-
-		if (fgets(line, sizeof line, stdin) == NULL)
-		    return;
-
-		docmd(line);
-	}
-#endif /* not HAVE_LIBREADLINE || HAVE_LIBEDIT */
+	ntp_readline_uninit();
 }
 
 
@@ -1099,7 +1171,7 @@ docmd(
 	const char *cmdline
 	)
 {
-	char *tokens[1+MAXARGS+2];
+	char *tokens[1+MAXARGS+MOREARGS+2];
 	struct parse pcmd;
 	int ntok;
 	int i, ti;
@@ -1156,6 +1228,21 @@ docmd(
 		i++;
 	}
 
+	/* Any extra args are assumed to be "OPT|NTP_STR". */
+	for ( ; i < MAXARGS + MOREARGS;) {
+	     if ((i+ti) >= ntok)
+		  break;
+		rval = getarg(tokens[i+ti], (int)(OPT|NTP_STR), &pcmd.argval[i]);
+		if (rval == -1) {
+			ti++;
+			continue;
+		}
+		if (rval == 0)
+			return;
+		pcmd.nargs++;
+		i++;
+	}
+
 	i += ti;
 	if (i < ntok && *tokens[i] == '>') {
 		char *fname;
@@ -1294,6 +1381,9 @@ findcmd(
 /*
  * getarg - interpret an argument token
  *
+ * string is always set.
+ * type is set to the decoded type.
+ *
  * return:	 0 - failure
  *		 1 - success
  *		-1 - skip to next token
@@ -1309,11 +1399,15 @@ getarg(
 	char *cp, *np;
 	static const char *digits = "0123456789";
 
-	switch (code & ~OPT) {
+	memset(argp, 0, sizeof(*argp));
+
+	argp->string = str;
+	argp->type   = code & ~OPT;
+
+	switch (argp->type) {
 	    case NTP_STR:
-		argp->string = str;
 		break;
-	    case ADD:
+	    case NTP_ADD:
 		if (!strcmp("-6", str)) {
 			ai_fam_templ = AF_INET6;
 			return -1;
@@ -1325,8 +1419,8 @@ getarg(
 			return 0;
 		}
 		break;
-	    case INT:
-	    case UINT:
+	    case NTP_INT:
+	    case NTP_UINT:
 		isneg = 0;
 		np = str;
 		if (*np == '-') {
@@ -1347,7 +1441,7 @@ getarg(
 		} while (*(++np) != '\0');
 
 		if (isneg) {
-			if ((code & ~OPT) == UINT) {
+			if ((code & ~OPT) == NTP_UINT) {
 				(void) fprintf(stderr,
 					       "***Value %s should be unsigned\n", str);
 				return 0;
@@ -1379,7 +1473,7 @@ getarg(
 static int
 getnetnum(
 	const char *hname,
-	struct sockaddr_storage *num,
+	sockaddr_u *num,
 	char *fullhost,
 	int af
 	)
@@ -1387,9 +1481,7 @@ getnetnum(
 	int sockaddr_len;
 	struct addrinfo hints, *ai = NULL;
 
-	sockaddr_len = (af == AF_INET)
-			   ? sizeof(struct sockaddr_in)
-			   : sizeof(struct sockaddr_in6);
+	sockaddr_len = SIZEOF_SOCKADDR(af);
 	memset((char *)&hints, 0, sizeof(struct addrinfo));
 	hints.ai_flags = AI_CANONNAME;
 #ifdef AI_ADDRCONFIG
@@ -1399,7 +1491,7 @@ getnetnum(
 	/* decodenetnum only works with addresses */
 	if (decodenetnum(hname, num)) {
 		if (fullhost != 0) {
-			getnameinfo((struct sockaddr *)num, sockaddr_len, 
+			getnameinfo(&num->sa, sockaddr_len, 
 				    fullhost, sizeof(fullhost), NULL, 0, 
 				    NI_NUMERICHOST); 
 		}
@@ -1422,13 +1514,13 @@ getnetnum(
  */
 char *
 nntohost(
-	struct sockaddr_storage *netnum
+	sockaddr_u *netnum
 	)
 {
 	if (!showhostnames)
-	    return stoa(netnum);
+		return stoa(netnum);
 
-	if ((netnum->ss_family == AF_INET) && ISREFCLOCKADR(netnum))
+	if (ISREFCLOCKADR(netnum))
 		return refnumtoa(netnum);
 	return socktohost(netnum);
 }
@@ -1447,57 +1539,56 @@ help(
 	FILE *fp
 	)
 {
-	int i;
-	int n;
 	struct xcmd *xcp;
 	char *cmd;
-	const char *cmdsort[100];
-	int length[100];
-	int maxlength;
-	int numperline;
-	static const char *spaces = "                    ";	/* 20 spaces */
+	const char *list[100];
+	int word, words;     
+        int row, rows;
+	int col, cols;
 
 	if (pcmd->nargs == 0) {
-		n = 0;
+		words = 0;
 		for (xcp = builtins; xcp->keyword != 0; xcp++) {
 			if (*(xcp->keyword) != '?')
-			    cmdsort[n++] = xcp->keyword;
+			    list[words++] = xcp->keyword;
 		}
-		for (xcp = opcmds; xcp->keyword != 0; xcp++)
-		    cmdsort[n++] = xcp->keyword;
+                for (xcp = opcmds; xcp->keyword != 0; xcp++)
+		    list[words++] = xcp->keyword;
 
+		qsort(
 #ifdef QSORT_USES_VOID_P
-		qsort(cmdsort, (size_t)n, sizeof(char *), helpsort);
+		    (void *)
 #else
-		qsort((char *)cmdsort, (size_t)n, sizeof(char *), helpsort);
+		    (char *)
 #endif
-
-		maxlength = 0;
-		for (i = 0; i < n; i++) {
-			length[i] = strlen(cmdsort[i]);
-			if (length[i] > maxlength)
-			    maxlength = length[i];
+			(list), (size_t)(words), sizeof(char *), helpsort);
+		col = 0;
+		for (word = 0; word < words; word++) {
+			int length = strlen(list[word]);
+			if (col < length) {
+			    col = length;
+                        }
 		}
-		maxlength++;
-		numperline = 76 / maxlength;
 
-		(void) fprintf(fp, "Commands available:\n");
-		for (i = 0; i < n; i++) {
-			if ((i % numperline) == (numperline-1)
-			    || i == (n-1))
-			    (void) fprintf(fp, "%s\n", cmdsort[i]);
-			else
-			    (void) fprintf(fp, "%s%s", cmdsort[i],
-					   spaces+20-maxlength+length[i]);
+		cols = SCREENWIDTH / ++col;
+                rows = (words + cols - 1) / cols;
+
+		(void) fprintf(fp, "ntpdc commands:\n");
+
+		for (row = 0; row < rows; row++) {
+                        for (word = row; word < words; word += rows) {
+				(void) fprintf(fp, "%-*.*s", col, col-1, list[word]);
+                        }
+			(void) fprintf(fp, "\n");
 		}
 	} else {
 		cmd = pcmd->argval[0].string;
-		n = findcmd(cmd, builtins, opcmds, &xcp);
-		if (n == 0) {
+		words = findcmd(cmd, builtins, opcmds, &xcp);
+		if (words == 0) {
 			(void) fprintf(stderr,
 				       "Command `%s' is unknown\n", cmd);
 			return;
-		} else if (n >= 2) {
+		} else if (words >= 2) {
 			(void) fprintf(stderr,
 				       "Command `%s' is ambiguous\n", cmd);
 			return;
@@ -1549,7 +1640,7 @@ printusage(
 	opt46 = 0;
 	(void) fprintf(fp, "usage: %s", xcp->keyword);
 	for (i = 0; i < MAXARGS && xcp->arg[i] != NO; i++) {
-		if (opt46 == 0 && (xcp->arg[i] & ~OPT) == ADD) {
+		if (opt46 == 0 && (xcp->arg[i] & ~OPT) == NTP_ADD) {
 			(void) fprintf(fp, " [ -4|-6 ]");
 			opt46 = 1;
 		}
@@ -1675,12 +1766,15 @@ keyid(
 	)
 {
 	if (pcmd->nargs == 0) {
-		if (info_auth_keyid == 0)
+		if (info_auth_keyid == 0 && !keyid_entered)
 		    (void) fprintf(fp, "no keyid defined\n");
+		else if (info_auth_keyid == 0 && keyid_entered)
+		    (void) fprintf(fp, "no keyid will be sent\n");
 		else
 		    (void) fprintf(fp, "keyid is %lu\n", (u_long)info_auth_keyid);
 	} else {
 		info_auth_keyid = pcmd->argval[0].uval;
+		keyid_entered = 1;
 	}
 }
 
@@ -1694,23 +1788,36 @@ keytype(
 	FILE *fp
 	)
 {
-	if (pcmd->nargs == 0)
-	    fprintf(fp, "keytype is %s\n",
-		    (info_auth_keytype == KEY_TYPE_MD5) ? "MD5" : "???");
-	else
-	    switch (*(pcmd->argval[0].string)) {
-		case 'm':
-		case 'M':
-		    info_auth_keytype = KEY_TYPE_MD5;
-		    break;
+	const char *	digest_name;
+	size_t		digest_len;
+	int		key_type;
 
-		default:
-		    fprintf(fp, "keytype must be 'md5'\n");
-	    }
+	if (!pcmd->nargs) {
+		fprintf(fp, "keytype is %s with %lu octet digests\n",
+			keytype_name(info_auth_keytype),
+			(u_long)info_auth_hashlen);
+		return;
+	}
+
+	digest_name = pcmd->argval[0].string;
+	digest_len = 0;
+	key_type = keytype_from_text(digest_name, &digest_len);
+
+	if (!key_type) {
+		fprintf(fp, "keytype must be 'md5'%s\n",
+#ifdef OPENSSL
+			" or a digest type provided by OpenSSL");
+#else
+			"");
+#endif
+		return;
+	}
+
+	info_auth_keytype = key_type;
+	info_auth_hashlen = digest_len;
 }
 
 
-
 /*
  * passwd - get an authentication key
  */
@@ -1735,7 +1842,7 @@ passwd(
 			   (u_char *)pcmd->argval[0].string);
 		authtrust(info_auth_keyid, 1);
 	} else {
-		pass = getpass("MD5 Password: ");
+		pass = getpass_keytype(info_auth_keytype);
 		if (*pass == '\0')
 		    (void) fprintf(fp, "Password unchanged\n");
 		else {
@@ -1877,7 +1984,7 @@ getkeyid(
 #ifndef SYS_WINNT
 	if ((fi = fdopen(open("/dev/tty", 2), "r")) == NULL)
 #else
-	    if ((fi = _fdopen((int)GetStdHandle(STD_INPUT_HANDLE), "r")) == NULL)
+	if ((fi = _fdopen(open("CONIN$", _O_TEXT), "r")) == NULL)
 #endif /* SYS_WINNT */
 		fi = stdin;
 	    else
